@@ -1,33 +1,22 @@
 {-# LANGUAGE OverloadedStrings #-}
-{-# LANGUAGE RecordWildCards   #-}
 
 module Dhall.Import.HTTP
     ( fetchFromHttpUrl
-    , Manager
+    , originHeadersFileExpr
     ) where
 
-import Control.Monad.IO.Class (MonadIO(..))
+import Control.Monad.IO.Class           (MonadIO (..))
 import Control.Monad.Trans.State.Strict (StateT)
-import Data.ByteString (ByteString)
-import Data.CaseInsensitive (CI)
-import Data.Semigroup ((<>))
+import Data.ByteString                  (ByteString)
+import Data.CaseInsensitive             (CI)
+import Dhall.Core                       (URL (..), Expr (..))
+import Dhall.Import.Types               (Import, Status)
+import Dhall.Parser                     (Src)
+import Dhall.URL                        (renderURL)
 
-import qualified Data.Text as Text
+import qualified Data.Text      as Text
 import qualified JavaScript.XHR
 
-import Data.Void (Void)
-import Dhall.Core (URL(..))
-import Dhall.URL (renderURL)
-import Dhall.Import.Types
-
-{-| The GHCJS implementation does not require a `Manager`
-
-    The purpose of this synonym is so that "Dhall.Import.Types" can import a
-    `Manager` type from "Dhall.Import.HTTP" that does the correct thing for
-    both the GHC and GHCJS implementations
--}
-type Manager = Void
-
 fetchFromHttpUrl
     :: URL
     -> Maybe [(CI ByteString, ByteString)]
@@ -46,5 +35,8 @@ fetchFromHttpUrl childURL Nothing = do
         _   -> fail (childURLString <> " returned a non-200 status code: " <> show statusCode)
 
     return body
-fetchFromHttpUrl _ _ = do
+fetchFromHttpUrl _ _ =
     fail "Dhall does not yet support custom headers when built using GHCJS"
+
+originHeadersFileExpr :: IO (Expr Src Import)
+originHeadersFileExpr = return Missing
