@@ -1,3 +1,4 @@
+{-# LANGUAGE NamedFieldPuns    #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards   #-}
 {-# LANGUAGE ViewPatterns      #-}
@@ -7,44 +8,53 @@
 module Dhall.Freeze
     ( -- * Freeze
       freeze
+    , freezeWithManager
     , freezeExpression
+    , freezeExpressionWithManager
     , freezeImport
+    , freezeImportWithManager
     , freezeRemoteImport
+    , freezeRemoteImportWithManager
 
       -- * Types
     , Scope(..)
     , Intent(..)
     ) where
 
-import Data.Bifunctor (first)
-import Data.Monoid ((<>))
-import Data.Text
-import Dhall.Parser (Src)
-import Dhall.Pretty (CharacterSet)
-import Dhall.Syntax (Expr(..), Import(..), ImportHashed(..), ImportType(..))
+import Data.Foldable       (for_)
+import Data.List.NonEmpty  (NonEmpty)
+import Data.Maybe          (fromMaybe)
+import Dhall.Pretty        (CharacterSet, detectCharacterSet)
+import Dhall.Syntax
+    ( Expr (..)
+    , Import (..)
+    , ImportHashed (..)
+    , ImportType (..)
+    )
 import Dhall.Util
     ( Censor
-    , CheckFailed(..)
-    , Header(..)
-    , Input(..)
-    , OutputMode(..)
+    , CheckFailed (..)
+    , Header (..)
+    , Input (..)
+    , OutputMode (..)
+    , Transitivity (..)
+    , handleMultipleChecksFailed
     )
 import System.Console.ANSI (hSupportsANSI)
 
-import qualified Control.Exception                         as Exception
-import qualified Control.Monad.Trans.State.Strict          as State
-import qualified Data.Text.IO                              as Text.IO
-import qualified Data.Text.Prettyprint.Doc                 as Pretty
-import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Pretty
-import qualified Data.Text.Prettyprint.Doc.Render.Text     as Pretty.Text
-import qualified Dhall.Core                                as Core
+import qualified Control.Exception                  as Exception
+import qualified Control.Monad.Trans.State.Strict   as State
+import qualified Data.Text.IO                       as Text.IO
+import qualified Dhall.Core                         as Core
 import qualified Dhall.Import
 import qualified Dhall.Optics
-import qualified Dhall.Parser                              as Parser
 import qualified Dhall.Pretty
 import qualified Dhall.TypeCheck
-import qualified Dhall.Util                                as Util
-import qualified System.AtomicWrite.Writer.LazyText        as AtomicWrite.LazyText
+import qualified Dhall.Util                         as Util
+import qualified Prettyprinter                      as Pretty
+import qualified Prettyprinter.Render.Terminal      as Pretty
+import qualified Prettyprinter.Render.Text          as Pretty.Text
+import qualified System.AtomicWrite.Writer.LazyText as AtomicWrite.LazyText
 import qualified System.FilePath
 import qualified System.IO
 
@@ -54,7 +64,15 @@ freezeImport
     -- ^ Current working directory
     -> Import
     -> IO Import
-freezeImport directory import_ = do
+freezeImport = freezeImportWithManager Dhall.Import.defaultNewManager
+
+-- | See 'freezeImport'.
+freezeImportWithManager
+    :: IO Dhall.Import.Manager
+    -> FilePath
+    -> Import
+    -> IO Import
+freezeImportWithManager newManager directory import_ = do
     let unprotectedImport =
             import_
                 { importHashed =
@@ -63,7 +81,7 @@ freezeImport directory import_ = do
                         }
                 }
 
-    let status = Dhall.Import.emptyStatus directory
+    let status = Dhall.Import.emptyStatusWithManager newManager directory
 
     expression <- State.evalStateT (Dhall.Import.loadWith (Embed unprotectedImport)) status
 
@@ -90,35 +108,19 @@ freezeRemoteImport
     -- ^ Current working directory
     -> Import
     -> IO Import
-freezeRemoteImport directory import_ = do
+freezeRemoteImport = freezeRemoteImportWithManager Dhall.Import.defaultNewManager
+
+-- | See 'freezeRemoteImport'.
+freezeRemoteImportWithManager
+    :: IO Dhall.Import.Manager
+    -> FilePath
+    -> Import
+    -> IO Import
+freezeRemoteImportWithManager newManager directory import_ =
     case importType (importHashed import_) of
-        Remote {} -> freezeImport directory import_
+        Remote {} -> freezeImportWithManager newManager directory import_
         _         -> return import_
 
-writeExpr :: Input -> (Text, Expr Src Import) -> CharacterSet -> IO ()
-writeExpr input (header, expr) characterSet = do
-    let doc =  Pretty.pretty header
-            <> Dhall.Pretty.prettyCharacterSet characterSet expr
-            <> "\n"
-
-    let stream = Dhall.Pretty.layout doc
-
-    let unAnnotated = Pretty.unAnnotateS stream
-
-    case input of
-        InputFile file ->
-            AtomicWrite.LazyText.atomicWriteFile
-                file
-                (Pretty.Text.renderLazy unAnnotated)
-
-        StandardInput -> do
-            supportsANSI <- System.Console.ANSI.hSupportsANSI System.IO.stdout
-            if supportsANSI
-               then
-                 Pretty.renderIO System.IO.stdout (Dhall.Pretty.annToAnsiStyle <$> stream)
-               else
-                 Pretty.renderIO System.IO.stdout unAnnotated
-
 -- | Specifies which imports to freeze
 data Scope
     = OnlyRemoteImports
@@ -140,58 +142,102 @@ data Intent
 -- | Implementation of the @dhall freeze@ subcommand
 freeze
     :: OutputMode
-    -> Input
+    -> Transitivity
+    -> NonEmpty Input
     -> Scope
     -> Intent
-    -> CharacterSet
+    -> Maybe CharacterSet
     -> Censor
     -> IO ()
-freeze outputMode input scope intent characterSet censor = do
-    let directory = case input of
-            StandardInput  -> "."
-            InputFile file -> System.FilePath.takeDirectory file
+freeze = freezeWithManager Dhall.Import.defaultNewManager
 
-    let rewrite = freezeExpression directory scope intent
+-- | See 'freeze'.
+freezeWithManager
+    :: IO Dhall.Import.Manager
+    -> OutputMode
+    -> Transitivity
+    -> NonEmpty Input
+    -> Scope
+    -> Intent
+    -> Maybe CharacterSet
+    -> Censor
+    -> IO ()
+freezeWithManager newManager outputMode transitivity0 inputs scope intent chosenCharacterSet censor =
+    handleMultipleChecksFailed "freeze" "frozen" go inputs
+  where
+    go input = do
+        let directory = case input of
+                StandardInput ->
+                    "."
+                InputFile file ->
+                    System.FilePath.takeDirectory file
 
-    case outputMode of
-        Write -> do
-            (Header header, parsedExpression) <- do
-                Util.getExpressionAndHeader censor input
+        let status = Dhall.Import.emptyStatusWithManager newManager directory
 
-            frozenExpression <- rewrite parsedExpression
+        (inputName, originalText, transitivity) <- case input of
+            InputFile file -> do
+                text <- Text.IO.readFile file
 
-            writeExpr input (header, frozenExpression) characterSet
+                return (file, text, transitivity0)
 
-        Check -> do
-            originalText <- case input of
-                InputFile file -> Text.IO.readFile file
-                StandardInput  -> Text.IO.getContents
+            StandardInput -> do
+                text <- Text.IO.getContents
 
-            let name = case input of
-                    InputFile file -> file
-                    StandardInput  -> "(input)"
+                return ("(input)", text, NonTransitive)
 
-            (Header header, parsedExpression) <- do
-                Core.throws (first Parser.censor (Parser.exprAndHeaderFromText name originalText))
+        (Header header, parsedExpression) <- Util.getExpressionAndHeaderFromStdinText censor inputName originalText
 
-            frozenExpression <- rewrite parsedExpression
+        let characterSet = fromMaybe (detectCharacterSet parsedExpression) chosenCharacterSet
 
-            let doc =  Pretty.pretty header
-                    <> Dhall.Pretty.prettyCharacterSet characterSet frozenExpression
-                    <> "\n"
+        case transitivity of
+            Transitive ->
+                for_ parsedExpression $ \import_ -> do
+                    maybeFilepath <- Dhall.Import.dependencyToFile status import_
 
-            let stream = Dhall.Pretty.layout doc
+                    for_ maybeFilepath $ \filepath ->
+                        go (InputFile filepath)
 
-            let modifiedText = Pretty.Text.renderStrict stream
+            NonTransitive ->
+                return ()
 
-            if originalText == modifiedText
-                then return ()
-                else do
-                    let command = "freeze"
+        frozenExpression <- freezeExpressionWithManager newManager directory scope intent parsedExpression
 
-                    let modified = "frozen"
+        let doc =  Pretty.pretty header
+                <> Dhall.Pretty.prettyCharacterSet characterSet frozenExpression
+                <> "\n"
 
-                    Exception.throwIO CheckFailed{..}
+        let stream = Dhall.Pretty.layout doc
+
+        let modifiedText = Pretty.Text.renderStrict stream
+
+        case outputMode of
+            Write -> do
+                let unAnnotated = Pretty.unAnnotateS stream
+
+                case input of
+                    InputFile file ->
+                        if originalText == modifiedText
+                            then return ()
+                            else
+                                AtomicWrite.LazyText.atomicWriteFile
+                                    file
+                                    (Pretty.Text.renderLazy unAnnotated)
+
+                    StandardInput -> do
+                        supportsANSI <- System.Console.ANSI.hSupportsANSI System.IO.stdout
+                        if supportsANSI
+                           then
+                             Pretty.renderIO System.IO.stdout (Dhall.Pretty.annToAnsiStyle <$> stream)
+                           else
+                             Pretty.renderIO System.IO.stdout unAnnotated
+
+                return (Right ())
+
+            Check ->
+                return $
+                    if originalText == modifiedText
+                        then Right ()
+                        else Left CheckFailed{..}
 
 {-| Slightly more pure version of the `freeze` function
 
@@ -205,13 +251,23 @@ freezeExpression
     -> Intent
     -> Expr s Import
     -> IO (Expr s Import)
-freezeExpression directory scope intent expression = do
+freezeExpression = freezeExpressionWithManager Dhall.Import.defaultNewManager
+
+-- | See 'freezeExpression'.
+freezeExpressionWithManager
+    :: IO Dhall.Import.Manager
+    -> FilePath
+    -> Scope
+    -> Intent
+    -> Expr s Import
+    -> IO (Expr s Import)
+freezeExpressionWithManager newManager directory scope intent expression = do
     let freezeScope =
             case scope of
-                AllImports        -> freezeImport
-                OnlyRemoteImports -> freezeRemoteImport
+                AllImports        -> freezeImportWithManager
+                OnlyRemoteImports -> freezeRemoteImportWithManager
 
-    let freezeFunction = freezeScope directory
+    let freezeFunction = freezeScope newManager directory
 
     let cache
             -- This case is necessary because `transformOf` is a bottom-up
@@ -245,7 +301,7 @@ freezeExpression directory scope intent expression = do
                         Import{ importHashed = ImportHashed{ hash = Nothing } }
                     )
                 )
-            ) = do
+            ) =
                 {- Here we could actually compare the `_expectedHash` and
                    `_actualHash` to see if they differ, but we choose not to do
                    so and instead automatically accept the `_actualHash`.  This
@@ -280,7 +336,7 @@ freezeExpression directory scope intent expression = do
                         }
 
                 return (ImportAlt (Embed frozenImport) (Embed thawedImport))
-        cache expression_ = do
+        cache expression_ =
             return expression_
 
     case intent of
