@@ -2,7 +2,7 @@
     @dhall@ executable
 -}
 
-{-# LANGUAGE DeriveAnyClass    #-}
+{-# LANGUAGE ApplicativeDo     #-}
 {-# LANGUAGE LambdaCase        #-}
 {-# LANGUAGE NamedFieldPuns    #-}
 {-# LANGUAGE OverloadedStrings #-}
@@ -12,6 +12,7 @@ module Dhall.Main
     ( -- * Options
       Options(..)
     , Mode(..)
+    , ResolveMode(..)
     , parseOptions
     , parserInfoOptions
 
@@ -21,47 +22,60 @@ module Dhall.Main
     ) where
 
 import Control.Applicative (optional, (<|>))
-import Control.Exception (Handler(..), SomeException)
-import Control.Monad (when)
-import Data.Bifunctor (first)
-import Data.List.NonEmpty (NonEmpty(..))
-import Data.Monoid ((<>))
-import Data.Text (Text)
-import Data.Text.Prettyprint.Doc (Doc, Pretty)
-import Data.Void (Void)
-import Dhall.Freeze (Intent(..), Scope(..))
-import Dhall.Import (Imported(..), Depends(..), SemanticCacheMode(..), _semanticCacheMode)
-import Dhall.Parser (Src)
-import Dhall.Pretty (Ann, CharacterSet(..), annToAnsiStyle)
-import Dhall.TypeCheck (Censored(..), DetailedTypeError(..), TypeError)
-import Dhall.Version (dhallVersionString)
+import Control.Exception   (Handler (..), SomeException)
+import Control.Monad       (when)
+import Data.Foldable       (for_)
+import Data.List.NonEmpty  (NonEmpty (..), nonEmpty)
+import Data.Maybe          (fromMaybe)
+import Data.Text           (Text)
+import Data.Void           (Void)
+import Dhall.Freeze        (Intent (..), Scope (..))
+import Dhall.Import
+    ( Depends (..)
+    , Imported (..)
+    , SemanticCacheMode (..)
+    , _semanticCacheMode
+    )
+import Dhall.Parser        (Src)
+import Dhall.Pretty
+    ( Ann
+    , CharacterSet (..)
+    , annToAnsiStyle
+    , detectCharacterSet
+    )
+import Dhall.Schemas       (Schemas (..))
+import Dhall.TypeCheck     (Censored (..), DetailedTypeError (..), TypeError)
+import Dhall.Version       (dhallVersionString)
 import Options.Applicative (Parser, ParserInfo)
-import System.Exit (ExitCode, exitFailure)
-import System.IO (Handle)
-import Text.Dot ((.->.))
+import Prettyprinter       (Doc, Pretty)
+import System.Exit         (ExitCode, exitFailure)
+import System.IO           (Handle)
+import Text.Dot            ((.->.))
 
 import Dhall.Core
-    ( Expr(Annot)
-    , Import(..)
-    , ImportHashed(..)
-    , ImportType(..)
-    , URL(..)
+    ( Expr (Annot)
+    , Import (..)
+    , ImportHashed (..)
+    , ImportType (..)
+    , URL (..)
     , pretty
     )
 import Dhall.Util
-    ( Censor(..)
-    , CheckFailed(..)
+    ( Censor (..)
+    , CheckFailed (..)
     , Header (..)
-    , Input(..)
-    , OutputMode(..)
-    , Output(..)
+    , Input (..)
+    , Output (..)
+    , OutputMode (..)
+    , Transitivity (..)
+    , handleMultipleChecksFailed
     )
 
 import qualified Codec.CBOR.JSON
 import qualified Codec.CBOR.Read
 import qualified Codec.CBOR.Write
 import qualified Control.Exception
-import qualified Control.Monad.Trans.State.Strict          as State
+import qualified Control.Monad.Trans.State.Strict   as State
 import qualified Data.Aeson
 import qualified Data.Aeson.Encode.Pretty
 import qualified Data.ByteString.Lazy
@@ -69,51 +83,45 @@ import qualified Data.ByteString.Lazy.Char8
 import qualified Data.Map
 import qualified Data.Text
 import qualified Data.Text.IO
-import qualified Data.Text.Prettyprint.Doc                 as Pretty
-import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Pretty
-import qualified Data.Text.Prettyprint.Doc.Render.Text     as Pretty.Text
 import qualified Dhall
 import qualified Dhall.Binary
 import qualified Dhall.Core
 import qualified Dhall.Diff
-import qualified Dhall.DirectoryTree                       as DirectoryTree
+import qualified Dhall.DirectoryTree                as DirectoryTree
 import qualified Dhall.Format
 import qualified Dhall.Freeze
 import qualified Dhall.Import
 import qualified Dhall.Import.Types
 import qualified Dhall.Lint
-import qualified Dhall.Parser                              as Parser
 import qualified Dhall.Map
-import qualified Dhall.Tags
 import qualified Dhall.Pretty
 import qualified Dhall.Repl
+import qualified Dhall.Schemas
+import qualified Dhall.Tags
 import qualified Dhall.TypeCheck
 import qualified Dhall.Util
 import qualified GHC.IO.Encoding
 import qualified Options.Applicative
-import qualified System.AtomicWrite.Writer.LazyText        as AtomicWrite.LazyText
+import qualified Prettyprinter                      as Pretty
+import qualified Prettyprinter.Render.Terminal      as Pretty
+import qualified Prettyprinter.Render.Text          as Pretty.Text
+import qualified System.AtomicWrite.Writer.LazyText as AtomicWrite.LazyText
 import qualified System.Console.ANSI
-import qualified System.Exit                               as Exit
-import qualified System.IO
+import qualified System.Exit                        as Exit
 import qualified System.FilePath
+import qualified System.IO
 import qualified Text.Dot
 import qualified Text.Pretty.Simple
 
 -- | Top-level program options
 data Options = Options
-    { mode    :: Mode
-    , explain :: Bool
-    , plain   :: Bool
-    , ascii   :: Bool
-    , censor  :: Censor
+    { mode               :: Mode
+    , explain            :: Bool
+    , plain              :: Bool
+    , chosenCharacterSet :: Maybe CharacterSet
+    , censor             :: Censor
     }
 
-ignoreSemanticCache :: Mode -> Bool
-ignoreSemanticCache Default {..} = semanticCacheMode == IgnoreSemanticCache
-ignoreSemanticCache Resolve {..} = semanticCacheMode == IgnoreSemanticCache
-ignoreSemanticCache Type {..}    = semanticCacheMode == IgnoreSemanticCache
-ignoreSemanticCache _            = False
-
 -- | The subcommands for the @dhall@ executable
 data Mode
     = Default
@@ -137,11 +145,11 @@ data Mode
           }
     | Normalize { file :: Input , alpha :: Bool }
     | Repl
-    | Format { input :: Input, outputMode :: OutputMode }
-    | Freeze { input :: Input, all_ :: Bool, cache :: Bool, outputMode :: OutputMode }
-    | Hash { file :: Input }
+    | Format { deprecatedInPlace :: Bool, transitivity :: Transitivity, outputMode :: OutputMode, inputs :: NonEmpty Input }
+    | Freeze { deprecatedInPlace :: Bool, transitivity :: Transitivity, all_ :: Bool, cache :: Bool, outputMode :: OutputMode, inputs :: NonEmpty Input }
+    | Hash { file :: Input, cache :: Bool }
     | Diff { expr1 :: Text, expr2 :: Text }
-    | Lint { input :: Input, outputMode :: OutputMode }
+    | Lint { deprecatedInPlace :: Bool, transitivity :: Transitivity, outputMode :: OutputMode, inputs :: NonEmpty Input }
     | Tags
           { input :: Input
           , output :: Output
@@ -149,15 +157,20 @@ data Mode
           , followSymlinks :: Bool
           }
     | Encode { file :: Input, json :: Bool }
-    | Decode { file :: Input, json :: Bool }
-    | Text { file :: Input }
+    | Decode { file :: Input, json :: Bool, quiet :: Bool }
+    | Text { file :: Input, output :: Output }
     | DirectoryTree { file :: Input, path :: FilePath }
-    | SyntaxTree { file :: Input }
+    | Schemas { file :: Input, outputMode :: OutputMode, schemas :: Text }
+    | SyntaxTree { file :: Input, noted :: Bool }
 
+-- | This specifies how to resolve transitive dependencies
 data ResolveMode
     = Dot
+    -- ^ Generate a DOT file for @graphviz@
     | ListTransitiveDependencies
+    -- ^ List all transitive dependencies as text, one per line
     | ListImmediateDependencies
+    -- ^ List immediate dependencies as text, one per line
 
 -- | Groups of subcommands
 data Group
@@ -184,7 +197,7 @@ parseOptions =
     <$> parseMode
     <*> switch "explain" "Explain error messages in more detail"
     <*> switch "plain" "Disable syntax highlighting"
-    <*> switch "ascii" "Format code using only ASCII syntax"
+    <*> parseCharacterSet
     <*> parseCensor
   where
     switch name description =
@@ -198,6 +211,19 @@ parseOptions =
         f True  = Censor
         f False = NoCensor
 
+    parseCharacterSet =
+            Options.Applicative.flag'
+                (Just Unicode)
+                (   Options.Applicative.long "unicode"
+                <>  Options.Applicative.help "Format code using only Unicode syntax"
+                )
+        <|> Options.Applicative.flag'
+                (Just ASCII)
+                (   Options.Applicative.long "ascii"
+                <>  Options.Applicative.help "Format code using only ASCII syntax"
+                )
+        <|> pure Nothing
+
 subcommand :: Group -> String -> String -> Parser a -> Parser a
 subcommand group name description parser =
     Options.Applicative.hsubparser
@@ -218,22 +244,27 @@ parseMode =
             Manipulate
             "format"
             "Standard code formatter for the Dhall language"
-            (Format <$> parseInplace <*> parseCheck "formatted")
+            (Format <$> deprecatedInPlace <*> parseTransitiveSwitch <*> parseCheck "formatted" <*> parseFiles)
     <|> subcommand
             Manipulate
             "freeze"
             "Add integrity checks to remote import statements of an expression"
-            (Freeze <$> parseInplace <*> parseAllFlag <*> parseCacheFlag <*> parseCheck "frozen")
+            (Freeze <$> deprecatedInPlace <*> parseTransitiveSwitch <*> parseAllFlag <*> parseCacheFlag <*> parseCheck "frozen" <*> parseFiles)
     <|> subcommand
             Manipulate
             "lint"
             "Improve Dhall code by using newer language features and removing dead code"
-            (Lint <$> parseInplace <*> parseCheck "linted")
+            (Lint <$> deprecatedInPlace <*> parseTransitiveSwitch <*> parseCheck "linted" <*> parseFiles)
+    <|> subcommand
+            Manipulate
+            "rewrite-with-schemas"
+            "Simplify Dhall code using a schemas record"
+            (Dhall.Main.Schemas <$> parseInplaceNonTransitive <*> parseCheck "rewritten" <*> parseSchemasRecord)
     <|> subcommand
             Generate
             "text"
             "Render a Dhall expression that evaluates to a Text literal"
-            (Text <$> parseFile)
+            (Text <$> parseFile <*> parseOutput)
     <|> subcommand
             Generate
             "to-directory-tree"
@@ -263,7 +294,7 @@ parseMode =
             Convert
             "decode"
             "Decode a Dhall expression from binary"
-            (Decode <$> parseFile <*> parseJSONFlag)
+            (Decode <$> parseFile <*> parseJSONFlag <*> parseQuiet)
     <|> subcommand
             Miscellaneous
             "repl"
@@ -278,7 +309,7 @@ parseMode =
             Miscellaneous
             "hash"
             "Compute semantic hashes for Dhall expressions"
-            (Hash <$> parseFile)
+            (Hash <$> parseFile <*> parseCache)
     <|> subcommand
             Miscellaneous
             "tags"
@@ -293,7 +324,7 @@ parseMode =
             Debugging
             "haskell-syntax-tree"
             "Output the parsed syntax tree (for debugging)"
-            (SyntaxTree <$> parseFile)
+            (SyntaxTree <$> parseFile <*> parseNoted)
     <|> (   Default
         <$> parseFile
         <*> parseOutput
@@ -303,6 +334,12 @@ parseMode =
         <*> parseVersion
         )
   where
+    deprecatedInPlace =
+        Options.Applicative.switch
+            (   Options.Applicative.long "inplace"
+            <>  Options.Applicative.internal -- completely hidden from help
+            )
+
     argument =
             fmap Data.Text.pack
         .   Options.Applicative.strArgument
@@ -317,6 +354,22 @@ parseMode =
                 (   Options.Applicative.long "file"
                 <>  Options.Applicative.help "Read expression from a file instead of standard input"
                 <>  Options.Applicative.metavar "FILE"
+                <>  Options.Applicative.action "file"
+                )
+
+    parseFiles = fmap f (Options.Applicative.many p)
+      where
+        -- Parse explicit stdin in the input filepaths
+        parseStdin inputs
+            | any (== InputFile "-") inputs = StandardInput : filter (/= InputFile "-") inputs
+            | otherwise = inputs
+
+        f = fromMaybe (pure StandardInput) . nonEmpty . parseStdin . fmap InputFile
+
+        p = Options.Applicative.strArgument
+                (   Options.Applicative.help "Read expression from files instead of standard input"
+                <>  Options.Applicative.metavar "FILES"
+                <>  Options.Applicative.action "file"
                 )
 
     parseOutput = fmap f (optional p)
@@ -328,6 +381,7 @@ parseMode =
                 (   Options.Applicative.long "output"
                 <>  Options.Applicative.help "Write result to a file instead of standard output"
                 <>  Options.Applicative.metavar "FILE"
+                <>  Options.Applicative.action "file"
                 )
 
     parseAlpha =
@@ -380,20 +434,26 @@ parseMode =
     parseQuiet =
         Options.Applicative.switch
             (   Options.Applicative.long "quiet"
-            <>  Options.Applicative.help "Don't print the inferred type"
+            <>  Options.Applicative.help "Don't print the result"
             )
 
-    parseInplace = fmap f (optional p)
-      where
-        f  Nothing    = StandardInput
-        f (Just file) = InputFile file
-
-        p = Options.Applicative.strOption
+    parseInplace =
+        Options.Applicative.strOption
             (   Options.Applicative.long "inplace"
             <>  Options.Applicative.help "Modify the specified file in-place"
             <>  Options.Applicative.metavar "FILE"
+            <>  Options.Applicative.action "file"
             )
 
+    parseTransitiveSwitch = Options.Applicative.flag NonTransitive Transitive
+        (   Options.Applicative.long "transitive"
+        <>  Options.Applicative.help "Modify the input and its transitive relative imports in-place"
+        )
+
+    parseInplaceNonTransitive =
+            fmap InputFile parseInplace
+        <|> pure StandardInput
+
     parseInput = fmap f (optional p)
       where
         f  Nothing    = StandardInput
@@ -403,6 +463,8 @@ parseMode =
             (   Options.Applicative.long "path"
             <>  Options.Applicative.help "Index all files in path recursively. Will get list of files from STDIN if omitted."
             <>  Options.Applicative.metavar "PATH"
+            <>  Options.Applicative.action "file"
+            <>  Options.Applicative.action "directory"
             )
 
     parseTagsOutput = fmap f (optional p)
@@ -414,6 +476,7 @@ parseMode =
             (   Options.Applicative.long "output"
             <>  Options.Applicative.help "The name of the file that the tags are written to. Defaults to \"tags\""
             <>  Options.Applicative.metavar "FILENAME"
+            <>  Options.Applicative.action "file"
             )
 
     parseSuffixes = fmap f (optional p)
@@ -463,11 +526,31 @@ parseMode =
             <>  Options.Applicative.help ("Only check if the input is " <> processed)
             )
 
+    parseSchemasRecord =
+        Options.Applicative.strOption
+            (   Options.Applicative.long "schemas"
+            <>  Options.Applicative.help "A record of schemas"
+            <>  Options.Applicative.metavar "EXPR"
+            )
+
     parseDirectoryTreeOutput =
         Options.Applicative.strOption
             (   Options.Applicative.long "output"
             <>  Options.Applicative.help "The destination path to create"
             <>  Options.Applicative.metavar "PATH"
+            <>  Options.Applicative.action "directory"
+            )
+
+    parseNoted =
+        Options.Applicative.switch
+            (   Options.Applicative.long "noted"
+            <>  Options.Applicative.help "Print `Note` constructors"
+            )
+
+    parseCache =
+        Options.Applicative.switch
+            (   Options.Applicative.long "cache"
+            <>  Options.Applicative.help "Cache the hashed expression"
             )
 
 -- | `ParserInfo` for the `Options` type
@@ -489,10 +572,6 @@ noHeaders i =
 -- | Run the command specified by the `Options` type
 command :: Options -> IO ()
 command (Options {..}) = do
-    let characterSet = case ascii of
-            True  -> ASCII
-            False -> Unicode
-
     GHC.IO.Encoding.setLocaleEncoding System.IO.utf8
 
     let rootDirectory = \case
@@ -501,8 +580,17 @@ command (Options {..}) = do
 
     let toStatus = Dhall.Import.emptyStatus . rootDirectory
 
-    let getExpression          = Dhall.Util.getExpression          censor
-    let getExpressionAndHeader = Dhall.Util.getExpressionAndHeader censor
+    let getExpression = Dhall.Util.getExpression censor
+
+    -- The characterSet detection used here only works on the source
+    -- expression, before any transformation is applied. This helper is there
+    -- make sure the detection is done on the correct expr.
+    let getExpressionAndCharacterSet file = do
+            expr <- getExpression file
+
+            let characterSet = fromMaybe (detectCharacterSet expr) chosenCharacterSet
+
+            return (expr, characterSet)
 
     let handle io =
             Control.Exception.catches io
@@ -544,7 +632,7 @@ command (Options {..}) = do
                         Data.Text.IO.hPutStrLn System.IO.stderr "\ESC[2mUse \"dhall --explain\" for detailed errors\ESC[0m"
                         Control.Exception.throwIO (Imported ps e)
 
-            handleExitCode e = do
+            handleExitCode e =
                 Control.Exception.throwIO (e :: ExitCode)
 
     let renderDoc :: Handle -> Doc Ann -> IO ()
@@ -560,8 +648,8 @@ command (Options {..}) = do
             Pretty.renderIO h ansiStream
             Data.Text.IO.hPutStrLn h ""
 
-    let render :: Pretty a => Handle -> Expr Src a -> IO ()
-        render h expression = do
+    let render :: Pretty a => Handle -> CharacterSet -> Expr Src a -> IO ()
+        render h characterSet expression = do
             let doc = Dhall.Pretty.prettyCharacterSet characterSet expression
 
             renderDoc h doc
@@ -572,10 +660,8 @@ command (Options {..}) = do
 
             AtomicWrite.LazyText.atomicWriteFile file (Pretty.Text.renderLazy stream)
 
-    when (not $ ignoreSemanticCache mode) Dhall.Import.warnAboutMissingCaches
-
     handle $ case mode of
-        Version -> do
+        Version ->
             putStrLn dhallVersionString
 
         Default {..} -> do
@@ -585,7 +671,7 @@ command (Options {..}) = do
                     Exit.exitSuccess
                 else return ()
 
-            expression <- getExpression file
+            (expression, characterSet) <- getExpressionAndCharacterSet file
 
             resolvedExpression <-
                 Dhall.Import.loadRelativeTo (rootDirectory file) semanticCacheMode expression
@@ -605,7 +691,7 @@ command (Options {..}) = do
                         else alphaNormalizedExpression
 
             case output of
-                StandardOutput -> render System.IO.stdout annotatedExpression
+                StandardOutput -> render System.IO.stdout characterSet annotatedExpression
 
                 OutputFile file_ ->
                     writeDocToFile
@@ -664,15 +750,15 @@ command (Options {..}) = do
                  $   _cache
 
         Resolve { resolveMode = Nothing, ..} -> do
-            expression <- getExpression file
+            (expression, characterSet) <- getExpressionAndCharacterSet file
 
             resolvedExpression <-
                 Dhall.Import.loadRelativeTo (rootDirectory file) semanticCacheMode expression
 
-            render System.IO.stdout resolvedExpression
+            render System.IO.stdout characterSet resolvedExpression
 
         Normalize {..} -> do
-            expression <- getExpression file
+            (expression, characterSet) <- getExpressionAndCharacterSet file
 
             resolvedExpression <- Dhall.Import.assertNoImports expression
 
@@ -685,10 +771,10 @@ command (Options {..}) = do
                     then Dhall.Core.alphaNormalize normalizedExpression
                     else normalizedExpression
 
-            render System.IO.stdout alphaNormalizedExpression
+            render System.IO.stdout characterSet alphaNormalizedExpression
 
         Type {..} -> do
-            expression <- getExpression file
+            (expression, characterSet) <- getExpressionAndCharacterSet file
 
             resolvedExpression <-
                 Dhall.Import.loadRelativeTo (rootDirectory file) semanticCacheMode expression
@@ -697,10 +783,12 @@ command (Options {..}) = do
 
             if quiet
                 then return ()
-                else render System.IO.stdout inferredType
+                else render System.IO.stdout characterSet inferredType
 
-        Repl -> do
-            Dhall.Repl.repl characterSet explain
+        Repl ->
+            Dhall.Repl.repl
+                (fromMaybe Unicode chosenCharacterSet) -- Default to Unicode if no characterSet specified
+                explain
 
         Diff {..} -> do
             expression1 <- Dhall.inputExpr expr1
@@ -716,14 +804,20 @@ command (Options {..}) = do
                 else Exit.exitFailure
 
         Format {..} -> do
-            Dhall.Format.format (Dhall.Format.Format {..})
+            when deprecatedInPlace $
+                System.IO.hPutStrLn System.IO.stderr "Warning: the flag \"--inplace\" is deprecated"
+
+            Dhall.Format.format Dhall.Format.Format{..}
 
         Freeze {..} -> do
+            when deprecatedInPlace $
+                System.IO.hPutStrLn System.IO.stderr "Warning: the flag \"--inplace\" is deprecated"
+
             let scope = if all_ then AllImports else OnlyRemoteImports
 
             let intent = if cache then Cache else Secure
 
-            Dhall.Freeze.freeze outputMode input scope intent characterSet censor
+            Dhall.Freeze.freeze outputMode transitivity inputs scope intent chosenCharacterSet censor
 
         Hash {..} -> do
             expression <- getExpression file
@@ -736,51 +830,78 @@ command (Options {..}) = do
             let normalizedExpression =
                     Dhall.Core.alphaNormalize (Dhall.Core.normalize resolvedExpression)
 
+            if cache
+                then Dhall.Import.writeExpressionToSemanticCache normalizedExpression
+                else return ()
+
             Data.Text.IO.putStrLn (Dhall.Import.hashExpressionToCode normalizedExpression)
 
-        Lint {..} -> do
-            case outputMode of
-                Write -> do
-                    (Header header, expression) <- do
-                        getExpressionAndHeader input
+        Lint { transitivity = transitivity0, ..} -> do
+            when deprecatedInPlace $
+                System.IO.hPutStrLn System.IO.stderr "Warning: the flag \"--inplace\" is deprecated"
 
-                    let lintedExpression = Dhall.Lint.lint expression
+            handleMultipleChecksFailed "lint" "linted" go inputs
+          where
+            go input = do
+                let directory = case input of
+                        StandardInput  -> "."
+                        InputFile file -> System.FilePath.takeDirectory file
 
-                    let doc =   Pretty.pretty header
-                            <>  Dhall.Pretty.prettyCharacterSet characterSet lintedExpression
+                let status = Dhall.Import.emptyStatus directory
 
-                    case input of
-                        InputFile file -> writeDocToFile file doc
+                (inputName, originalText, transitivity) <- case input of
+                    InputFile file -> do
+                        text <- Data.Text.IO.readFile file
 
-                        StandardInput -> renderDoc System.IO.stdout doc
+                        return (file, text, transitivity0)
+                    StandardInput -> do
+                        text <- Data.Text.IO.getContents
 
-                Check -> do
-                    originalText <- case input of
-                        InputFile file -> Data.Text.IO.readFile file
-                        StandardInput  -> Data.Text.IO.getContents
+                        return ("(input)", text, NonTransitive)
 
-                    let name = case input of
-                            InputFile file -> file
-                            StandardInput  -> "(input)"
+                (Header header, parsedExpression) <-
+                    Dhall.Util.getExpressionAndHeaderFromStdinText censor inputName originalText
 
-                    (Header header, expression) <- do
-                        Dhall.Core.throws (first Parser.censor (Parser.exprAndHeaderFromText name originalText))
+                let characterSet = fromMaybe (detectCharacterSet parsedExpression) chosenCharacterSet
 
-                    let lintedExpression = Dhall.Lint.lint expression
+                case transitivity of
+                    Transitive ->
+                        for_ parsedExpression $ \import_ -> do
+                            maybeFilepath <- Dhall.Import.dependencyToFile status import_
 
-                    let doc =   Pretty.pretty header
-                            <>  Dhall.Pretty.prettyCharacterSet characterSet lintedExpression
+                            for_ maybeFilepath $ \filepath ->
+                                go (InputFile filepath)
 
-                    let stream = Dhall.Pretty.layout doc
+                    NonTransitive ->
+                        return ()
 
-                    let modifiedText = Pretty.Text.renderStrict stream <> "\n"
+                let lintedExpression = Dhall.Lint.lint parsedExpression
 
-                    if originalText == modifiedText
-                        then return ()
-                        else do
-                            let modified = "linted"
+                let doc =   Pretty.pretty header
+                        <>  Dhall.Pretty.prettyCharacterSet characterSet lintedExpression
 
-                            Control.Exception.throwIO CheckFailed{ command = "lint", ..}
+                let stream = Dhall.Pretty.layout doc
+
+                let modifiedText = Pretty.Text.renderStrict stream <> "\n"
+
+                case outputMode of
+                    Write -> do
+                        case input of
+                            InputFile file ->
+                                if originalText == modifiedText
+                                    then return ()
+                                    else writeDocToFile file doc
+
+                            StandardInput ->
+                                renderDoc System.IO.stdout doc
+
+                        return (Right ())
+
+                    Check ->
+                        return $
+                            if originalText == modifiedText
+                                then Right ()
+                                else Left CheckFailed{..}
 
         Encode {..} -> do
             expression <- getExpression file
@@ -797,16 +918,16 @@ command (Options {..}) = do
 
                     Data.ByteString.Lazy.Char8.putStrLn jsonBytes
 
-                else do
+                else
                     Data.ByteString.Lazy.putStr bytes
 
         Decode {..} -> do
-            bytes <- do
+            bytes <-
                 case file of
                     InputFile f   -> Data.ByteString.Lazy.readFile f
                     StandardInput -> Data.ByteString.Lazy.getContents
 
-            expression <- do
+            expression <-
                 if json
                     then do
                         value <- case Data.Aeson.eitherDecode' bytes of
@@ -818,13 +939,19 @@ command (Options {..}) = do
                         let cborgBytes = Codec.CBOR.Write.toLazyByteString encoding
 
                         Dhall.Core.throws (Dhall.Binary.decodeExpression cborgBytes)
-                    else do
+                    else
                         Dhall.Core.throws (Dhall.Binary.decodeExpression bytes)
 
 
-            let doc = Dhall.Pretty.prettyCharacterSet characterSet (Dhall.Core.renote expression :: Expr Src Import)
+            if quiet
+                then return ()
+                else do
+                    let doc =
+                            Dhall.Pretty.prettyCharacterSet
+                                (fromMaybe Unicode chosenCharacterSet) -- default to Unicode
+                                (Dhall.Core.renote expression :: Expr Src Import)
 
-            renderDoc System.IO.stdout doc
+                    renderDoc System.IO.stdout doc
 
         Text {..} -> do
             expression <- getExpression file
@@ -837,8 +964,11 @@ command (Options {..}) = do
             let normalizedExpression = Dhall.Core.normalize resolvedExpression
 
             case normalizedExpression of
-                Dhall.Core.TextLit (Dhall.Core.Chunks [] text) -> do
-                    Data.Text.IO.putStr text
+                Dhall.Core.TextLit (Dhall.Core.Chunks [] text) ->
+                    let write = case output of
+                          StandardOutput -> Data.Text.IO.putStr
+                          OutputFile file_ -> Data.Text.IO.writeFile file_
+                    in write text
                 _ -> do
                     let invalidDecoderExpected :: Expr Void Void
                         invalidDecoderExpected = Dhall.Core.Text
@@ -869,13 +999,18 @@ command (Options {..}) = do
 
             DirectoryTree.toDirectoryTree path normalizedExpression
 
+        Dhall.Main.Schemas{..} ->
+            Dhall.Schemas.schemasCommand Dhall.Schemas.Schemas{ input = file, ..}
+
         SyntaxTree {..} -> do
             expression <- getExpression file
 
-            let denoted :: Expr Void Import
-                denoted = Dhall.Core.denote expression
-
-            Text.Pretty.Simple.pPrintNoColor denoted
+            if noted then
+                Text.Pretty.Simple.pPrintNoColor expression
+            else
+                let denoted :: Expr Void Import
+                    denoted = Dhall.Core.denote expression
+                in Text.Pretty.Simple.pPrintNoColor denoted
 
 -- | Entry point for the @dhall@ executable
 main :: IO ()
