@@ -1,6 +1,8 @@
-{-# LANGUAGE BangPatterns       #-}
-{-# LANGUAGE OverloadedStrings  #-}
-{-# LANGUAGE RankNTypes         #-}
+{-# LANGUAGE BangPatterns      #-}
+{-# LANGUAGE LambdaCase        #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RankNTypes        #-}
+{-# LANGUAGE ViewPatterns      #-}
 
 module Dhall.Normalize (
       alphaNormalize
@@ -12,37 +14,38 @@ module Dhall.Normalize (
     , ReifiedNormalizer (..)
     , judgmentallyEqual
     , subst
-    , shift
+    , Syntax.shift
     , isNormalized
     , isNormalizedWith
     , freeIn
     ) where
 
-import Control.Applicative (empty)
+import Control.Applicative   (empty)
 import Data.Foldable
-import Data.Functor.Identity (Identity(..))
-import Data.Semigroup (Semigroup(..))
-import Data.Sequence (ViewL(..), ViewR(..))
+import Data.Functor.Identity (Identity (..))
+import Data.List.NonEmpty    (NonEmpty (..))
+import Data.Sequence         (ViewL (..), ViewR (..))
 import Data.Traversable
-import Instances.TH.Lift ()
-import Prelude hiding (succ)
+import Instances.TH.Lift     ()
+import Prelude               hiding (succ)
 
 import Dhall.Syntax
-    ( Expr(..)
-    , Var(..)
-    , Binding(Binding)
-    , Chunks(..)
-    , DhallDouble(..)
-    , Const(..)
-    , PreferAnnotation(..)
+    ( Binding (Binding)
+    , Chunks (..)
+    , DhallDouble (..)
+    , Expr (..)
+    , FieldSelection (..)
+    , FunctionBinding (..)
+    , PreferAnnotation (..)
+    , RecordField (..)
+    , Var (..)
     )
 
 import qualified Data.Sequence
 import qualified Data.Set
-import qualified Data.Text
+import qualified Data.Text     as Text
 import qualified Dhall.Eval    as Eval
 import qualified Dhall.Map
-import qualified Dhall.Set
 import qualified Dhall.Syntax  as Syntax
 import qualified Lens.Family   as Lens
 
@@ -56,115 +59,28 @@ judgmentallyEqual :: Eq a => Expr s a -> Expr t a -> Bool
 judgmentallyEqual = Eval.judgmentallyEqual
 {-# INLINE judgmentallyEqual #-}
 
-{-| `shift` is used by both normalization and type-checking to avoid variable
-    capture by shifting variable indices
-
-    For example, suppose that you were to normalize the following expression:
-
-> λ(a : Type) → λ(x : a) → (λ(y : a) → λ(x : a) → y) x
-
-    If you were to substitute @y@ with @x@ without shifting any variable
-    indices, then you would get the following incorrect result:
-
-> λ(a : Type) → λ(x : a) → λ(x : a) → x  -- Incorrect normalized form
-
-    In order to substitute @x@ in place of @y@ we need to `shift` @x@ by @1@ in
-    order to avoid being misinterpreted as the @x@ bound by the innermost
-    lambda.  If we perform that `shift` then we get the correct result:
-
-> λ(a : Type) → λ(x : a) → λ(x : a) → x@1
-
-    As a more worked example, suppose that you were to normalize the following
-    expression:
-
->     λ(a : Type)
-> →   λ(f : a → a → a)
-> →   λ(x : a)
-> →   λ(x : a)
-> →   (λ(x : a) → f x x@1) x@1
-
-    The correct normalized result would be:
-
->     λ(a : Type)
-> →   λ(f : a → a → a)
-> →   λ(x : a)
-> →   λ(x : a)
-> →   f x@1 x
-
-    The above example illustrates how we need to both increase and decrease
-    variable indices as part of substitution:
-
-    * We need to increase the index of the outer @x\@1@ to @x\@2@ before we
-      substitute it into the body of the innermost lambda expression in order
-      to avoid variable capture.  This substitution changes the body of the
-      lambda expression to @(f x\@2 x\@1)@
-
-    * We then remove the innermost lambda and therefore decrease the indices of
-      both @x@s in @(f x\@2 x\@1)@ to @(f x\@1 x)@ in order to reflect that one
-      less @x@ variable is now bound within that scope
-
-    Formally, @(shift d (V x n) e)@ modifies the expression @e@ by adding @d@ to
-    the indices of all variables named @x@ whose indices are greater than
-    @(n + m)@, where @m@ is the number of bound variables of the same name
-    within that scope
-
-    In practice, @d@ is always @1@ or @-1@ because we either:
-
-    * increment variables by @1@ to avoid variable capture during substitution
-    * decrement variables by @1@ when deleting lambdas after substitution
-
-    @n@ starts off at @0@ when substitution begins and increments every time we
-    descend into a lambda or let expression that binds a variable of the same
-    name in order to avoid shifting the bound variables by mistake.
--}
-shift :: Int -> Var -> Expr s a -> Expr s a
-shift d (V x n) (Var (V x' n')) = Var (V x' n'')
-  where
-    n'' = if x == x' && n <= n' then n' + d else n'
-shift d (V x n) (Lam x' _A b) = Lam x' _A' b'
-  where
-    _A' = shift d (V x n ) _A
-    b'  = shift d (V x n') b
-      where
-        n' = if x == x' then n + 1 else n
-shift d (V x n) (Pi x' _A _B) = Pi x' _A' _B'
-  where
-    _A' = shift d (V x n ) _A
-    _B' = shift d (V x n') _B
-      where
-        n' = if x == x' then n + 1 else n
-shift d (V x n) (Let (Binding src0 f src1 mt src2 r) e) =
-    Let (Binding src0 f src1 mt' src2 r') e'
-  where
-    e' = shift d (V x n') e
-      where
-        n' = if x == f then n + 1 else n
-
-    mt' = fmap (fmap (shift d (V x n))) mt
-    r'  =             shift d (V x n)  r
-shift d v expression = Lens.over Syntax.subExpressions (shift d v) expression
-
 {-| Substitute all occurrences of a variable with an expression
 
 > subst x C B  ~  B[x := C]
 -}
 subst :: Var -> Expr s a -> Expr s a -> Expr s a
 subst _ _ (Const a) = Const a
-subst (V x n) e (Lam y _A b) = Lam y _A' b'
+subst (V x n) e (Lam cs (FunctionBinding src0 y src1 src2 _A) b) =
+    Lam cs (FunctionBinding src0 y src1 src2 _A') b'
   where
-    _A' = subst (V x n )                  e  _A
-    b'  = subst (V x n') (shift 1 (V y 0) e)  b
+    _A' = subst (V x n )                         e  _A
+    b'  = subst (V x n') (Syntax.shift 1 (V y 0) e)  b
     n'  = if x == y then n + 1 else n
-subst (V x n) e (Pi y _A _B) = Pi y _A' _B'
+subst (V x n) e (Pi cs y _A _B) = Pi cs y _A' _B'
   where
-    _A' = subst (V x n )                  e  _A
-    _B' = subst (V x n') (shift 1 (V y 0) e) _B
+    _A' = subst (V x n )                         e  _A
+    _B' = subst (V x n') (Syntax.shift 1 (V y 0) e) _B
     n'  = if x == y then n + 1 else n
 subst v e (Var v') = if v == v' then e else Var v'
 subst (V x n) e (Let (Binding src0 f src1 mt src2 r) b) =
     Let (Binding src0 f src1 mt' src2 r') b'
   where
-    b' = subst (V x n') (shift 1 (V f 0) e) b
+    b' = subst (V x n') (Syntax.shift 1 (V f 0) e) b
       where
         n' = if x == f then n + 1 else n
 
@@ -190,15 +106,16 @@ boundedType Double           = True
 boundedType Text             = True
 boundedType (App List _)     = False
 boundedType (App Optional t) = boundedType t
-boundedType (Record kvs)     = all boundedType kvs
+boundedType (Record kvs)     = all (boundedType . recordFieldValue) kvs
 boundedType (Union kvs)      = all (all boundedType) kvs
 boundedType _                = False
 
 {-| α-normalize an expression by renaming all bound variables to @\"_\"@ and
     using De Bruijn indices to distinguish them
 
->>> alphaNormalize (Lam "a" (Const Type) (Lam "b" (Const Type) (Lam "x" "a" (Lam "y" "b" "x"))))
-Lam "_" (Const Type) (Lam "_" (Const Type) (Lam "_" (Var (V "_" 1)) (Lam "_" (Var (V "_" 1)) (Var (V "_" 1)))))
+>>> mfb = Syntax.makeFunctionBinding
+>>> alphaNormalize (Lam mempty (mfb "a" (Const Type)) (Lam mempty (mfb "b" (Const Type)) (Lam mempty (mfb "x" "a") (Lam mempty (mfb "y" "b") "x"))))
+Lam Nothing (FunctionBinding {functionBindingSrc0 = Nothing, functionBindingVariable = "_", functionBindingSrc1 = Nothing, functionBindingSrc2 = Nothing, functionBindingAnnotation = Const Type}) (Lam Nothing (FunctionBinding {functionBindingSrc0 = Nothing, functionBindingVariable = "_", functionBindingSrc1 = Nothing, functionBindingSrc2 = Nothing, functionBindingAnnotation = Const Type}) (Lam Nothing (FunctionBinding {functionBindingSrc0 = Nothing, functionBindingVariable = "_", functionBindingSrc1 = Nothing, functionBindingSrc2 = Nothing, functionBindingAnnotation = Var (V "_" 1)}) (Lam Nothing (FunctionBinding {functionBindingSrc0 = Nothing, functionBindingVariable = "_", functionBindingSrc1 = Nothing, functionBindingSrc2 = Nothing, functionBindingAnnotation = Var (V "_" 1)}) (Var (V "_" 1)))))
 
     α-normalization does not affect free variables:
 
@@ -226,7 +143,7 @@ normalize = Eval.normalize
 {-| Reduce an expression to its normal form, performing beta reduction and applying
     any custom definitions.
 
-    `normalizeWith` is designed to be used with function `typeWith`. The `typeWith`
+    `normalizeWith` is designed to be used with function `Dhall.TypeCheck.typeWith`. The `Dhall.TypeCheck.typeWith`
     function allows typing of Dhall functions in a custom typing context whereas
     `normalizeWith` allows evaluating Dhall expressions in a custom context.
 
@@ -243,6 +160,7 @@ normalize = Eval.normalize
 normalizeWith :: Eq a => Maybe (ReifiedNormalizer a) -> Expr s a -> Expr t a
 normalizeWith (Just ctx) t = runIdentity (normalizeWithM (getReifiedNormalizer ctx) t)
 normalizeWith _          t = Eval.normalize t
+{-# INLINABLE normalizeWith #-}
 
 {-| This function generalizes `normalizeWith` by allowing the custom normalizer
     to use an arbitrary `Monad`
@@ -254,14 +172,15 @@ normalizeWithM
     :: (Monad m, Eq a) => NormalizerM m a -> Expr s a -> m (Expr t a)
 normalizeWithM ctx e0 = loop (Syntax.denote e0)
  where
- loop e =  case e of
+ loop =  \case
     Const k -> pure (Const k)
     Var v -> pure (Var v)
-    Lam x _A b -> Lam x <$> _A' <*> b'
+    Lam cs (FunctionBinding { functionBindingVariable = x, functionBindingAnnotation = _A }) b ->
+        Lam cs <$> (Syntax.makeFunctionBinding x <$> _A') <*> b'
       where
         _A' = loop _A
         b'  = loop b
-    Pi x _A _B -> Pi x <$> _A' <*> _B'
+    Pi cs x _A _B -> Pi cs x <$> _A' <*> _B'
       where
         _A' = loop _A
         _B' = loop _B
@@ -273,14 +192,14 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
               f' <- loop f
               a' <- loop a
               case f' of
-                Lam x _A b₀ -> do
+                Lam _ (FunctionBinding _ x _ _ _A) b₀ -> do
 
-                    let a₂ = shift 1 (V x 0) a'
+                    let a₂ = Syntax.shift 1 (V x 0) a'
                     let b₁ = subst (V x 0) a₂ b₀
-                    let b₂ = shift (-1) (V x 0) b₁
+                    let b₂ = Syntax.shift (-1) (V x 0) b₁
 
                     loop b₂
-                _ -> do
+                _ ->
                   case App f' a' of
                     App (App (App (App NaturalFold (NaturalLit n0)) t) succ') zero -> do
                       t' <- loop t
@@ -293,14 +212,14 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                         strict =       strictLoop (fromIntegral n0 :: Integer)
                         lazy   = loop (  lazyLoop (fromIntegral n0 :: Integer))
 
-                        strictLoop !0 = loop zero
+                        strictLoop 0 = loop zero
                         strictLoop !n = App succ' <$> strictLoop (n - 1) >>= loop
 
-                        lazyLoop !0 = zero
+                        lazyLoop 0 = zero
                         lazyLoop !n = App succ' (lazyLoop (n - 1))
                     App NaturalBuild g -> loop (App (App (App g Natural) succ) zero)
                       where
-                        succ = Lam "n" Natural (NaturalPlus "n" (NaturalLit 1))
+                        succ = Lam mempty (Syntax.makeFunctionBinding "n" Natural) (NaturalPlus "n" (NaturalLit 1))
 
                         zero = NaturalLit 0
                     App NaturalIsZero (NaturalLit n) -> pure (BoolLit (n == 0))
@@ -308,7 +227,7 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                     App NaturalOdd (NaturalLit n) -> pure (BoolLit (odd n))
                     App NaturalToInteger (NaturalLit n) -> pure (IntegerLit (toInteger n))
                     App NaturalShow (NaturalLit n) ->
-                        pure (TextLit (Chunks [] (Data.Text.pack (show n))))
+                        pure (TextLit (Chunks [] (Text.pack (show n))))
                     App (App NaturalSubtract (NaturalLit x)) (NaturalLit y)
                         -- Use an `Integer` for the subtraction, due to the
                         -- following issue:
@@ -327,32 +246,24 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                     App IntegerNegate (IntegerLit n) ->
                         pure (IntegerLit (negate n))
                     App IntegerShow (IntegerLit n)
-                        | 0 <= n    -> pure (TextLit (Chunks [] ("+" <> Data.Text.pack (show n))))
-                        | otherwise -> pure (TextLit (Chunks [] (Data.Text.pack (show n))))
+                        | 0 <= n    -> pure (TextLit (Chunks [] ("+" <> Text.pack (show n))))
+                        | otherwise -> pure (TextLit (Chunks [] (Text.pack (show n))))
                     -- `(read . show)` is used instead of `fromInteger` because `read` uses
                     -- the correct rounding rule.
                     -- See https://gitlab.haskell.org/ghc/ghc/issues/17231.
                     App IntegerToDouble (IntegerLit n) -> pure (DoubleLit ((DhallDouble . read . show) n))
                     App DoubleShow (DoubleLit (DhallDouble n)) ->
-                        pure (TextLit (Chunks [] (Data.Text.pack (show n))))
-                    App (App OptionalBuild _A₀) g ->
-                        loop (App (App (App g optional) just) nothing)
-                      where
-                        optional = App Optional _A₀
-
-                        just = Lam "a" _A₀ (Some "a")
-
-                        nothing = App None _A₀
+                        pure (TextLit (Chunks [] (Text.pack (show n))))
                     App (App ListBuild _A₀) g -> loop (App (App (App g list) cons) nil)
                       where
-                        _A₁ = shift 1 "a" _A₀
+                        _A₁ = Syntax.shift 1 "a" _A₀
 
                         list = App List _A₀
 
                         cons =
-                            Lam "a" _A₀
-                                (Lam "as"
-                                    (App List _A₁)
+                            Lam mempty (Syntax.makeFunctionBinding "a" _A₀)
+                                (Lam mempty
+                                    (Syntax.makeFunctionBinding "as" (App List _A₁))
                                     (ListAppend (ListLit Nothing (pure "a")) "as")
                                 )
 
@@ -367,7 +278,7 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                         strictNil = loop nil
                         lazyNil   =      nil
 
-                        strictCons y ys = do
+                        strictCons y ys =
                           App (App cons y) <$> ys >>= loop
                         lazyCons   y ys =       App (App cons y) ys
                     App (App ListLength _) (ListLit _ ys) ->
@@ -388,8 +299,8 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
 
                         _A₂ = Record (Dhall.Map.fromList kts)
                           where
-                            kts = [ ("index", Natural)
-                                  , ("value", _A₀)
+                            kts = [ ("index", Syntax.makeRecordField Natural)
+                                  , ("value", Syntax.makeRecordField _A₀)
                                   ]
 
                         t | null as₀  = Just (App List _A₂)
@@ -398,30 +309,73 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                         adapt n a_ =
                             RecordLit (Dhall.Map.fromList kvs)
                           where
-                            kvs = [ ("index", NaturalLit (fromIntegral n))
-                                  , ("value", a_)
+                            kvs = [ ("index", Syntax.makeRecordField $ NaturalLit (fromIntegral n))
+                                  , ("value", Syntax.makeRecordField a_)
                                   ]
                     App (App ListReverse _) (ListLit t xs) ->
                         loop (ListLit t (Data.Sequence.reverse xs))
-
-                    App (App OptionalFold t0) x0 -> do
-                        t1 <- loop t0
-                        let optional = Var (V "optional" 0)
-                        let lam term = (Lam "optional"
-                                           (Const Type)
-                                           (Lam "some"
-                                               (Pi "_" t1 optional)
-                                               (Lam "none" optional term)))
-                        x1 <- loop x0
-                        pure $ case x1 of
-                            App None _ -> lam (Var (V "none" 0))
-                            Some x'    -> lam (App (Var (V "some" 0)) x')
-                            _          -> App (App OptionalFold t1) x1
-
                     App TextShow (TextLit (Chunks [] oldText)) ->
                         loop (TextLit (Chunks [] newText))
                       where
                         newText = Eval.textShow oldText
+                    App
+                        (App (App TextReplace (TextLit (Chunks [] ""))) _)
+                        haystack ->
+                            return haystack
+                    App (App
+                            (App TextReplace (TextLit (Chunks [] needleText)))
+                            (TextLit (Chunks [] replacementText))
+                        )
+                        (TextLit (Chunks xys z)) -> do
+                            let xys' = do
+                                    (x, y) <- xys
+
+                                    let x' = Text.replace needleText replacementText x
+                                    return (x', y)
+
+                            let z' = Text.replace needleText replacementText z
+
+                            return (TextLit (Chunks xys' z'))
+                    App (App
+                            (App TextReplace (TextLit (Chunks [] needleText)))
+                            replacement
+                        )
+                        (TextLit (Chunks [] lastText)) -> do
+                            let (prefix, suffix) =
+                                    Text.breakOn needleText lastText
+
+                            if Text.null suffix
+                                then return (TextLit (Chunks [] lastText))
+                                else do
+                                    let remainder =
+                                            Text.drop
+                                                (Text.length needleText)
+                                                suffix
+
+                                    loop (TextAppend (TextLit (Chunks [(prefix, replacement)] "")) (App (App (App TextReplace (TextLit (Chunks [] needleText))) replacement) (TextLit (Chunks [] remainder))))
+                    App (App
+                            (App TextReplace (TextLit (Chunks [] needleText)))
+                            replacement
+                        )
+                        (TextLit
+                            (Chunks
+                                ((firstText, firstInterpolation) : chunks)
+                                lastText
+                            )
+                        ) -> do
+                            let (prefix, suffix) =
+                                    Text.breakOn needleText firstText
+
+                            if Text.null suffix
+                                then do
+                                    loop (TextAppend (TextLit (Chunks [(firstText, firstInterpolation)] "")) (App (App (App TextReplace (TextLit (Chunks [] needleText))) replacement) (TextLit (Chunks chunks lastText))))
+                                else do
+                                    let remainder =
+                                            Text.drop
+                                                (Text.length needleText)
+                                                suffix
+
+                                    loop (TextAppend (TextLit (Chunks [(prefix, replacement)] "")) (App (App (App TextReplace (TextLit (Chunks [] needleText))) replacement) (TextLit (Chunks ((remainder, firstInterpolation) : chunks) lastText))))
                     _ -> do
                         res2 <- ctx (App f' a')
                         case res2 of
@@ -429,9 +383,9 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                             Just app' -> loop app'
     Let (Binding _ f _ _ _ r) b -> loop b''
       where
-        r'  = shift   1  (V f 0) r
+        r'  = Syntax.shift   1  (V f 0) r
         b'  = subst (V f 0) r' b
-        b'' = shift (-1) (V f 0) b'
+        b'' = Syntax.shift (-1) (V f 0) b'
     Annot x _ -> loop x
     Bool -> pure Bool
     BoolLit b -> pure (BoolLit b)
@@ -524,7 +478,14 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
             TextLit c -> pure [Chunks [] x, c]
             _         -> pure [Chunks [(x, y')] mempty]
     TextAppend x y -> loop (TextLit (Chunks [("", x), ("", y)] ""))
+    TextReplace -> pure TextReplace
     TextShow -> pure TextShow
+    Date -> pure Date
+    DateLiteral d -> pure (DateLiteral d)
+    Time -> pure Time
+    TimeLiteral t p -> pure (TimeLiteral t p)
+    TimeZone -> pure TimeZone
+    TimeZoneLiteral z -> pure (TimeZoneLiteral z)
     List -> pure List
     ListLit t es
         | Data.Sequence.null es -> ListLit <$> t' <*> pure Data.Sequence.empty
@@ -550,38 +511,44 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
       where
         a' = loop a
     None -> pure None
-    OptionalFold -> pure OptionalFold
-    OptionalBuild -> pure OptionalBuild
     Record kts -> Record . Dhall.Map.sort <$> kts'
       where
-        kts' = traverse loop kts
+        f (RecordField s0 expr s1 s2) = (\e -> RecordField s0 e s1 s2) <$> loop expr
+        kts' = traverse f kts
     RecordLit kvs -> RecordLit . Dhall.Map.sort <$> kvs'
       where
-        kvs' = traverse loop kvs
+        f (RecordField s0 expr s1 s2) = (\e -> RecordField s0 e s1 s2) <$> loop expr
+        kvs' = traverse f kvs
     Union kts -> Union . Dhall.Map.sort <$> kts'
       where
         kts' = traverse (traverse loop) kts
-    Combine mk x y -> decide <$> loop x <*> loop y
+    Combine cs mk x y -> decide <$> loop x <*> loop y
       where
         decide (RecordLit m) r | Data.Foldable.null m =
             r
         decide l (RecordLit n) | Data.Foldable.null n =
             l
         decide (RecordLit m) (RecordLit n) =
-            RecordLit (Dhall.Map.unionWith decide m n)
+            RecordLit (Dhall.Map.unionWith f m n)
+          where
+            f (RecordField _ expr _ _) (RecordField _ expr' _ _) =
+              Syntax.makeRecordField $ decide expr expr'
         decide l r =
-            Combine mk l r
-    CombineTypes x y -> decide <$> loop x <*> loop y
+            Combine cs mk l r
+    CombineTypes cs x y -> decide <$> loop x <*> loop y
       where
         decide (Record m) r | Data.Foldable.null m =
             r
         decide l (Record n) | Data.Foldable.null n =
             l
         decide (Record m) (Record n) =
-            Record (Dhall.Map.unionWith decide m n)
+            Record (Dhall.Map.unionWith f m n)
+          where
+            f (RecordField _ expr _ _) (RecordField _ expr' _ _) =
+              Syntax.makeRecordField $ decide expr expr'
         decide l r =
-            CombineTypes l r
-    Prefer _ x y -> decide <$> loop x <*> loop y
+            CombineTypes cs l r
+    Prefer cs _ x y -> decide <$> loop x <*> loop y
       where
         decide (RecordLit m) r | Data.Foldable.null m =
             r
@@ -592,37 +559,40 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
         decide l r | Eval.judgmentallyEqual l r =
             l
         decide l r =
-            Prefer PreferFromSource l r
-    RecordCompletion x y -> do
-        loop (Annot (Prefer PreferFromCompletion (Field x "default") y) (Field x "Type"))
+            Prefer cs PreferFromSource l r
+    RecordCompletion x y ->
+        loop (Annot (Prefer mempty PreferFromCompletion (Field x def) y) (Field x typ))
+      where
+        def = Syntax.makeFieldSelection "default"
+        typ = Syntax.makeFieldSelection "Type"
     Merge x y t      -> do
         x' <- loop x
         y' <- loop y
         case x' of
             RecordLit kvsX ->
                 case y' of
-                    Field (Union ktsY) kY ->
+                    Field (Union ktsY) (Syntax.fieldSelectionLabel -> kY) ->
                         case Dhall.Map.lookup kY ktsY of
                             Just Nothing ->
-                                case Dhall.Map.lookup kY kvsX of
+                                case recordFieldValue <$> Dhall.Map.lookup kY kvsX of
                                     Just vX -> return vX
                                     Nothing -> Merge x' y' <$> t'
                             _ ->
                                 Merge x' y' <$> t'
-                    App (Field (Union ktsY) kY) vY ->
+                    App (Field (Union ktsY) (Syntax.fieldSelectionLabel -> kY)) vY ->
                         case Dhall.Map.lookup kY ktsY of
                             Just (Just _) ->
-                                case Dhall.Map.lookup kY kvsX of
+                                case recordFieldValue <$> Dhall.Map.lookup kY kvsX of
                                     Just vX -> loop (App vX vY)
                                     Nothing -> Merge x' y' <$> t'
                             _ ->
                                 Merge x' y' <$> t'
                     Some a ->
-                        case Dhall.Map.lookup "Some" kvsX of
+                        case recordFieldValue <$> Dhall.Map.lookup "Some" kvsX of
                             Just vX -> loop (App vX a)
                             Nothing -> Merge x' y' <$> t'
                     App None _ ->
-                        case Dhall.Map.lookup "None" kvsX of
+                        case recordFieldValue <$> Dhall.Map.lookup "None" kvsX of
                             Just vX -> return vX
                             Nothing -> Merge x' y' <$> t'
                     _ -> Merge x' y' <$> t'
@@ -637,12 +607,12 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                 let entry (key, value) =
                         RecordLit
                             (Dhall.Map.fromList
-                                [ ("mapKey"  , TextLit (Chunks [] key))
-                                , ("mapValue", value                  )
+                                [ ("mapKey"  , Syntax.makeRecordField $ TextLit (Chunks [] key))
+                                , ("mapValue", Syntax.makeRecordField value                  )
                                 ]
                             )
 
-                let keyValues = Data.Sequence.fromList (map entry (Dhall.Map.toList kvsX))
+                let keyValues = Data.Sequence.fromList (map entry (Dhall.Map.toList $ recordFieldValue <$> kvsX))
 
                 let listType = case t' of
                         Just _ | null keyValues ->
@@ -651,52 +621,52 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
                             Nothing
 
                 return (ListLit listType keyValues)
-            _ -> do
+            _ ->
                 return (ToMap x' t')
-    Field r x        -> do
+    Field r k@FieldSelection{fieldSelectionLabel = x}        -> do
         let singletonRecordLit v = RecordLit (Dhall.Map.singleton x v)
 
         r' <- loop r
         case r' of
             RecordLit kvs ->
                 case Dhall.Map.lookup x kvs of
-                    Just v  -> pure v
-                    Nothing -> Field <$> (RecordLit <$> traverse loop kvs) <*> pure x
-            Project r_ _ -> loop (Field r_ x)
-            Prefer _ (RecordLit kvs) r_ -> case Dhall.Map.lookup x kvs of
-                Just v -> pure (Field (Prefer PreferFromSource (singletonRecordLit v) r_) x)
-                Nothing -> loop (Field r_ x)
-            Prefer _ l (RecordLit kvs) -> case Dhall.Map.lookup x kvs of
-                Just v -> pure v
-                Nothing -> loop (Field l x)
-            Combine m (RecordLit kvs) r_ -> case Dhall.Map.lookup x kvs of
-                Just v -> pure (Field (Combine m (singletonRecordLit v) r_) x)
-                Nothing -> loop (Field r_ x)
-            Combine m l (RecordLit kvs) -> case Dhall.Map.lookup x kvs of
-                Just v -> pure (Field (Combine m l (singletonRecordLit v)) x)
-                Nothing -> loop (Field l x)
-            _ -> pure (Field r' x)
+                    Just v  -> pure $ recordFieldValue v
+                    Nothing -> Field <$> (RecordLit <$> traverse (Syntax.recordFieldExprs loop) kvs) <*> pure k
+            Project r_ _ -> loop (Field r_ k)
+            Prefer cs _ (RecordLit kvs) r_ -> case Dhall.Map.lookup x kvs of
+                Just v -> pure (Field (Prefer cs PreferFromSource (singletonRecordLit v) r_) k)
+                Nothing -> loop (Field r_ k)
+            Prefer _ _ l (RecordLit kvs) -> case Dhall.Map.lookup x kvs of
+                Just v -> pure $ recordFieldValue v
+                Nothing -> loop (Field l k)
+            Combine cs m (RecordLit kvs) r_ -> case Dhall.Map.lookup x kvs of
+                Just v -> pure (Field (Combine cs m (singletonRecordLit v) r_) k)
+                Nothing -> loop (Field r_ k)
+            Combine cs m l (RecordLit kvs) -> case Dhall.Map.lookup x kvs of
+                Just v -> pure (Field (Combine cs m l (singletonRecordLit v)) k)
+                Nothing -> loop (Field l k)
+            _ -> pure (Field r' k)
     Project x (Left fields)-> do
         x' <- loop x
-        let fieldsSet = Dhall.Set.toSet fields
+        let fieldsSet = Data.Set.fromList fields
         case x' of
             RecordLit kvs ->
                 pure (RecordLit (Dhall.Map.restrictKeys kvs fieldsSet))
             Project y _ ->
                 loop (Project y (Left fields))
-            Prefer _ l (RecordLit rKvs) -> do
+            Prefer cs _ l (RecordLit rKvs) -> do
                 let rKs = Dhall.Map.keysSet rKvs
-                let l' = Project l (Left (Dhall.Set.fromSet (Data.Set.difference fieldsSet rKs)))
+                let l' = Project l (Left (Data.Set.toList (Data.Set.difference fieldsSet rKs)))
                 let r' = RecordLit (Dhall.Map.restrictKeys rKvs fieldsSet)
-                loop (Prefer PreferFromSource l' r')
+                loop (Prefer cs PreferFromSource l' r')
             _ | null fields -> pure (RecordLit mempty)
-              | otherwise   -> pure (Project x' (Left (Dhall.Set.sort fields)))
+              | otherwise   -> pure (Project x' (Left (Data.Set.toList (Data.Set.fromList fields))))
     Project r (Right e1) -> do
         e2 <- loop e1
 
         case e2 of
-            Record kts -> do
-                loop (Project r (Left (Dhall.Set.fromSet (Dhall.Map.keysSet kts))))
+            Record kts ->
+                loop (Project r (Left (Data.Set.toList (Dhall.Map.keysSet kts))))
             _ -> do
                 r' <- loop r
                 pure (Project r' (Right e2))
@@ -704,13 +674,31 @@ normalizeWithM ctx e0 = loop (Syntax.denote e0)
         t' <- loop t
 
         pure (Assert t')
-    Equivalent l r -> do
+    Equivalent cs l r -> do
         l' <- loop l
         r' <- loop r
 
-        pure (Equivalent l' r')
-    With e' k v -> do
-        loop (Syntax.desugarWith (With e' k v))
+        pure (Equivalent cs l' r')
+    With e ks v -> do
+        e' <- loop e
+        v' <- loop v
+
+        case e' of
+            RecordLit kvs ->
+                case ks of
+                    k :| [] ->
+                        return (RecordLit (Dhall.Map.insert k (Syntax.makeRecordField v') kvs))
+                    k₀ :| k₁ : ks' -> do
+                        let e₁ =
+                                case Dhall.Map.lookup k₀ kvs of
+                                    Nothing -> RecordLit mempty
+                                    Just r  -> Syntax.recordFieldValue r
+
+                        e₂ <- loop (With e₁ (k₁ :| ks') v')
+
+                        return (RecordLit (Dhall.Map.insert k₀ (Syntax.makeRecordField e₂) kvs))
+            _ ->
+                return (With e' ks v')
     Note _ e' -> loop e'
     ImportAlt l _r -> loop l
     Embed a -> pure (Embed a)
@@ -750,8 +738,9 @@ isNormalized e0 = loop (Syntax.denote e0)
     loop e = case e of
       Const _ -> True
       Var _ -> True
-      Lam _ a b -> loop a && loop b
-      Pi _ a b -> loop a && loop b
+      Lam _ (FunctionBinding Nothing _ Nothing Nothing a) b -> loop a && loop b
+      Lam _ _ _ -> False
+      Pi _ _ a b -> loop a && loop b
       App f a -> loop f && loop a && case App f a of
           App (Lam _ _ _) _ -> False
           App (App (App (App NaturalFold (NaturalLit _)) _) _) _ -> False
@@ -770,7 +759,6 @@ isNormalized e0 = loop (Syntax.denote e0)
           App IntegerShow (IntegerLit _) -> False
           App IntegerToDouble (IntegerLit _) -> False
           App DoubleShow (DoubleLit _) -> False
-          App (App OptionalBuild _) _ -> False
           App (App ListBuild _) _ -> False
           App (App (App (App (App (App ListFold _) (ListLit _ _)) _) _) _) _ -> False
           App (App ListLength _) (ListLit _ _) -> False
@@ -778,10 +766,12 @@ isNormalized e0 = loop (Syntax.denote e0)
           App (App ListLast _) (ListLit _ _) -> False
           App (App ListIndexed _) (ListLit _ _) -> False
           App (App ListReverse _) (ListLit _ _) -> False
-          App (App OptionalFold _) (Some _) -> False
-          App (App OptionalFold _) (App None _) -> False
           App TextShow (TextLit (Chunks [] _)) ->
               False
+          App (App (App TextReplace (TextLit (Chunks [] ""))) _) _ ->
+              False
+          App (App (App TextReplace (TextLit (Chunks [] _))) _) (TextLit _) ->
+              False
           _ -> True
       Let _ _ -> False
       Annot _ _ -> False
@@ -854,7 +844,14 @@ isNormalized e0 = loop (Syntax.denote e0)
               TextLit _ -> False
               _         -> True
       TextAppend _ _ -> False
+      TextReplace -> True
       TextShow -> True
+      Date -> True
+      DateLiteral _ -> True
+      Time -> True
+      TimeLiteral _ _ -> True
+      TimeZone -> True
+      TimeZoneLiteral _ -> True
       List -> True
       ListLit t es -> all loop t && all loop es
       ListAppend x y -> loop x && loop y && decide x y
@@ -873,24 +870,28 @@ isNormalized e0 = loop (Syntax.denote e0)
       Optional -> True
       Some a -> loop a
       None -> True
-      OptionalFold -> True
-      OptionalBuild -> True
-      Record kts -> Dhall.Map.isSorted kts && all loop kts
-      RecordLit kvs -> Dhall.Map.isSorted kvs && all loop kvs
+      Record kts -> Dhall.Map.isSorted kts && all decide kts
+        where
+          decide (RecordField Nothing exp' Nothing Nothing) = loop exp'
+          decide _ = False
+      RecordLit kvs -> Dhall.Map.isSorted kvs && all decide kvs
+        where
+          decide (RecordField Nothing exp' Nothing Nothing) = loop exp'
+          decide _ = False
       Union kts -> Dhall.Map.isSorted kts && all (all loop) kts
-      Combine _ x y -> loop x && loop y && decide x y
+      Combine _ _ x y -> loop x && loop y && decide x y
         where
           decide (RecordLit m) _ | Data.Foldable.null m = False
           decide _ (RecordLit n) | Data.Foldable.null n = False
           decide (RecordLit _) (RecordLit _) = False
           decide  _ _ = True
-      CombineTypes x y -> loop x && loop y && decide x y
+      CombineTypes _ x y -> loop x && loop y && decide x y
         where
           decide (Record m) _ | Data.Foldable.null m = False
           decide _ (Record n) | Data.Foldable.null n = False
           decide (Record _) (Record _) = False
           decide  _ _ = True
-      Prefer _ x y -> loop x && loop y && decide x y
+      Prefer _ _ x y -> loop x && loop y && decide x y
         where
           decide (RecordLit m) _ | Data.Foldable.null m = False
           decide _ (RecordLit n) | Data.Foldable.null n = False
@@ -908,26 +909,27 @@ isNormalized e0 = loop (Syntax.denote e0)
       ToMap x t -> case x of
           RecordLit _ -> False
           _ -> loop x && all loop t
-      Field r k -> case r of
+      Field r (FieldSelection Nothing k Nothing) -> case r of
           RecordLit _ -> False
           Project _ _ -> False
-          Prefer _ (RecordLit m) _ -> Dhall.Map.keys m == [k] && loop r
-          Prefer _ _ (RecordLit _) -> False
-          Combine _ (RecordLit m) _ -> Dhall.Map.keys m == [k] && loop r
-          Combine _ _ (RecordLit m) -> Dhall.Map.keys m == [k] && loop r
+          Prefer _ _ (RecordLit m) _ -> Dhall.Map.keys m == [k] && loop r
+          Prefer _ _ _ (RecordLit _) -> False
+          Combine _ _ (RecordLit m) _ -> Dhall.Map.keys m == [k] && loop r
+          Combine _ _ _ (RecordLit m) -> Dhall.Map.keys m == [k] && loop r
           _ -> loop r
+      Field _ _ -> False
       Project r p -> loop r &&
           case p of
               Left s -> case r of
                   RecordLit _ -> False
                   Project _ _ -> False
-                  Prefer _ _ (RecordLit _) -> False
-                  _ -> not (Dhall.Set.null s) && Dhall.Set.isSorted s
+                  Prefer _ _ _ (RecordLit _) -> False
+                  _ -> not (null s) && Data.Set.toList (Data.Set.fromList s) == s
               Right e' -> case e' of
                   Record _ -> False
                   _ -> loop e'
       Assert t -> loop t
-      Equivalent l r -> loop l && loop r
+      Equivalent _ l r -> loop l && loop r
       With{} -> False
       Note _ e' -> loop e'
       ImportAlt _ _ -> False
@@ -939,7 +941,7 @@ isNormalized e0 = loop (Syntax.denote e0)
 True
 >>> "x" `freeIn` "y"
 False
->>> "x" `freeIn` Lam "x" (Const Type) "x"
+>>> "x" `freeIn` Lam mempty (Syntax.makeFunctionBinding "x" (Const Type)) "x"
 False
 -}
 freeIn :: Eq a => Var -> Expr s a -> Bool
@@ -951,3 +953,8 @@ variable@(V var i) `freeIn` expression =
     denote' = Syntax.denote
 
     strippedExpression = denote' expression
+{-# INLINABLE freeIn #-}
+
+{- $setup
+>>> import Dhall.Syntax (Const(..))
+-}
