@@ -1,4 +1,5 @@
-{-# LANGUAGE RecordWildCards #-}
+{-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards   #-}
 
 -- | This module contains Dhall's parsing logic
 
@@ -21,22 +22,19 @@ module Dhall.Parser (
     ) where
 
 import Control.Exception (Exception)
-import Data.Semigroup (Semigroup(..))
-import Data.Text (Text)
-import Data.Void (Void)
+import Data.Text         (Text)
+import Data.Void         (Void)
+import Dhall.Src         (Src (..))
 import Dhall.Syntax
-import Dhall.Src (Src(..))
-import Prelude hiding (const, pi)
-import Text.Megaparsec (ParseErrorBundle(..), PosState(..))
+import Text.Megaparsec   (ParseErrorBundle (..), PosState (..))
 
-import qualified Data.Char
-import qualified Data.Text
+import qualified Data.Text       as Text
 import qualified Dhall.Core      as Core
 import qualified Text.Megaparsec
 
 import Dhall.Parser.Combinators
-import Dhall.Parser.Token hiding (text)
 import Dhall.Parser.Expression
+import Dhall.Parser.Token       hiding (text)
 
 -- | Parser for a top-level Dhall expression
 expr :: Parser (Expr Src Import)
@@ -78,7 +76,7 @@ instance Show ParseError where
 
 instance Exception ParseError
 
--- | Parse an expression from `Text` containing a Dhall program
+-- | Parse an expression from `Text.Text` containing a Dhall program
 exprFromText
   :: String -- ^ User-friendly name describing the input expression,
             --   used in parsing error messages
@@ -94,8 +92,15 @@ newtype Header = Header Text deriving Show
 
 -- | Create a header with stripped leading spaces and trailing newlines
 createHeader :: Text -> Header
-createHeader =
-    Header . Data.Text.dropWhile Data.Char.isSpace . Data.Text.dropWhileEnd (/= '\n')
+createHeader text = Header (prefix <> newSuffix)
+  where
+    isWhitespace c = c == ' ' || c == '\n' || c == '\r' || c == '\t'
+
+    prefix = Text.dropAround isWhitespace text
+
+    newSuffix
+        | Text.null prefix = ""
+        | otherwise        = "\n"
 
 -- | Like `exprFromText` but also returns the leading comments and whitespace
 -- (i.e. header) up to the last newline before the code begins
