@@ -1,47 +1,37 @@
 {-# LANGUAGE NamedFieldPuns      #-}
-{-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE OverloadedLists     #-}
 {-# LANGUAGE OverloadedStrings   #-}
+{-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 
 -- | Parsing Dhall expressions.
 module Dhall.Parser.Expression where
 
-import Control.Applicative (liftA2, Alternative(..), optional)
-import Data.ByteArray.Encoding (Base(..))
-import Data.Foldable (foldl')
-import Data.Functor (void)
-import Data.List.NonEmpty (NonEmpty(..))
-import Data.Semigroup (Semigroup(..))
-import Data.Text (Text)
+import Control.Applicative     (Alternative (..), liftA2, optional)
+import Data.Foldable           (foldl')
+import Data.List.NonEmpty      (NonEmpty (..))
+import Data.Text               (Text)
+import Dhall.Src               (Src (..))
 import Dhall.Syntax
-import Dhall.Src (Src(..))
-import Prelude hiding (const, pi)
 import Text.Parser.Combinators (choice, try, (<?>))
 
 import qualified Control.Monad
+import qualified Control.Monad.Combinators          as Combinators
 import qualified Control.Monad.Combinators.NonEmpty as Combinators.NonEmpty
-import qualified Data.ByteArray.Encoding
-import qualified Data.ByteString
+import qualified Data.ByteString.Base16             as Base16
 import qualified Data.Char                          as Char
 import qualified Data.List
 import qualified Data.List.NonEmpty                 as NonEmpty
 import qualified Data.Sequence
 import qualified Data.Text
 import qualified Data.Text.Encoding
+import qualified Data.Time                          as Time
 import qualified Dhall.Crypto
 import qualified Text.Megaparsec
 
 import Dhall.Parser.Combinators
 import Dhall.Parser.Token
 
--- | Get the current source position
-getSourcePos :: Text.Megaparsec.MonadParsec e s m =>
-                m Text.Megaparsec.SourcePos
-getSourcePos =
-    Text.Megaparsec.getSourcePos
-{-# INLINE getSourcePos #-}
-
 -- | Get the current source offset (in tokens)
 getOffset :: Text.Megaparsec.MonadParsec e s m => m Int
 getOffset = Text.Megaparsec.stateOffset <$> Text.Megaparsec.getParserState
@@ -59,19 +49,27 @@ setOffset o = Text.Megaparsec.updateParserState $ \state ->
 -}
 src :: Parser a -> Parser Src
 src parser = do
-    before      <- getSourcePos
+    before      <- Text.Megaparsec.getSourcePos
     (tokens, _) <- Text.Megaparsec.match parser
-    after       <- getSourcePos
+    after       <- Text.Megaparsec.getSourcePos
     return (Src before after tokens)
 
+-- | Same as `src`, except also return the parsed value
+srcAnd :: Parser a -> Parser (Src, a)
+srcAnd parser = do
+    before      <- Text.Megaparsec.getSourcePos
+    (tokens, x) <- Text.Megaparsec.match parser
+    after       <- Text.Megaparsec.getSourcePos
+    return (Src before after tokens, x)
+
 {-| Wrap a `Parser` to still match the same text, but to wrap the resulting
     `Expr` in a `Note` constructor containing the `Src` span
 -}
 noted :: Parser (Expr Src a) -> Parser (Expr Src a)
 noted parser = do
-    before      <- getSourcePos
+    before      <- Text.Megaparsec.getSourcePos
     (tokens, e) <- Text.Megaparsec.match parser
-    after       <- getSourcePos
+    after       <- Text.Megaparsec.getSourcePos
     let src₀ = Src before after tokens
     case e of
         Note src₁ _ | laxSrcEq src₀ src₁ -> return e
@@ -107,13 +105,182 @@ importExpression embedded = importExpression_
 data Parsers a = Parsers
     { completeExpression_ :: Parser (Expr Src a)
     , importExpression_   :: Parser (Expr Src a)
+    , letBinding          :: Parser (Binding Src a)
     }
 
--- | Given a parser for imports, 
-parsers :: Parser a -> Parsers a
-parsers embedded = Parsers {..}
+{-| Parse a numeric `TimeZone`
+
+    This corresponds to the @time-numoffset@ rule from the official grammar
+-}
+timeNumOffset :: Parser (Expr s a)
+timeNumOffset = do
+    s <- signPrefix
+
+    hour <- timeHour
+
+    _ <- text ":"
+
+    minute <- timeMinute
+
+    let minutes = s (hour * 60 + minute)
+
+    return (TimeZoneLiteral (Time.TimeZone minutes Prelude.False ""))
+
+{-| Parse a numeric `TimeZone` or a @Z@
+
+    This corresponds to the @time-offset@ rule from the official grammar
+-}
+timeOffset :: Parser (Expr s a)
+timeOffset =
+        (do _ <- text "Z"
+
+            return (TimeZoneLiteral (Time.TimeZone 0 Prelude.False ""))
+        )
+    <|> timeNumOffset
+
+{-| Parse a `Time`
+
+    This corresponds to the @partial-time@ rule from the official grammar
+-}
+partialTime :: Parser (Expr s a)
+partialTime = do
+    hour <- timeHour
+
+    _ <- text ":"
+
+    minute <- timeMinute
+
+    _ <- text ":"
+
+    second <- timeSecond
+
+    (fraction, precision) <- timeSecFrac <|> pure (0, 0)
+
+    let time = Time.TimeOfDay hour minute (second + fraction)
+
+    return (TimeLiteral time precision)
+
+{-| Parse a `Date`
+
+    This corresponds to the @full-date@ rule from the official grammar
+-}
+fullDate :: Parser (Expr s a)
+fullDate = do
+    year <- dateFullYear
+
+    _ <- text "-"
+
+    month <- dateMonth
+
+    _ <- text "-"
+
+    day <- dateMday
+
+    case Time.fromGregorianValid year month day of
+        Nothing -> fail "Invalid calendar day"
+        Just d  -> return (DateLiteral d)
+
+{-| Parse a `Date`, `Time`, `TimeZone` or any valid permutation of them as a
+    record
+
+    This corresponds to the @temporal-literal@ rule from the official grammar
+-}
+temporalLiteral :: Parser (Expr s a)
+temporalLiteral =
+        try (do
+            date <- fullDate
+
+            _ <- text "T" <|> text "t"
+
+            time <- partialTime
+
+            timeZone <- timeOffset
+
+            return
+                (RecordLit
+                    [   ("date"    , makeRecordField date)
+                    ,   ("time"    , makeRecordField time)
+                    ,   ("timeZone", makeRecordField timeZone)
+                    ]
+                )
+        )
+    <|> try (do
+            date <- fullDate
+
+            _ <- text "T" <|> text "t"
+
+            time <- partialTime
+
+            return
+                (RecordLit
+                    [   ("date", makeRecordField date)
+                    ,   ("time", makeRecordField time)
+                    ]
+                )
+        )
+    <|> try (do
+            time <- partialTime
+
+            timeZone <- timeOffset
+
+            return
+                (RecordLit
+                    [   ("time"    , makeRecordField time)
+                    ,   ("timeZone", makeRecordField timeZone)
+                    ]
+                )
+        )
+    <|> try fullDate
+    <|> try partialTime
+    <|> try timeNumOffset
+
+-- | Given a parser for imports,
+parsers :: forall a. Parser a -> Parsers a
+parsers embedded = Parsers{..}
   where
-    completeExpression_ = whitespace *> expression <* whitespace
+    completeExpression_ =
+            many shebang
+        *>  whitespace
+        *>  expression
+        <*  whitespace
+        <*  optional lineCommentPrefix
+
+    shebang = do
+        _ <- text "#!"
+
+        let predicate c = ('\x20' <= c && c <= '\x10FFFF') || c == '\t'
+
+        _ <- Dhall.Parser.Combinators.takeWhile predicate
+
+        endOfLine
+
+    letBinding = do
+        src0 <- try (_let *> src nonemptyWhitespace)
+
+        c <- label
+
+        src1 <- src whitespace
+
+        d <- optional (do
+            _colon
+
+            src2 <- src nonemptyWhitespace
+
+            e <- expression
+
+            whitespace
+
+            return (Just src2, e) )
+
+        _equal
+
+        src3 <- src whitespace
+
+        f <- expression
+
+        whitespace
+
+        return (Binding (Just src0) c (Just src1) d (Just src3) f)
 
     expression =
         noted
@@ -128,22 +295,22 @@ parsers embedded = Parsers {..}
             ) <?> "expression"
       where
         alternative0 = do
-            _lambda
+            cs <- _lambda
             whitespace
             _openParens
-            whitespace
+            src0 <- src whitespace
             a <- label
-            whitespace
+            src1 <- src whitespace
             _colon
-            nonemptyWhitespace
+            src2 <- src nonemptyWhitespace
             b <- expression
             whitespace
             _closeParens
             whitespace
-            _arrow
+            cs' <- _arrow
             whitespace
             c <- expression
-            return (Lam a b c)
+            return (Lam (Just (cs <> cs')) (FunctionBinding (Just src0) a (Just src1) (Just src2) b) c)
 
         alternative1 = do
             try (_if *> nonemptyWhitespace)
@@ -157,35 +324,7 @@ parsers embedded = Parsers {..}
             return (BoolIf a b c)
 
         alternative2 = do
-            let binding = do
-                    src0 <- try (_let *> src nonemptyWhitespace)
-
-                    c <- label
-
-                    src1 <- src whitespace
-
-                    d <- optional (do
-                        _colon
-
-                        src2 <- src nonemptyWhitespace
-
-                        e <- expression
-
-                        whitespace
-
-                        return (Just src2, e) )
-
-                    _equal
-
-                    src3 <- src whitespace
-
-                    f <- expression
-
-                    whitespace
-
-                    return (Binding (Just src0) c (Just src1) d (Just src3) f)
-
-            as <- NonEmpty.some1 binding
+            as <- NonEmpty.some1 letBinding
 
             try (_in *> nonemptyWhitespace)
 
@@ -211,7 +350,7 @@ parsers embedded = Parsers {..}
             return (Dhall.Syntax.wrapInLets as b)
 
         alternative3 = do
-            try (_forall *> whitespace *> _openParens)
+            cs <- try (_forall <* whitespace <* _openParens)
             whitespace
             a <- label
             whitespace
@@ -221,10 +360,10 @@ parsers embedded = Parsers {..}
             whitespace
             _closeParens
             whitespace
-            _arrow
+            cs' <- _arrow
             whitespace
             c <- expression
-            return (Pi a b c)
+            return (Pi (Just (cs <> cs')) a b c)
 
         alternative4 = do
             try (_assert *> whitespace *> _colon)
@@ -233,53 +372,18 @@ parsers embedded = Parsers {..}
             return (Assert a)
 
         alternative5 = do
-            a0 <- applicationExpression
+            (a0Info, a0) <- applicationExpressionWithInfo
 
             let (parseFirstOperatorExpression, parseOperatorExpression) =
                     operatorExpression (pure a0)
 
-            a <- parseFirstOperatorExpression
+            let alternative5A = do
+                    case a0Info of
+                        ImportExpr -> return ()
+                        _          -> empty
 
-            let alternative4A = do
-                    _arrow
-                    whitespace
-                    b <- expression
-                    whitespace
-                    return (Pi "_" a b)
-
-            let alternative4B = do
-                    _colon
-                    nonemptyWhitespace
-                    b <- expression
-                    case shallowDenote a of
-                        ListLit Nothing [] ->
-                            return (ListLit (Just b) [])
-                        Merge c d Nothing ->
-                            return (Merge c d (Just b))
-                        ToMap c Nothing ->
-                            return (ToMap c (Just b))
-                        _ -> return (Annot a b)
-
-            let alternative4C = do
-                    case shallowDenote a of
-                        Equivalent{}   -> empty
-                        ImportAlt{}    -> empty
-                        BoolOr{}       -> empty
-                        NaturalPlus{}  -> empty
-                        TextAppend{}   -> empty
-                        ListAppend{}   -> empty
-                        BoolAnd{}      -> empty
-                        Combine{}      -> empty
-                        Prefer{}       -> empty
-                        CombineTypes{} -> empty
-                        NaturalTimes{} -> empty
-                        BoolEQ{}       -> empty
-                        BoolNE{}       -> empty
-                        App{}          -> empty
-                        _              -> return ()
-
-                    bs <- many (do
-                        try (whitespace *> _with *> nonemptyWhitespace)
+                    bs <- some (do
+                        try (nonemptyWhitespace *> _with *> nonemptyWhitespace)
 
                         keys <- Combinators.NonEmpty.sepBy1 anyLabel (try (whitespace *> _dot) *> whitespace)
 
@@ -295,7 +399,48 @@ parsers embedded = Parsers {..}
 
                     return (foldl (\e f -> f e) a0 bs)
 
-            alternative4A <|> alternative4B <|> alternative4C <|> pure a
+            let alternative5B = do
+                    a <- parseFirstOperatorExpression
+
+                    whitespace
+
+                    let alternative5B0 = do
+                            cs <- _arrow
+                            whitespace
+                            b <- expression
+                            whitespace
+                            return (Pi (Just cs) "_" a b)
+
+                    let alternative5B1 = do
+                            _colon
+                            nonemptyWhitespace
+                            case (shallowDenote a, a0Info) of
+                                (ListLit Nothing [], _) -> do
+                                    b <- expression
+
+                                    return (ListLit (Just b) [])
+                                (Merge c d Nothing, NakedMergeOrSomeOrToMap) -> do
+                                    b <- expression
+
+                                    return (Merge c d (Just b))
+                                (ToMap c Nothing, NakedMergeOrSomeOrToMap) -> do
+                                    b <- expression
+
+                                    return (ToMap c (Just b))
+                                _ -> do
+                                    b <- expression
+
+                                    return (Annot a b)
+
+                    let alternative5B2 =
+                            case shallowDenote a of
+                                ListLit Nothing [] ->
+                                    fail "Empty list literal without annotation"
+                                _ -> pure a
+
+                    alternative5B0 <|> alternative5B1 <|> alternative5B2
+
+            alternative5A <|> alternative5B
 
     -- The firstApplicationExpression argument is necessary in order to
     -- left-factor the parsers for function types and @with@ expressions to
@@ -314,55 +459,68 @@ parsers embedded = Parsers {..}
 
         nil = (firstApplicationExpression, applicationExpression)
 
-    makeOperatorExpression firstSubExpression operatorParser subExpression =
-            noted (do
-                a <- firstSubExpression
+    makeOperatorExpression firstSubExpression operatorParser subExpression = do
+            a <- firstSubExpression
 
-                whitespace
+            bs <- Text.Megaparsec.many $ do
+                (Src _ _ textOp, op0) <- srcAnd (try (whitespace *> operatorParser))
 
-                b <- Text.Megaparsec.many $ do
-                    op <- operatorParser
+                r0 <- subExpression
 
-                    r  <- subExpression
+                let l@(Note (Src startL _ textL) _) `op` r@(Note (Src _ endR textR) _) =
+                        Note (Src startL endR (textL <> textOp <> textR)) (l `op0` r)
+                    -- We shouldn't hit this branch if things are working, but
+                    -- that is not enforced in the types
+                    l `op` r =
+                        l `op0` r
 
-                    whitespace
+                return (`op` r0)
 
-                    return (\l -> l `op` r)
-                return (foldl' (\x f -> f x) a b))
+            return (foldl' (\x f -> f x) a bs)
 
     operatorParsers :: [Parser (Expr s a -> Expr s a -> Expr s a)]
     operatorParsers =
-        [ Equivalent              <$ _equivalent   <* whitespace
-        , ImportAlt               <$ _importAlt    <* nonemptyWhitespace
-        , BoolOr                  <$ _or           <* whitespace
-        , NaturalPlus             <$ _plus         <* nonemptyWhitespace
-        , TextAppend              <$ _textAppend   <* whitespace
-        , ListAppend              <$ _listAppend   <* whitespace
-        , BoolAnd                 <$ _and          <* whitespace
-        , Combine Nothing         <$ _combine      <* whitespace
-        , Prefer PreferFromSource <$ _prefer       <* whitespace
-        , CombineTypes            <$ _combineTypes <* whitespace
-        , NaturalTimes            <$ _times        <* whitespace
+        [ Equivalent . Just           <$> _equivalent   <* whitespace
+        , ImportAlt                   <$ _importAlt     <* nonemptyWhitespace
+        , BoolOr                      <$ _or            <* whitespace
+        , NaturalPlus                 <$ _plus          <* nonemptyWhitespace
+        , TextAppend                  <$ _textAppend    <* whitespace
+        , ListAppend                  <$ _listAppend    <* whitespace
+        , BoolAnd                     <$ _and           <* whitespace
+        , (\cs -> Combine (Just cs) Nothing)         <$> _combine <* whitespace
+        , (\cs -> Prefer (Just cs) PreferFromSource) <$> _prefer  <* whitespace
+        , CombineTypes . Just         <$> _combineTypes <* whitespace
+        , NaturalTimes                <$ _times         <* whitespace
         -- Make sure that `==` is not actually the prefix of `===`
-        , BoolEQ                  <$ try (_doubleEqual <* Text.Megaparsec.notFollowedBy (char '=')) <* whitespace
-        , BoolNE                  <$ _notEqual     <* whitespace
+        , BoolEQ                      <$ try (_doubleEqual <* Text.Megaparsec.notFollowedBy (char '=')) <* whitespace
+        , BoolNE                      <$ _notEqual      <* whitespace
         ]
 
-    applicationExpression = do
+    applicationExpression = snd <$> applicationExpressionWithInfo
+
+    applicationExpressionWithInfo :: Parser (ApplicationExprInfo, Expr Src a)
+    applicationExpressionWithInfo = do
             let alternative0 = do
-                    _ <- try (_Some <* nonemptyWhitespace)
+                    try (_merge *> nonemptyWhitespace)
+
+                    a <- importExpression_ <* nonemptyWhitespace
+
+                    return (\b -> Merge a b Nothing, Just "second argument to ❰merge❱")
+
+            let alternative1 = do
+                    try (_Some *> nonemptyWhitespace)
 
                     return (Some, Just "argument to ❰Some❱")
 
-            let alternative1 = do
-                    _ <- try (_toMap *> nonemptyWhitespace)
+            let alternative2 = do
+                    try (_toMap *> nonemptyWhitespace)
 
                     return (\a -> ToMap a Nothing, Just "argument to ❰toMap❱")
 
-            let alternative2 = do
+            let alternative3 =
                     return (id, Nothing)
 
-            (f, maybeMessage) <- alternative0 <|> alternative1 <|> alternative2
+            (f, maybeMessage) <- alternative0 <|> alternative1 <|> alternative2 <|> alternative3
 
             let adapt parser =
                     case maybeMessage of
@@ -376,7 +534,15 @@ parsers embedded = Parsers {..}
                 b <- importExpression_
                 return (sep, b)
 
-            return (foldl' app (f a) bs)
+            let c = foldl' app (f a) bs
+
+            let info =
+                    case (maybeMessage, bs) of
+                        (Just _ , []) -> NakedMergeOrSomeOrToMap
+                        (Nothing, []) -> ImportExpr
+                        _             -> ApplicationExpr
+
+            return (info, c)
           where
             app a (sep, b)
                 | Note (Src left _ bytesL) _ <- a
@@ -416,25 +582,43 @@ parsers embedded = Parsers {..}
             let projectBySet        xs e = Project e (Left  xs)
             let projectByExpression xs e = Project e (Right xs)
 
-            let alternatives =
-                        fmap field               anyLabel
-                    <|> fmap projectBySet        labels
-                    <|> fmap projectByExpression recordType
+            let alternatives = do
+                    src0 <- src whitespace
+
+                    let fieldSelection = do
+                            l <- anyLabel
+
+                            pos <- Text.Megaparsec.getSourcePos
+
+                            -- FIXME: Suffix whitespace can't be parsed given our limitation
+                            -- about whitespace treatment, but for @dhall-docs@ this
+                            -- is enough
+                            let src1 = Src pos pos ""
+
+                            return (FieldSelection (Just src0) l (Just src1))
+
+                    let result =
+                                fmap field               fieldSelection
+                            <|> fmap projectBySet        labels
+                            <|> fmap projectByExpression recordType
+
+                    result
+
+            b <- Text.Megaparsec.many (try (whitespace *> _dot *> alternatives))
 
-            b <- Text.Megaparsec.many (try (whitespace *> _dot *> whitespace *> alternatives))
             return (foldl' (\e k -> k e) a b) )
 
     primitiveExpression =
             noted
                 ( choice
-                    [ alternative00
+                    [ temporalLiteral
+                    , alternative00
                     , alternative01
                     , alternative02
-                    , alternative03
+                    , textLiteral
                     , alternative04
-                    , alternative05
-                    , alternative06
-                    , alternative07
+                    , unionType
+                    , listLiteral
                     , alternative37
                     , alternative09
                     , builtin
@@ -458,34 +642,25 @@ parsers embedded = Parsers {..}
                 a <- try integerLiteral
                 return (IntegerLit a)
 
-            alternative03 = textLiteral
-
             alternative04 = (do
                 _openBrace
 
-                whitespace
+                src0 <- src whitespace
+                mComma <- optional _comma
 
-                _ <- optional (_comma *> whitespace)
+                -- `src1` corresponds to the prefix whitespace of the first key-value
+                -- pair. This is done to avoid using `try` to recover the consumed
+                -- whitespace when the comma is not consumed
+                src1 <- case mComma of
+                    Nothing -> return src0
+                    Just _ -> src whitespace
 
-                a <- recordTypeOrLiteral
-
-                whitespace
+                a <- recordTypeOrLiteral src1
 
                 _closeBrace
 
                 return a ) <?> "literal"
 
-            alternative05 = unionType
-
-            alternative06 = listLiteral
-
-            alternative07 = do
-                try (_merge *> nonemptyWhitespace)
-                a <- importExpression_
-                nonemptyWhitespace
-                b <- importExpression_ <?> "second argument to ❰merge❱"
-                return (Merge a b Nothing)
-
             alternative09 = do
                 a <- try doubleInfinity
                 return (DoubleLit (DhallDouble a))
@@ -533,7 +708,8 @@ parsers embedded = Parsers {..}
 
                     'D' ->
                         choice
-                            [ DoubleShow       <$ _DoubleShow
+                            [ Date             <$ _Date
+                            , DoubleShow       <$ _DoubleShow
                             , Double           <$ _Double
                             ]
                     'L' ->
@@ -547,18 +723,16 @@ parsers embedded = Parsers {..}
                             , ListReverse      <$ _ListReverse
                             , List             <$ _List
                             ]
-                    'O' ->
-                        choice
-                            [ OptionalFold     <$ _OptionalFold
-                            , OptionalBuild    <$ _OptionalBuild
-                            , Optional         <$ _Optional
-                            ]
+                    'O' ->    Optional         <$ _Optional
                     'B' ->    Bool             <$ _Bool
                     'S' ->    Const Sort       <$ _Sort
                     'T' ->
                         choice
-                            [ TextShow         <$ _TextShow
+                            [ TextReplace      <$ _TextReplace
+                            , TextShow         <$ _TextShow
                             , Text             <$ _Text
+                            , TimeZone         <$ _TimeZone
+                            , Time             <$ _Time
                             , BoolLit True     <$ _True
                             , Const Type       <$ _Type
                             ]
@@ -647,7 +821,7 @@ parsers embedded = Parsers {..}
 
                     let fourCharacterEscapeSequence = do
                             ns <- Control.Monad.replicateM 4 hexNumber
-                            
+
                             let number = toNumber ns
 
                             Control.Monad.guard (validCodepoint number)
@@ -687,7 +861,7 @@ parsers embedded = Parsers {..}
                 , unescapedCharacterFast
                 , unescapedCharacterSlow
                 , tab
-                , endOfLine
+                , endOfLine_
                 ]
           where
                 escapeSingleQuotes = do
@@ -726,7 +900,7 @@ parsers embedded = Parsers {..}
                   where
                     predicate c = c == '$' || c == '\''
 
-                endOfLine = do
+                endOfLine_ = do
                     a <- "\n" <|> "\r\n"
                     b <- singleQuoteContinue
                     return (Chunks [] a <> b)
@@ -738,99 +912,125 @@ parsers embedded = Parsers {..}
 
     singleQuoteLiteral = do
             _ <- text "''"
+
             _ <- endOfLine
+
             a <- singleQuoteContinue
 
             return (Dhall.Syntax.toDoubleQuoted a)
-          where
-            endOfLine = (void (char '\n') <|> void (text "\r\n")) <?> "newline"
 
     textLiteral = (do
             literal <- doubleQuotedLiteral <|> singleQuoteLiteral
             return (TextLit literal) ) <?> "literal"
 
-    recordTypeOrLiteral =
+    recordTypeOrLiteral firstSrc0 =
             choice
-                [ alternative0
-                , alternative1
-                , alternative2
+                [ emptyRecordLiteral
+                , nonEmptyRecordTypeOrLiteral firstSrc0
+                , emptyRecordType
                 ]
-          where
-            alternative0 = do
-                _equal
-                return (RecordLit mempty)
 
-            alternative1 = nonEmptyRecordTypeOrLiteral
+    emptyRecordLiteral = do
+        _equal
 
-            alternative2 = return (Record mempty)
+        _ <- optional (try (whitespace *> _comma))
 
-    nonEmptyRecordTypeOrLiteral = do
+        whitespace
+        return (RecordLit mempty)
+
+    emptyRecordType = return (Record mempty)
+
+    nonEmptyRecordTypeOrLiteral firstSrc0 = do
             let nonEmptyRecordType = do
-                    a <- try (anyLabel <* whitespace <* _colon)
+                    (firstKeySrc1, a) <- try $ do
+                        a <- anyLabelOrSome
+                        s <- src whitespace
+                        _colon
+                        return (s, a)
 
-                    nonemptyWhitespace
+                    firstKeySrc2 <- src nonemptyWhitespace
 
                     b <- expression
 
-                    whitespace
+                    e <- Text.Megaparsec.many $ do
+                        (src0', c) <- try $ do
+                            _comma
+                            src0' <- src whitespace
+                            c <- anyLabelOrSome
+                            return (src0', c)
 
-                    e <- Text.Megaparsec.many (do
-                        _comma
-
-                        whitespace
-
-                        c <- anyLabel
-
-                        whitespace
+                        src1 <- src whitespace
 
                         _colon
 
-                        nonemptyWhitespace
+                        src2 <- src nonemptyWhitespace
 
                         d <- expression
 
                         whitespace
 
-                        return (c, d) )
+                        return (c, RecordField (Just src0') d (Just src1) (Just src2))
 
-                    m <- toMap ((a, b) : e)
+                    _ <- optional (whitespace *> _comma)
+                    whitespace
+
+                    m <- toMap ((a, RecordField (Just firstSrc0) b (Just firstKeySrc1) (Just firstKeySrc2)) : e)
 
                     return (Record m)
 
-            let keysValue = do
-                    keys <- Combinators.NonEmpty.sepBy1 anyLabel (try (whitespace *> _dot) *> whitespace)
+            let keysValue maybeSrc = do
+                    firstSrc0' <- case maybeSrc of
+                        Just src0 -> return src0
+                        Nothing -> src whitespace
+                    firstLabel <- anyLabelOrSome
+                    firstSrc1 <- src whitespace
+
+                    let parseLabelWithWhsp = try $ do
+                            _dot
+                            src0 <- src whitespace
+                            l <- anyLabelOrSome
+                            src1 <- src whitespace
+                            return (src0, l, src1)
+
+                    restKeys <- Combinators.many parseLabelWithWhsp
+                    let keys = (firstSrc0', firstLabel, firstSrc1) :| restKeys
 
                     let normalRecordEntry = do
-                            try (whitespace *> _equal)
+                            try _equal
 
-                            whitespace
+                            lastSrc2 <- src whitespace
 
                             value <- expression
 
-                            let cons key (key', values) =
-                                    (key, RecordLit [ (key', values) ])
+                            let cons (s0, key, s1) (key', values) =
+                                    (key, RecordField (Just s0) (RecordLit [ (key', values) ]) (Just s1) Nothing)
 
-                            let nil = (NonEmpty.last keys, value)
+                            let (lastSrc0, lastLabel, lastSrc1) = NonEmpty.last keys
+                            let nil = (lastLabel, RecordField (Just lastSrc0) value (Just lastSrc1) (Just lastSrc2))
 
                             return (foldr cons nil (NonEmpty.init keys))
 
                     let punnedEntry =
                             case keys of
-                                x :| [] -> return (x, Var (V x 0))
+                                (s0, x, s1) :| [] -> return (x, RecordField (Just s0) (Var (V x 0)) (Just s1) Nothing)
                                 _       -> empty
 
                     (normalRecordEntry <|> punnedEntry) <* whitespace
 
             let nonEmptyRecordLiteral = do
-                    as <- Text.Megaparsec.sepBy1 keysValue (_comma *> whitespace)
+                    a <- keysValue (Just firstSrc0)
 
-                    {- The `flip` is necessary because `toMapWith` is internally
-                       based on `Data.Map.fromListWithKey` which combines keys
-                       in reverse order
-                    -}
-                    let combine k = liftA2 (flip (Combine (Just k)))
+                    as <- many (try (_comma *> keysValue Nothing))
 
-                    m <- toMapWith combine as
+                    _ <- optional (whitespace *> _comma)
+
+                    whitespace
+
+                    let combine k = liftA2 $ \rf rf' -> makeRecordField $ Combine mempty (Just k)
+                                                            (recordFieldValue rf')
+                                                            (recordFieldValue rf)
+
+                    m <- toMapWith combine (a : as)
 
                     return (RecordLit m)
 
@@ -841,34 +1041,59 @@ parsers embedded = Parsers {..}
 
             whitespace
 
-            _ <- optional (_bar *> whitespace)
-
             let unionTypeEntry = do
-                    a <- anyLabel
+                    a <- anyLabelOrSome
+
                     whitespace
+
                     b <- optional (_colon *> nonemptyWhitespace *> expression <* whitespace)
+
                     return (a, b)
 
-            kvs <- Text.Megaparsec.sepBy unionTypeEntry (_bar *> whitespace)
+            let nonEmptyUnionType = do
+                    kv <- try (optional (_bar *> whitespace) *> unionTypeEntry)
 
-            m <- toMap kvs
+                    kvs <- many (try (_bar *> whitespace *> unionTypeEntry))
 
-            _closeAngle
+                    m <- toMap (kv : kvs)
 
-            return (Union m) ) <?> "literal"
+                    _ <- optional (_bar *> whitespace)
+
+                    _closeAngle
+
+                    return (Union m)
+
+            let emptyUnionType = do
+                    try (optional (_bar *> whitespace) *> _closeAngle)
+
+                    return (Union mempty)
+
+            nonEmptyUnionType <|> emptyUnionType ) <?> "literal"
 
     listLiteral = (do
             _openBracket
 
             whitespace
 
-            _ <- optional (_comma *> whitespace)
+            let nonEmptyListLiteral = do
+                    a <- try (optional (_comma *> whitespace) *> expression)
 
-            a <- Text.Megaparsec.sepBy (expression <* whitespace) (_comma *> whitespace)
+                    whitespace
 
-            _closeBracket
+                    as <- many (try (_comma *> whitespace *> expression) <* whitespace)
 
-            return (ListLit Nothing (Data.Sequence.fromList a)) ) <?> "literal"
+                    _ <- optional (_comma *> whitespace)
+
+                    _closeBracket
+
+                    return (ListLit Nothing (Data.Sequence.fromList (a : as)))
+
+            let emptyListLiteral = do
+                    try (optional (_comma *> whitespace) *> _closeBracket)
+
+                    return (ListLit Nothing mempty)
+
+            nonEmptyListLiteral <|> emptyListLiteral) <?> "literal"
 
 {-| Parse an environment variable import
 
@@ -973,9 +1198,9 @@ importHash_ = do
     _ <- text "sha256:"
     t <- count 64 (satisfy hexdig <?> "hex digit")
     let strictBytes16 = Data.Text.Encoding.encodeUtf8 t
-    strictBytes <- case Data.ByteArray.Encoding.convertFromBase Base16 strictBytes16 of
+    strictBytes <- case Base16.decode strictBytes16 of
         Left  string      -> fail string
-        Right strictBytes -> return (strictBytes :: Data.ByteString.ByteString)
+        Right strictBytes -> return strictBytes
     case Dhall.Crypto.sha256DigestFromByteString strictBytes of
       Nothing -> fail "Invalid sha256 hash"
       Just h  -> pure h
@@ -1004,3 +1229,13 @@ import_ = (do
       try (whitespace *> _as *> nonemptyWhitespace)
 
       (_Text >> pure RawText) <|> (_Location >> pure Location)
+
+-- | 'ApplicationExprInfo' distinguishes certain subtypes of application
+-- expressions.
+data ApplicationExprInfo
+    = NakedMergeOrSomeOrToMap
+    -- ^ @merge x y@, @Some x@ or @toMap x@, unparenthesized.
+    | ImportExpr
+    -- ^ An import expression.
+    | ApplicationExpr
+    -- ^ Any other application expression.
