@@ -1,11 +1,15 @@
-{-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE RecordWildCards   #-}
 
 -- | Parse Dhall tokens. Even though we don't have a tokenizer per-se this
 ---  module is useful for keeping some small parsing utilities.
 module Dhall.Parser.Token (
+    endOfLine,
     validCodepoint,
     whitespace,
+    lineComment,
+    lineCommentPrefix,
+    blockComment,
     nonemptyWhitespace,
     bashEnvironmentVariable,
     posixEnvironmentVariable,
@@ -14,16 +18,25 @@ module Dhall.Parser.Token (
     char,
     file_,
     label,
+    anyLabelOrSome,
     anyLabel,
     labels,
     httpRaw,
     hexdig,
     identifier,
     hexNumber,
+    signPrefix,
     doubleLiteral,
     doubleInfinity,
     naturalLiteral,
     integerLiteral,
+    dateFullYear,
+    dateMonth,
+    dateMday,
+    timeHour,
+    timeMinute,
+    timeSecond,
+    timeSecFrac,
     _Optional,
     _if,
     _then,
@@ -57,14 +70,16 @@ module Dhall.Parser.Token (
     _ListLast,
     _ListIndexed,
     _ListReverse,
-    _OptionalFold,
-    _OptionalBuild,
     _Bool,
     _Natural,
     _Integer,
     _Double,
     _Text,
+    _TextReplace,
     _TextShow,
+    _Date,
+    _Time,
+    _TimeZone,
     _List,
     _True,
     _False,
@@ -108,19 +123,18 @@ module Dhall.Parser.Token (
     _with,
     ) where
 
-import           Dhall.Parser.Combinators
+import Dhall.Parser.Combinators
 
-import Control.Applicative (Alternative(..), optional)
-import Data.Bits ((.&.))
-import Data.Functor (void, ($>))
-import Data.Semigroup (Semigroup(..))
-import Data.Text (Text)
+import Control.Applicative     (Alternative (..), optional)
+import Data.Bits               ((.&.))
+import Data.Fixed              (Pico)
+import Data.Functor            (void, ($>))
+import Data.Ratio              ((%))
+import Data.Text               (Text)
 import Dhall.Syntax
-import Dhall.Set (Set)
-import Prelude hiding (const, pi)
 import Text.Parser.Combinators (choice, try, (<?>))
 
-import qualified Control.Monad
+import qualified Control.Monad              as Monad
 import qualified Data.Char                  as Char
 import qualified Data.Foldable
 import qualified Data.HashSet
@@ -128,22 +142,26 @@ import qualified Data.List                  as List
 import qualified Data.List.NonEmpty
 import qualified Data.Scientific            as Scientific
 import qualified Data.Text
-import qualified Dhall.Set
-import qualified Network.URI.Encode         as URI.Encode
 import qualified Text.Megaparsec
 import qualified Text.Megaparsec.Char.Lexer
 import qualified Text.Parser.Char
-import qualified Text.Parser.Token
 import qualified Text.Parser.Combinators
+import qualified Text.Parser.Token
 
 import Numeric.Natural (Natural)
-import Prelude hiding (const, pi)
+
+-- | Match an end-of-line character sequence
+endOfLine :: Parser Text
+endOfLine =
+    (   Text.Parser.Char.text "\n"  
+    <|> Text.Parser.Char.text "\r\n"
+    ) <?> "newline"
 
 -- | Returns `True` if the given `Int` is a valid Unicode codepoint
 validCodepoint :: Int -> Bool
 validCodepoint c =
-    not (category == Char.Surrogate 
-      || c .&. 0xFFFE == 0xFFFE 
+    not (category == Char.Surrogate
+      || c .&. 0xFFFE == 0xFFFE
       || c .&. 0xFFFF == 0xFFFF)
   where
     category = Char.generalCategory (Char.chr c)
@@ -181,13 +199,14 @@ hexdig c =
     ||  ('A' <= c && c <= 'F')
     ||  ('a' <= c && c <= 'f')
 
+-- | Parse a leading @+@ or @-@ sign
 signPrefix :: Num a => Parser (a -> a)
 signPrefix = (do
     let positive = fmap (\_ -> id    ) (char '+')
     let negative = fmap (\_ -> negate) (char '-')
     positive <|> negative ) <?> "sign"
 
-{-| Parse a `Double` literal
+{-| Parse a `Dhall.Syntax.Double` literal
 
     This corresponds to the @double-literal@ rule from the official grammar
 -}
@@ -259,7 +278,7 @@ doubleInfinity = (do
     a <- text "Infinity" >> return (1.0/0.0)
     return (sign a) ) <?> "literal"
 
-{-| Parse an `Integer` literal
+{-| Parse an `Dhall.Syntax.Integer` literal
 
     This corresponds to the @integer-literal@ rule from the official grammar
 -}
@@ -269,7 +288,7 @@ integerLiteral = (do
     a    <- naturalLiteral
     return (sign (fromIntegral a)) ) <?> "literal"
 
-{-| Parse a `Natural` literal 
+{-| Parse a `Dhall.Syntax.Natural` literal
 
     This corresponds to the @natural-literal@ rule from the official grammar
 -}
@@ -299,6 +318,99 @@ naturalLiteral = (do
           where
             step acc x = acc * 10 + x
 
+{-| Parse a 4-digit year
+
+    This corresponds to the @date-fullyear@ rule from the official grammar
+-}
+dateFullYear :: Parser Integer
+dateFullYear = do
+    digits <- Monad.replicateM 4 (Text.Parser.Char.satisfy digit)
+
+    return (digits `base` 10)
+
+{-| Parse a 2-digit month
+
+    This corresponds to the @date-month@ rule from the official grammar
+-}
+dateMonth :: Parser Int
+dateMonth = do
+    digits <- Monad.replicateM 2 (Text.Parser.Char.satisfy digit)
+
+    let month = digits `base` 10
+
+    if 1 <= month && month <= 12
+        then return month
+        else fail "Invalid month"
+
+{-| Parse a 2-digit day of the month
+
+    This corresponds to the @date-mday@ rule from the official grammar
+-}
+dateMday :: Parser Int
+dateMday = do
+    digits <- Monad.replicateM 2 (Text.Parser.Char.satisfy digit)
+
+    let day = digits `base` 10
+
+    if 1 <= day && day <= 31
+        then return day
+        else fail "Invalid day"
+
+{-| Parse a 2-digit hour
+
+    This corresponds to the @time-hour@ rule from the official grammar
+-}
+timeHour :: Parser Int
+timeHour = do
+    digits <- Monad.replicateM 2 (Text.Parser.Char.satisfy digit)
+
+    let hour = digits `base` 10
+
+    if 0 <= hour && hour < 24
+        then return hour
+        else fail "Invalid hour"
+
+{-| Parse a 2-digit minute
+
+    This corresponds to the @time-minute@ rule from the official grammar
+-}
+timeMinute :: Parser Int
+timeMinute = do
+    digits <- Monad.replicateM 2 (Text.Parser.Char.satisfy digit)
+
+    let minute = digits `base` 10
+
+    if 0 <= minute && minute < 60
+        then return minute
+        else fail "Invalid minute"
+
+{-| Parse a 2-digit second
+
+    This corresponds to the @time-second@ rule from the official grammar
+-}
+timeSecond :: Parser Pico
+timeSecond = do
+    digits <- Monad.replicateM 2 (Text.Parser.Char.satisfy digit)
+
+    let second = digits `base` 10
+
+    if 0 <= second && second < 60
+        then return second
+        else fail "Invalid second"
+
+{-| Parse the fractional component of a second
+
+    This corresponds to the @time-secfrac@ rule from the official grammar
+-}
+timeSecFrac :: Parser (Pico, Word)
+timeSecFrac = do
+    _ <- Text.Parser.Char.text "."
+
+    digits <- some (Text.Parser.Char.satisfy digit)
+
+    let precision = fromIntegral (length digits)
+
+    return (fromRational ((digits `base` 10) % (10 ^ precision)), precision)
 
 {-| Parse an identifier (i.e. a variable or built-in)
 
@@ -325,8 +437,8 @@ whitespaceChunk =
     choice
         [ void (Dhall.Parser.Combinators.takeWhile1 predicate)
         , void (Text.Parser.Char.text "\r\n" <?> "newline")
-        , lineComment
-        , blockComment
+        , void lineComment
+        , void blockComment
         ] <?> "whitespace"
   where
     predicate c = c == ' ' || c == '\t' || c == '\n'
@@ -353,29 +465,31 @@ hexNumber = choice [ hexDigit, hexUpper, hexLower ]
       where
         predicate c = 'a' <= c && c <= 'f'
 
-lineComment :: Parser ()
-lineComment = do
+-- | Same as `lineComment` except that this doesn't parse the end-of-line
+--   character
+lineCommentPrefix :: Parser Text
+lineCommentPrefix = do
     _ <- text "--"
 
     let predicate c = ('\x20' <= c && c <= '\x10FFFF') || c == '\t'
 
-    _ <- Dhall.Parser.Combinators.takeWhile predicate
+    commentText <- Dhall.Parser.Combinators.takeWhile predicate
 
-    endOfLine
+    return ("--" <> commentText)
 
-    return ()
-  where
-    endOfLine =
-        (   void (Text.Parser.Char.char '\n'  )
-        <|> void (Text.Parser.Char.text "\r\n")
-        ) <?> "newline"
+-- | Parse a Dhall's single-line comment, starting from `--` and until the
+--   last character of the line /before/ the end-of-line character
+lineComment :: Parser Text
+lineComment = try (lineCommentPrefix <* endOfLine)
 
-blockComment :: Parser ()
+-- | Parsed text doesn't include opening braces
+blockComment :: Parser Text
 blockComment = do
     _ <- text "{-"
-    blockCommentContinue
+    c <- blockCommentContinue
+    pure ("{-" <> c <> "-}")
 
-blockCommentChunk :: Parser ()
+blockCommentChunk :: Parser Text
 blockCommentChunk =
     choice
         [ blockComment  -- Nested block comment
@@ -384,36 +498,36 @@ blockCommentChunk =
         , endOfLine
         ]
   where
-    characters = void (Dhall.Parser.Combinators.takeWhile1 predicate)
+    characters = (Dhall.Parser.Combinators.takeWhile1 predicate)
       where
         predicate c =
                 '\x20' <= c && c <= '\x10FFFF' && c /= '-' && c /= '{'
             ||  c == '\n'
             ||  c == '\t'
 
-    character = void (Text.Parser.Char.satisfy predicate)
+    character = (Dhall.Parser.Combinators.satisfy predicate)
       where
         predicate c = '\x20' <= c && c <= '\x10FFFF' || c == '\n' || c == '\t'
 
-    endOfLine = void (Text.Parser.Char.text "\r\n" <?> "newline")
-
-blockCommentContinue :: Parser ()
+blockCommentContinue :: Parser Text
 blockCommentContinue = endOfComment <|> continue
   where
-    endOfComment = void (text "-}")
+    endOfComment = void (text "-}") *> pure ""
 
     continue = do
-        blockCommentChunk
-        blockCommentContinue
+        c <- blockCommentChunk
+        c' <- blockCommentContinue
+        pure (c <> c')
 
 simpleLabel :: Bool -> Parser Text
-simpleLabel allowReserved = try (do
+simpleLabel allowReserved = try $ do
     c    <- Text.Parser.Char.satisfy headCharacter
     rest <- Dhall.Parser.Combinators.takeWhile tailCharacter
     let t = Data.Text.cons c rest
-    Control.Monad.guard (allowReserved || not (Data.HashSet.member t reservedIdentifiers))
-    return t )
-  where
+    let isNotAKeyword = not $ t `Data.HashSet.member` reservedKeywords
+    let isNotAReservedIdentifier = not $ t `Data.HashSet.member` reservedIdentifiers
+    Monad.guard (isNotAKeyword && (allowReserved || isNotAReservedIdentifier))
+    return t
 
 headCharacter :: Char -> Bool
 headCharacter c = alpha c || c == '_'
@@ -424,7 +538,7 @@ tailCharacter c = alphaNum c || c == '_' || c == '-' || c == '/'
 backtickLabel :: Parser Text
 backtickLabel = do
     _ <- char '`'
-    t <- takeWhile1 predicate
+    t <- Dhall.Parser.Combinators.takeWhile predicate
     _ <- char '`'
     return t
   where
@@ -438,21 +552,31 @@ backtickLabel = do
 
     This corresponds to the @labels@ rule in the official grammar
 -}
-labels :: Parser (Set Text)
+labels :: Parser [Text]
 labels = do
     _openBrace
+
     whitespace
-    xs <- nonEmptyLabels <|> emptyLabels
-    _closeBrace
-    return xs
+
+    nonEmptyLabels <|> emptyLabels
   where
-    emptyLabels = pure Dhall.Set.empty
+    emptyLabels = do
+        try (optional (_comma *> whitespace) *> _closeBrace)
+
+        pure []
 
     nonEmptyLabels = do
-        x  <- anyLabel
+        x  <- try (optional (_comma *> whitespace) *> anyLabelOrSome)
+
         whitespace
-        xs <- many (do _comma; whitespace; l <- anyLabel; whitespace; return l)
-        noDuplicates (x : xs)
+
+        xs <- many (try (_comma *> whitespace *> anyLabelOrSome) <* whitespace)
+
+        _ <- optional (_comma *> whitespace)
+
+        _closeBrace
+
+        return (x : xs)
 
 {-| Parse a label (e.g. a variable\/field\/alternative name)
 
@@ -472,6 +596,14 @@ anyLabel = (do
     t <- backtickLabel <|> simpleLabel True
     return t ) <?> "any label"
 
+{-| Same as `anyLabel` except that `Some` is allowed
+
+    This corresponds to the @any-label-or-some@ rule in the official grammar
+-}
+
+anyLabelOrSome :: Parser Text
+anyLabelOrSome = try anyLabel <|> ("Some" <$ _Some)
+
 {-| Parse a valid Bash environment variable name
 
     This corresponds to the @bash-environment-variable@ rule in the official
@@ -526,7 +658,7 @@ quotedPathCharacter c =
     ||  ('\x23' <= c && c <= '\x2E')
     ||  ('\x30' <= c && c <= '\x10FFFF')
 
-{-| The `pathComponent` function uses this type to distinguish whether to parse
+{-| The @pathComponent@ function uses this type to distinguish whether to parse
     a URL path component or a file path component
 -}
 data ComponentType = URLComponent | FileComponent
@@ -538,23 +670,20 @@ pathComponent componentType = do
 
     let pathData =
             case componentType of
-                FileComponent -> do
+                FileComponent ->
                     Text.Megaparsec.takeWhile1P Nothing Dhall.Syntax.pathCharacter
-                URLComponent -> do
+                URLComponent ->
                     star pchar
 
     let quotedPathData = do
             _ <- char '"'
             t <- Text.Megaparsec.takeWhile1P Nothing quotedPathCharacter
             _ <- char '"'
+            return t
 
-            case componentType of
-              FileComponent -> do
-                return t
-              URLComponent -> do
-                return (URI.Encode.encodeText t)
-
-    quotedPathData <|> pathData
+    case componentType of
+        FileComponent -> quotedPathData <|> pathData
+        URLComponent -> pathData
 
 -- | Parse a `File`
 file_ :: ComponentType -> Parser File
@@ -719,7 +848,7 @@ pctEncoded :: Parser Text
 pctEncoded = "%" <> count 2 (satisfy hexdig)
 
 subDelims :: Char -> Bool
-subDelims c = c `elem` ("!$&'()*+,;=" :: String)
+subDelims c = c `elem` ("!$&'*+;=" :: String)
 
 unreserved :: Char -> Bool
 unreserved c =
@@ -988,20 +1117,6 @@ _ListIndexed = builtin "List/indexed"
 _ListReverse :: Parser ()
 _ListReverse = builtin "List/reverse"
 
-{-| Parse the @Optional/fold@ built-in
-
-    This corresponds to the @Optional-fold@ rule from the official grammar
--}
-_OptionalFold :: Parser ()
-_OptionalFold = builtin "Optional/fold"
-
-{-| Parse the @Optional/build@ built-in
-
-    This corresponds to the @Optional-build@ rule from the official grammar
--}
-_OptionalBuild :: Parser ()
-_OptionalBuild = builtin "Optional/build"
-
 {-| Parse the @Bool@ built-in
 
     This corresponds to the @Bool@ rule from the official grammar
@@ -1044,6 +1159,13 @@ _Double = builtin "Double"
 _Text :: Parser ()
 _Text = builtin "Text"
 
+{-| Parse the @Text/replace@ built-in
+
+    This corresponds to the @Text-replace@ rule from the official grammar
+-}
+_TextReplace :: Parser ()
+_TextReplace = builtin "Text/replace"
+
 {-| Parse the @Text/show@ built-in
 
     This corresponds to the @Text-show@ rule from the official grammar
@@ -1051,6 +1173,27 @@ _Text = builtin "Text"
 _TextShow :: Parser ()
 _TextShow = builtin "Text/show"
 
+{-| Parse the @Date@ bult-in
+
+    This corresponds to the @Date@ rule from the official grammar
+-}
+_Date :: Parser ()
+_Date = builtin "Date"
+
+{-| Parse the @Time@ bult-in
+
+    This corresponds to the @Time@ rule from the official grammar
+-}
+_Time :: Parser ()
+_Time = builtin "Time"
+
+{-| Parse the @TimeZone@ bult-in
+
+    This corresponds to the @TimeZone@ rule from the official grammar
+-}
+_TimeZone :: Parser ()
+_TimeZone = builtin "TimeZone"
+
 {-| Parse the @List@ built-in
 
     This corresponds to the @List@ rule from the official grammar
@@ -1196,8 +1339,10 @@ _at :: Parser ()
 _at = reservedChar '@' <?> "\"@\""
 
 -- | Parse the equivalence symbol (@===@ or @≡@)
-_equivalent :: Parser ()
-_equivalent = (void (char '≡' <?> "\"≡\"") <|> void (text "===")) <?> "operator"
+_equivalent :: Parser CharacterSet
+_equivalent =
+        (Unicode <$ char '≡' <?> "\"≡\"")
+    <|> (ASCII <$ text "===" <?> "===")
 
 -- | Parse the @missing@ keyword
 _missing :: Parser ()
@@ -1210,32 +1355,40 @@ _importAlt :: Parser ()
 _importAlt = operatorChar '?'
 
 -- | Parse the record combine operator (@/\\@ or @∧@)
-_combine :: Parser ()
-_combine = (void (char '∧' <?> "\"∧\"") <|> void (text "/\\")) <?> "operator"
+_combine :: Parser CharacterSet
+_combine =
+        (Unicode <$ char '∧' <?> "\"∧\"")
+    <|> (ASCII <$ text "/\\" <?> "/\\")
 
 -- | Parse the record type combine operator (@//\\\\@ or @⩓@)
-_combineTypes :: Parser ()
-_combineTypes = (void (char '⩓' <?> "\"⩓\"") <|> void (text "//\\\\")) <?> "operator"
+_combineTypes :: Parser CharacterSet
+_combineTypes =
+        (Unicode <$ char '⩓' <?> "\"⩓\"")
+    <|> (ASCII <$ text "//\\\\" <?> "//\\\\")
 
 -- | Parse the record \"prefer\" operator (@//@ or @⫽@)
-_prefer :: Parser ()
-_prefer = (void (char '⫽' <?> "\"⫽\"") <|> void (text "//")) <?> "operator"
+_prefer :: Parser CharacterSet
+_prefer =
+        (Unicode <$ char '⫽' <?> "\"⫽\"")
+    <|> (ASCII <$ text "//" <?> "//")
 
 -- | Parse a lambda (@\\@ or @λ@)
-_lambda :: Parser ()
-_lambda = void (Text.Parser.Char.satisfy predicate) <?> "\\"
-  where
-    predicate 'λ'  = True
-    predicate '\\' = True
-    predicate _    = False
+_lambda :: Parser CharacterSet
+_lambda =
+        (Unicode <$ char 'λ' <?> "\"λ\"")
+    <|> (ASCII <$ char '\\' <?> "\\")
 
 -- | Parse a forall (@forall@ or @∀@)
-_forall :: Parser ()
-_forall = (void (char '∀' <?> "\"∀\"") <|> void (text "forall")) <?> "forall"
+_forall :: Parser CharacterSet
+_forall =
+        (Unicode <$ char '∀' <?> "\"∀\"")
+    <|> (ASCII <$ text "forall" <?> "forall")
 
 -- | Parse a right arrow (@->@ or @→@)
-_arrow :: Parser ()
-_arrow = (void (char '→' <?> "\"→\"") <|> void (text "->")) <?> "->"
+_arrow :: Parser CharacterSet
+_arrow =
+        (Unicode <$ char '→' <?> "\"→\"")
+    <|> (ASCII <$ text "->" <?> "->")
 
 -- | Parse a double colon (@::@)
 _doubleColon :: Parser ()
