@@ -3,6 +3,10 @@
 {-| Dhall is a programming language specialized for configuration files.  This
     module contains a tutorial explaining how to author configuration files
     using this language
+
+    You might also be interested in the
+    <https://docs.dhall-lang.org/tutorials/Language-Tour.html Language Tour>, a
+    Haskell-agnostic tour of Dhall's language features.
 -}
 module Dhall.Tutorial (
     -- * Introduction
@@ -80,6 +84,9 @@ module Dhall.Tutorial (
     -- * Prelude
     -- $prelude
 
+    -- * Limitations
+    -- $limitations
+
     -- * Conclusion
     -- $conclusion
 
@@ -99,35 +106,37 @@ import Dhall
 -- The simplest way to use Dhall is to ignore the programming language features
 -- and use it as a strongly typed configuration format.  For example, suppose
 -- that you create the following configuration file:
--- 
+--
 -- > -- ./config.dhall
 -- > { foo = 1
 -- > , bar = [3.0, 4.0, 5.0]
 -- > }
--- 
+--
 -- You can read the above configuration file into Haskell using the following
 -- code:
--- 
+--
 -- > -- example.hs
--- > 
+-- >
 -- > {-# LANGUAGE DeriveGeneric     #-}
 -- > {-# LANGUAGE OverloadedStrings #-}
--- > 
+-- >
 -- > import Dhall
--- > 
+-- >
 -- > data Example = Example { foo :: Natural, bar :: Vector Double }
 -- >     deriving (Generic, Show)
--- > 
+-- >
 -- > instance FromDhall Example
--- > 
+-- >
 -- > main :: IO ()
 -- > main = do
 -- >     x <- input auto "./config.dhall"
 -- >     print (x :: Example)
--- 
+--
+-- __WARNING__: You must not instantiate FromDhall with a recursive type. See [Limitations](#limitations).
+--
 -- If you compile and run the above example, the program prints the corresponding
 -- Haskell record:
--- 
+--
 -- > $ ./example
 -- > Example {foo = 1, bar = [3.0,4.0,5.0]}
 --
@@ -139,8 +148,8 @@ import Dhall
 -- >>> input auto "True" :: IO Bool
 -- True
 --
--- The `input` function can decode any value if we specify the value's expected
--- `Type`:
+-- The `Dhall.input` function can decode any value if we specify the value's
+-- expected `Dhall.Type`:
 --
 -- > input
 -- >     :: Type a  -- Expected type
@@ -150,7 +159,7 @@ import Dhall
 -- ... and we can either specify an explicit type like `bool`:
 --
 -- > bool :: Type Bool
--- > 
+-- >
 -- > input bool :: Text -> IO Bool
 -- >
 -- > input bool "True" :: IO Bool
@@ -171,7 +180,7 @@ import Dhall
 -- You can see what types `auto` supports \"out-of-the-box\" by browsing the
 -- instances for the `FromDhall` class.  For example, the following instance
 -- says that we can directly decode any Dhall expression that evaluates to a
--- @Bool@ into a Haskell `Bool`:
+-- @Bool@ into a Haskell `Prelude.Bool`:
 --
 -- > instance FromDhall Bool
 --
@@ -189,8 +198,8 @@ import Dhall
 -- >>> input auto "[True, False]" :: IO (Vector Bool)
 -- [True,False]
 --
--- We could also specify what type to decode by providing an explicit `Type`
--- instead of using `auto` with a type annotation:
+-- We could also specify what type to decode by providing an explicit
+-- `Dhall.Type` instead of using `auto` with a type annotation:
 --
 -- >>> input (vector bool) "[True, False]"
 -- [True,False]
@@ -201,12 +210,12 @@ import Dhall
 -- > {-# LANGUAGE DeriveAnyClass    #-}
 -- > {-# LANGUAGE DeriveGeneric     #-}
 -- > {-# LANGUAGE OverloadedStrings #-}
--- > 
+-- >
 -- > import Dhall
--- > 
+-- >
 -- > data Person = Person { age :: Natural, name :: Text }
 -- >     deriving (Generic, FromDhall, Show)
--- > 
+-- >
 -- > main :: IO ()
 -- > main = do
 -- >     x <- input auto "./config.dhall"
@@ -216,12 +225,12 @@ import Dhall
 -- decode:
 --
 -- > {-# LANGUAGE OverloadedStrings #-}
--- > 
+-- >
 -- > import Data.Functor.Identity
 -- > import Dhall
--- > 
+-- >
 -- > instance FromDhall a => FromDhall (Identity a)
--- > 
+-- >
 -- > main :: IO ()
 -- > main = do
 -- >     x <- input auto "./config.dhall"
@@ -244,7 +253,7 @@ import Dhall
 -- ...
 --
 -- The interpreter complains because the string @\"1\"@ cannot be decoded into a
--- Haskell value of type `Bool`.  This part of the type error:
+-- Haskell value of type `Prelude.Bool`.  This part of the type error:
 --
 -- > - Bool
 -- > + Natural
@@ -307,13 +316,13 @@ import Dhall
 -- expressions by their file path.
 --
 -- To illustrate this, let's create three files:
--- 
+--
 -- > $ echo "True"  > bool1
 -- > $ echo "False" > bool2
 -- > $ echo "./bool1 && ./bool2" > both
 --
 -- ... and read in all three files in a single expression:
--- 
+--
 -- >>> input auto "[ ./bool1 , ./bool2 , ./both ]" :: IO (Vector Bool)
 -- [True,False,False]
 --
@@ -349,7 +358,7 @@ import Dhall
 -- ... then the interpreter will reject the import:
 --
 -- >>> input auto "./file1" :: IO Natural
--- *** Exception: 
+-- *** Exception:
 -- ↳ ./file1
 --   ↳ ./file2
 -- ...
@@ -368,7 +377,7 @@ import Dhall
 --
 -- >>> input auto "https://raw.githubusercontent.com/dhall-lang/dhall-haskell/18e4e9a18dc53271146df3ccf5b4177c3552236b/examples/True" :: IO Bool
 -- True
--- 
+--
 -- ... or inside of a larger expression:
 --
 -- >>> input auto "False == https://raw.githubusercontent.com/dhall-lang/dhall-haskell/18e4e9a18dc53271146df3ccf5b4177c3552236b/examples/True" :: IO Bool
@@ -404,8 +413,8 @@ import Dhall
 --
 -- >>> writeFile "baz" "2.0"
 -- >>> input auto "./baz: Double" :: IO Double
--- *** Exception: 
--- ↳ ./baz: 
+-- *** Exception:
+-- ↳ ./baz:
 -- ...
 -- ...Error...: Missing file ...baz:
 -- ...
@@ -438,15 +447,18 @@ import Dhall
 --
 -- Every list can be followed by the type of the list.  The type annotation is
 -- required for empty lists but optional for non-empty lists.  You will get a
--- type error if you provide an empty list without a type annotation:
+-- parse error if you provide an empty list without a type annotation:
 --
 -- >>> input auto "[]" :: IO (Vector Natural)
 -- *** Exception:
--- ...Error...: An empty list requires a type annotation
+-- ...Error...: Invalid input
 -- ...
--- 1│ []
+-- (input):1:3:
+--   |
+-- 1 | []
+--   |   ^
+-- Empty list literal without annotation
 -- ...
--- (input):1:1
 --
 -- Also, list elements must all have the same type.  You will get an error if
 -- you try to store elements of different types in a list:
@@ -582,7 +594,7 @@ import Dhall
 -- using the @dhall@ command.
 --
 -- You can convert a record to a list of key-value pairs (a.k.a. a \"Map\") by
--- using the `toMap` keyword.  For example:
+-- using the @toMap@ keyword.  For example:
 --
 -- > $ dhall <<< 'toMap { foo = 1, bar = 2 }'
 -- > [ { mapKey = "bar", mapValue = 2 }, { mapKey = "foo", mapValue = 1 } ]
@@ -602,14 +614,17 @@ import Dhall
 -- >     [ n && True, n && False, n || True, n || False ]
 --
 -- ... or we can use Dhall's support for Unicode characters to use @λ@ (U+03BB)
--- instead of @\\@ and @→@ (U+2192) instead of @->@ (for people who are into that
--- sort of thing):
+-- instead of @\\@ and @→@ (U+2192) instead of @->@ (for people who are into
+-- that sort of thing):
 --
 -- > $ -- ./makeBools.dhall
 -- > λ(n : Bool) →
 -- >     [ n && True, n && False, n || True, n || False ]
 -- > <Ctrl-D>
 --
+-- We'll be sticking to ASCII for the remainder of the tutorial, though, while
+-- still pointing out Unicode equivalents as we go.
+--
 -- You can read this as a function of one argument named @n@ that has type
 -- @Bool@.  This function returns a @List@ of @Bool@s.  Each element of the
 -- @List@ depends on the input argument named @n@.
@@ -632,7 +647,7 @@ import Dhall
 -- Thanks to currying, this instance works for functions of multiple simple
 -- arguments:
 --
--- >>> dhallAnd <- input auto "λ(x : Bool) → λ(y : Bool) → x && y" :: IO (Bool -> Bool -> Bool)
+-- >>> dhallAnd <- input auto "\\(x : Bool) -> \\(y : Bool) -> x && y" :: IO (Bool -> Bool -> Bool)
 -- >>> dhallAnd True False
 -- False
 --
@@ -645,16 +660,16 @@ import Dhall
 -- > {-# LANGUAGE DeriveAnyClass    #-}
 -- > {-# LANGUAGE DeriveGeneric     #-}
 -- > {-# LANGUAGE OverloadedStrings #-}
--- > 
+-- >
 -- > module Main where
--- > 
+-- >
 -- > import Dhall
--- > 
+-- >
 -- > data Example0 = Example0 { foo :: Bool, bar :: Bool }
 -- >     deriving (Generic, ToDhall)
--- > 
+-- >
 -- > main = do
--- >     f <- input auto "λ(r : { foo : Bool, bar : Bool }) → r.foo && r.bar"
+-- >     f <- input auto "\\(r : { foo : Bool, bar : Bool }) -> r.foo && r.bar"
 -- >     print (f (Example0 { foo = True, bar = False }) :: Bool)
 -- >     print (f (Example0 { foo = True, bar = True  }) :: Bool)
 --
@@ -674,6 +689,12 @@ import Dhall
 -- > $ dhall --annotate <<< './makeBools.dhall'
 -- > (λ(n : Bool) → [ n, False, True, n ]) : ∀(n : Bool) → List Bool
 --
+-- The @dhall@ command emits Unicode output by default, but you can switch to
+-- ASCII output using the @--ascii@ flag:
+--
+-- > $ dhall --annotate --ascii <<< './makeBools.dhall'
+-- > (\(n : Bool) -> [ n, False, True, n ]) : forall (n : Bool) -> List Bool
+--
 -- The @--annotate@ flag adds a type signature to the output to let us know
 -- what type the interpreter inferred for our expression.  The type signature
 -- is @∀(n : Bool) → List Bool@ which says that @makeBools@ is a function of one
@@ -681,16 +702,16 @@ import Dhall
 -- @Bool@s.  The @∀@ (U+2200) symbol is shorthand for the ASCII @forall@
 -- keyword:
 --
--- > ∀(x : a) → b        -- This type ...
--- > 
--- > forall (x : a) → b  -- ... is the same as this type
+-- > ∀(x : a) → b         -- This type ...
+-- >
+-- > forall (x : a) -> b  -- ... is the same as this type
 --
 -- ... and Dhall's @forall@ keyword behaves the same way as Haskell's @forall@
 -- keyword for input values that are @Type@s:
 --
--- > forall (x : Type) → b  -- This Dhall type ...
--- 
--- > forall x . b           -- ... is the same as this Haskell type
+-- > forall (x : Type) -> b  -- This Dhall type ...
+--
+-- > forall x . b            -- ... is the same as this Haskell type
 --
 -- The part where Dhall differs from Haskell is that you can also use
 -- @∀@/@forall@ to give names to non-@Type@ arguments (such as the first
@@ -721,8 +742,8 @@ import Dhall
 --
 -- Remember that file paths are synonymous with their contents, so the above
 -- code is exactly equivalent to:
--- 
--- > $ dhall <<< '(λ(n : Bool) → [n && True, n && False, n || True, n || False]) True'
+--
+-- > $ dhall <<< '(\(n : Bool) -> [n && True, n && False, n || True, n || False]) True'
 -- > [True, False, True, True]
 --
 -- When you apply an anonymous function to an argument, you substitute the
@@ -730,9 +751,9 @@ import Dhall
 --
 -- >    Bound variable
 -- >    ⇩
--- > (λ(n : Bool) → ...) True
--- >                     ⇧
--- >                     Function argument
+-- > (\(n : Bool) -> ...) True
+-- >                      ⇧
+-- >                      Function argument
 --
 -- So in our above example, we would replace all occurrences of @n@ with @True@,
 -- like this:
@@ -755,7 +776,7 @@ import Dhall
 -- __Exercise__: Create a file named @getFoo@ that is a function of the following
 -- type:
 --
--- > ∀(r : { foo : Bool, bar : Text }) → Bool
+-- > forall (r : { foo : Bool, bar : Text }) -> Bool
 --
 -- This function should take a single input argument named @r@ that is a record
 -- with two fields.  The function should return the value of the @foo@ field.
@@ -779,7 +800,7 @@ import Dhall
 -- > $ dhall
 -- > ''
 -- >     #!/bin/bash
--- >     
+-- >
 -- >     echo "Hi!"
 -- > ''
 -- > <Ctrl-D>
@@ -866,14 +887,14 @@ import Dhall
 -- ... but fails with a type error if either shared field is not a record:
 --
 -- > $ dhall
--- > { foo = 1, bar = "ABC" } ∧ { foo = True }
+-- > { foo = 1, bar = "ABC" } /\ { foo = True }
 -- > <Ctrl-D>
 -- > Use "dhall --explain" for detailed errors
--- > 
+-- >
 -- > Error: Field collision
--- > 
+-- >
 -- > { foo = 1, bar = "ABC" } ∧ { foo = True }
--- > 
+-- >
 -- > (input):1:1
 --
 -- __Exercise__: Combine any record with the empty record.  What do you expect
@@ -882,7 +903,7 @@ import Dhall
 -- You can analogously combine record types using the @//\\\\@ operator (or @(⩓)@ U+2A53):
 --
 -- > $ dhall
--- > { foo : Natural } ⩓ { bar : Text }
+-- > { foo : Natural } //\\ { bar : Text }
 -- > <Ctrl-D>
 -- > { foo : Natural, bar : Text }
 --
@@ -890,7 +911,7 @@ import Dhall
 -- operator descends recursively into record types:
 --
 -- > $ dhall
--- > { foo : { bar : Text } } ⩓ { foo : { baz : Bool }, qux : Natural }
+-- > { foo : { bar : Text } } //\\ { foo : { baz : Bool }, qux : Natural }
 -- > <Ctrl-D>
 -- > { foo : { bar : Text, baz : Bool }, qux : Natural }
 
@@ -929,7 +950,7 @@ import Dhall
 -- function:
 --
 -- > $ dhall
--- > let twice = λ(x : Text) → x ++ x in twice "ha"
+-- > let twice = \(x : Text) -> x ++ x in twice "ha"
 -- > <Ctrl-D>
 -- > "haha"
 --
@@ -940,7 +961,7 @@ import Dhall
 -- > let twice (x : Text) = x ++ x in twice "ha"
 -- > <Ctrl-D>
 -- > Error: Invalid input
--- > 
+-- >
 -- > (input):1:11:
 -- >   |
 -- > 1 | let twice (x : Text) = x ++ x in twice "ha"
@@ -964,7 +985,7 @@ import Dhall
 --
 -- > $ dhall <<< 'let Name : Type = Text in [ "John", "Mary" ] : List Name'
 -- > List Text
--- > 
+-- >
 -- > [ "John", "Mary" ]
 --
 -- __Exercise:__ What do you think the following code will normalize to?
@@ -994,12 +1015,12 @@ import Dhall
 -- > let greet =
 -- >           \(args : { greeting : Text, name : Text })
 -- >       ->  "${args.greeting}, ${args.name}!"
--- > 
+-- >
 -- > let Greeting =
 -- >       { Type = { greeting : Text, name : Text }
 -- >       , default = { greeting = "Hello", name = "John" }
 -- >       }
--- > 
+-- >
 -- > in  ''
 -- >     ${greet Greeting::{=}}
 -- >     ${greet Greeting::{ greeting = "Hola" }}
@@ -1050,7 +1071,7 @@ import Dhall
 -- You can specify the value of a union constructor like this:
 --
 -- > let Union = < Left : Natural | Right : Bool>
--- > 
+-- >
 -- > in  [ Union.Left 0, Union.Right True ]
 --
 -- In other words, you can access a union constructor as a field of a union
@@ -1080,10 +1101,10 @@ import Dhall
 -- would write:
 --
 -- > $ cat > process <<EOF
--- >     λ(union : < Left : Natural | Right : Bool >)
--- > →   let handlers =
+-- >   \(union : < Left : Natural | Right : Bool >) ->
+-- >     let handlers =
 -- >             { Left  = Natural/even  -- Natural/even is a built-in function
--- >             , Right = λ(b : Bool) → b
+-- >             , Right = \(b : Bool) -> b
 -- >             }
 -- > in  merge handlers union
 -- > EOF
@@ -1100,7 +1121,7 @@ import Dhall
 --
 -- > merge handlers union [ : type ]
 --
--- ... where: 
+-- ... where:
 --
 -- * @union@ is the union you want to consume
 -- * @handlers@ is a record with one function per alternative of the union
@@ -1113,14 +1134,14 @@ import Dhall
 --
 -- So, for example:
 --
--- > merge { Left = Natural/even, Right = λ(b : Bool) → b } (< Left : Natural | Right : Bool >.Left 3)
+-- > merge { Left = Natural/even, Right = \(b : Bool) -> b } (< Left : Natural | Right : Bool >.Left 3)
 -- >     = Natural/even 3
 -- >     = False
 --
 -- ... and similarly:
 --
--- > merge { Left = Natural/even, Right = λ(b : Bool) → b } (< Left : Natural | Right : Bool >.Right True)
--- >     = (λ(b : Bool) → b) True
+-- > merge { Left = Natural/even, Right = \(b : Bool) -> b } (< Left : Natural | Right : Bool >.Right True)
+-- >     = (\(b : Bool) -> b) True
 -- >     = True
 --
 -- Notice that each handler has to return the same type of result (@Bool@ in
@@ -1144,7 +1165,7 @@ import Dhall
 -- Empty alternatives like @Empty@ require no argument:
 --
 -- > let Example = < Empty | Person : { name : Text, age : Natural } >
--- > 
+-- >
 -- > in  [ Example.Empty  -- Note the absence of any argument to `Empty`
 -- >     , Example.Person { name = "John", age = 23 }
 -- >     , Example.Person { name = "Amy" , age = 25 }
@@ -1153,13 +1174,13 @@ import Dhall
 -- ... and when you @merge@ an empty alternative the correspond handler takes no
 -- argument:
 --
--- >     λ(x : < Empty | Person : { name : Text, age : Natural } >)
--- > →   merge
--- >     {   Empty = "Unknown"  -- Note the absence of a `λ`
+-- > \(x : < Empty | Person : { name : Text, age : Natural } >) ->
+-- >   merge
+-- >     { Empty = "Unknown"  -- Note the absence of an anonymous function
 -- >
--- >     ,   Person =
--- >             λ(person : { name : Text, age : Natural })
--- >         →   "Name: ${person.name}, Age: ${Natural/show person.age}"
+-- >     , Person =
+-- >         \(person : { name : Text, age : Natural }) ->
+-- >           "Name: ${person.name}, Age: ${Natural/show person.age}"
 -- >     }
 -- >     x
 --
@@ -1191,18 +1212,18 @@ import Dhall
 --
 -- The equivalent function in Dhall is:
 --
--- > λ(a : Type) → λ(x : a) → x
+-- > \(a : Type) -> \(x : a) -> x
 --
 -- Notice how this function takes two arguments instead of one.  The first
 -- argument is the type of the second argument.
 --
 -- Let's illustrate how this works by actually using the above function:
 --
--- > $ echo "λ(a : Type) → λ(x : a) → x" > id
+-- > $ echo "\\(a : Type) -> \\(x : a) -> x" > id
 --
 -- Let's ask the interpreter for the type of this function:
 -- the first line:
--- 
+--
 -- > $ dhall type <<< './id'
 -- > ∀(a : Type) → ∀(x : a) → a
 --
@@ -1268,7 +1289,7 @@ import Dhall
 -- argument type then that means that the name of the argument is @"_"@.  This
 -- is true even for user-defined functions:
 --
--- > $ dhall type <<< 'λ(_ : Text) → 1'
+-- > $ dhall type <<< '\\(_ : Text) -> 1'
 -- > Text → Natural
 --
 -- The type @Text → Natural@ is the same as @∀(_ : Text) → Natural@
@@ -1301,8 +1322,8 @@ import Dhall
 -- complex example:
 --
 -- > $ dhall
--- >     let List/map = https://prelude.dhall-lang.org/List/map
--- > in  λ(f : Natural → Natural) → List/map Natural Natural f [1, 2, 3]
+-- > let List/map = https://prelude.dhall-lang.org/List/map
+-- > in  \(f : Natural → Natural) -> List/map Natural Natural f [1, 2, 3]
 -- > <Ctrl-D>
 -- > λ(f : Natural → Natural) → [f 1, f 2, f 3]
 --
@@ -1336,13 +1357,13 @@ import Dhall
 -- code.  For example, we can add some tests to our @not@ function like this:
 --
 -- > let not
--- >     : Bool → Bool
--- >     = λ(b : Bool) → b == False
--- > 
+-- >     : Bool -> Bool
+-- >     = \(b : Bool) -> b == False
+-- >
 -- > let example0 = assert : not False === True
--- > 
+-- >
 -- > let example1 = assert : not True === False
--- > 
+-- >
 -- > in  not
 --
 -- The expression @assert : not False == True@ is a type-checking assertion
@@ -1354,8 +1375,8 @@ import Dhall
 -- > -- ./test.dhall
 -- >
 -- > let not
--- >     : Bool → Bool
--- >     = λ(b : Bool) → b == False
+-- >     : Bool -> Bool
+-- >     = \(b : Bool) -> b == False
 -- >
 -- > let example0 = assert : not False === True
 -- >
@@ -1365,22 +1386,22 @@ import Dhall
 --
 -- The type-checker then rejects the assertion with the following error message:
 --
--- > $ dhall <<< './test.dhall' 
--- > 
+-- > $ dhall <<< './test.dhall'
+-- >
 -- > ↳ ./test.dhall
--- > 
+-- >
 -- > Error: Assertion failed
--- > 
+-- >
 -- > - False
 -- > + True
--- > 
+-- >
 -- > 7│                assert : not True === True -- Oops!
--- > 8│ 
--- > 
+-- > 8│
+-- >
 -- > ./test.dhall:7:16
--- > 
+-- >
 -- > 1│ ./test.dhall
--- > 
+-- >
 -- > (input):1:1
 --
 -- You can compare expressions that contain variables, too, which is equivalent
@@ -1398,18 +1419,18 @@ import Dhall
 -- if they don't share the same normal form, such as these:
 --
 -- > $ dhall <<< '\(n : Natural) -> assert : Natural/even (n + n) === True'
--- > 
+-- >
 -- > Use "dhall --explain" for detailed errors
--- > 
+-- >
 -- > n : Natural
--- > 
+-- >
 -- > Error: Assertion failed
--- > 
+-- >
 -- > - … …
 -- > + True
--- > 
+-- >
 -- > 1│                   assert : Natural/even (n + n) === True
--- > 
+-- >
 -- > (input):1:19
 --
 -- Here the interpreter is not smart enough to simplify @Natural/even (n + n)@
@@ -1456,9 +1477,9 @@ import Dhall
 -- Dhall will forward imports if you import an expression from a URL that
 -- contains a relative import.  For example, if you import an expression like
 -- this:
--- 
+--
 -- > http://example.com/example.dhall using ./headers
--- 
+--
 -- ... and @http:\/\/example.com/example.dhall@ contains a relative import of @./foo@
 -- then Dhall will import @http:\/\/example.com/foo@ using the same @./headers@ file.
 
@@ -1492,7 +1513,7 @@ import Dhall
 --
 -- > $ dhall <<< './foo'
 -- > Natural
--- > 
+-- >
 -- > 1
 --
 -- Any import protected by a semantic integrity check is automatically cached
@@ -1516,12 +1537,12 @@ import Dhall
 -- From Dhall's point of view, the hash is the true address and the file path
 -- is just a suggestion for how to obtain the import if it's not already cached.
 --
--- You can disable the cache by setting `XDG_CACHE_HOME` to `/var/empty` (an
+-- You can disable the cache by setting @XDG_CACHE_HOME@ to `/var/empty` (an
 -- empty and unwritable directory), like this:
 --
 -- > $ XDG_CACHE_HOME=/var/empty dhall <<< './foo'
 -- > Natural
--- > 
+-- >
 -- > 1
 --
 -- We'll use this trick to test changes to the protected import in the following
@@ -1539,10 +1560,10 @@ import Dhall
 --
 -- > $ XDG_CACHE_HOME=/var/empty dhall <<< './foo'  # This still succeeds
 -- > Natural
--- > 
+-- >
 -- > 1
 --
--- You can compute the Hash for any import by using the hash subcommand 
+-- You can compute the Hash for any import by using the hash subcommand
 -- of this package.  For example:
 --
 -- > $ dhall hash <<< './bar'
@@ -1557,15 +1578,15 @@ import Dhall
 -- text of the @./bar@ file technically never changed:
 --
 -- > XDG_CACHE_HOME=/var/empty dhall <<< './foo'
--- > 
+-- >
 -- > Error: Import integrity check failed
--- > 
+-- >
 -- > Expected hash:
--- > 
+-- >
 -- > ↳ d60d8415e36e86dae7f42933d3b0c4fe3ca238f057fba206c7e9fbf5d784fe15
--- > 
+-- >
 -- > Actual hash:
--- > 
+-- >
 -- > ↳ 4caf97e8c445d4d4b5c5b992973e098ed4ae88a355915f5a59db640a589bc9cb
 --
 -- This is because the @./bar@ file now represents a new value (@2@ instead of
@@ -1598,18 +1619,18 @@ import Dhall
 -- > let replicate = https://prelude.dhall-lang.org/List/replicate
 -- > in  replicate 5
 -- > $
--- > $ dhall freeze --inplace ./foo.dhall
+-- > $ dhall freeze ./foo.dhall
 -- > $ cat ./foo.dhall
 -- > let replicate =
 -- >       https://prelude.dhall-lang.org/List/replicate sha256:d4250b45278f2d692302489ac3e78280acb238d27541c837ce46911ff3baa347
--- > 
+-- >
 -- > in  replicate 5
 
 -- $rawText
 --
 -- Sometimes you want to import the contents of a raw text file as a Dhall
--- value of type `Text`.  For example, one of the fields of a record might be
--- the contents of a software license:
+-- value of type `Dhall.Syntax.Text`.  For example, one of the fields of a
+-- record might be the contents of a software license:
 --
 -- > { package = "dhall"
 -- > , author  = "Gabriel Gonzalez"
@@ -1623,7 +1644,7 @@ import Dhall
 -- > ''
 -- > Copyright (c) 2017 Gabriel Gonzalez
 -- > All rights reserved.
--- > 
+-- >
 -- > ...
 -- > (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 -- > SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
@@ -1645,57 +1666,57 @@ import Dhall
 -- > $ cat LICENSE
 -- > Copyright (c) 2017 Gabriel Gonzalez
 -- > All rights reserved.
--- > 
+-- >
 -- > ...
 -- > (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 -- > SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 -- $format
 --
--- A format subcommand is also available that you can use to 
+-- A format subcommand is also available that you can use to
 -- automatically format Dhall expressions.  For example, we can take the
 -- following unformatted Dhall expression:
 --
 -- > $ cat ./unformatted
--- > λ(a : Type) → λ(kvss : List (List { index : Natural, value : a })) → 
--- > List/build { index : Natural, value : a } (λ(list : Type) → λ(cons : { 
--- > index : Natural, value : a } → list → list) → λ(nil : list) → 
--- > (List/fold (List { index : Natural, value : a }) kvss { count : Natural, diff : 
--- > Natural → list } (λ(kvs : List { index : Natural, value : a }) → λ(y : { 
--- > count : Natural, diff : Natural → list }) → { count = y.count + List/length 
--- > { index : Natural, value : a } kvs, diff = λ(n : Natural) → List/fold { 
--- > index : Natural, value : a } kvs list (λ(kvOld : { index : Natural, value : a 
--- > }) → λ(z : list) → cons { index = kvOld.index + n, value = kvOld.value } 
--- > z) (y.diff (n + List/length { index : Natural, value : a } kvs)) }) { count = 
+-- > λ(a : Type) → λ(kvss : List (List { index : Natural, value : a })) →
+-- > List/build { index : Natural, value : a } (λ(list : Type) → λ(cons : {
+-- > index : Natural, value : a } → list → list) → λ(nil : list) →
+-- > (List/fold (List { index : Natural, value : a }) kvss { count : Natural, diff :
+-- > Natural → list } (λ(kvs : List { index : Natural, value : a }) → λ(y : {
+-- > count : Natural, diff : Natural → list }) → { count = y.count + List/length
+-- > { index : Natural, value : a } kvs, diff = λ(n : Natural) → List/fold {
+-- > index : Natural, value : a } kvs list (λ(kvOld : { index : Natural, value : a
+-- > }) → λ(z : list) → cons { index = kvOld.index + n, value = kvOld.value }
+-- > z) (y.diff (n + List/length { index : Natural, value : a } kvs)) }) { count =
 -- > 0, diff = λ(_ : Natural) → nil }).diff 0)
 --
 -- ... and run the expression through the formatter:
 --
 -- > $ dhall format < ./unformatted
--- >   λ(a : Type)
--- > → λ(kvss : List (List { index : Natural, value : a }))
--- > → List/build
+-- > λ(a : Type) →
+-- > λ(kvss : List (List { index : Natural, value : a })) →
+-- >   List/build
 -- >     { index : Natural, value : a }
--- >     (   λ(list : Type)
--- >       → λ(cons : { index : Natural, value : a } → list → list)
--- >       → λ(nil : list)
--- >       → ( List/fold
+-- >     ( λ(list : Type) →
+-- >       λ(cons : { index : Natural, value : a } → list → list) →
+-- >       λ(nil : list) →
+-- >         ( List/fold
 -- >             (List { index : Natural, value : a })
 -- >             kvss
 -- >             { count : Natural, diff : Natural → list }
--- >             (   λ(kvs : List { index : Natural, value : a })
--- >               → λ(y : { count : Natural, diff : Natural → list })
--- >               → { count =
+-- >             ( λ(kvs : List { index : Natural, value : a }) →
+-- >               λ(y : { count : Natural, diff : Natural → list }) →
+-- >                 { count =
 -- >                     y.count + List/length { index : Natural, value : a } kvs
 -- >                 , diff =
--- >                       λ(n : Natural)
--- >                     → List/fold
+-- >                     λ(n : Natural) →
+-- >                       List/fold
 -- >                         { index : Natural, value : a }
 -- >                         kvs
 -- >                         list
--- >                         (   λ(kvOld : { index : Natural, value : a })
--- >                           → λ(z : list)
--- >                           → cons
+-- >                         ( λ(kvOld : { index : Natural, value : a }) →
+-- >                           λ(z : list) →
+-- >                             cons
 -- >                               { index = kvOld.index + n, value = kvOld.value }
 -- >                               z
 -- >                         )
@@ -1713,20 +1734,19 @@ import Dhall
 -- normalizing them:
 --
 -- > $ dhall format
--- > let replicate = https://prelude.dhall-lang.org/List/replicate 
+-- > let replicate = https://prelude.dhall-lang.org/List/replicate
 -- > in replicate 5 (List (List Natural)) (replicate 5 (List Natural) (replicate 5 Natural 1))
 -- > <Ctrl-D>
 -- > let replicate = https://prelude.dhall-lang.org/List/replicate
--- > 
+-- >
 -- > in  replicate
 -- >       5
 -- >       (List (List Natural))
 -- >       (replicate 5 (List Natural) (replicate 5 Natural 1))
 --
--- You can also use the formatter to modify files in place using the
--- @--inplace@ flag (i.e. for formatting source code):
+-- You can also use the formatter to modify files in place:
 --
--- > $ dhall format --inplace ./unformatted
+-- > $ dhall format ./unformatted
 --
 -- Carefully note that the code formatter does not preserve all comments.
 -- Currently, the formatter only preserves two types of comments:
@@ -1752,7 +1772,7 @@ import Dhall
 -- >     x {- This comment will also be preserved-} =
 -- >     {- ... and this one will be preserved, too -}
 -- >       1
--- > 
+-- >
 -- > in  x
 --
 -- Note that you do not need to format the output of the
@@ -1760,7 +1780,7 @@ import Dhall
 -- multi-line expressions, too:
 --
 -- > $ dhall
--- > let replicate = https://prelude.dhall-lang.org/List/replicate 
+-- > let replicate = https://prelude.dhall-lang.org/List/replicate
 -- > in replicate 5 (List (List Natural)) (replicate 5 (List Natural) (replicate 5 Natural 1))
 -- > <Ctrl-D>
 -- > [ [ [ 1, 1, 1, 1, 1 ]
@@ -1844,11 +1864,11 @@ import Dhall
 -- > +2 + +2
 -- > <Ctrl-D>
 -- > Use "dhall --explain" for detailed errors
--- > 
+-- >
 -- > Error: ❰+❱ only works on ❰Natural❱s
--- > 
+-- >
 -- > +2 + +2
--- > 
+-- >
 -- > (input):1:1
 --
 -- In fact, there are no built-in functions for @Integer@s (or @Double@s) other
@@ -1868,52 +1888,52 @@ import Dhall
 -- new built-in functions.  This section contains a simple Haskell recipe to add
 -- a new @Natural/equal@ built-in function of type:
 --
--- > Natural/equal : Natural → Natural → Bool
+-- > Natural/equal : Natural -> Natural -> Bool
 --
 -- To do so, we:
--- 
+--
 -- * extend the type-checking context to include the type of @Natural/equal@
 -- * extend the normalizer to evaluate all occurrences of @Natural/equal@
 --
 -- ... like this:
 --
 -- > -- example.hs
--- > 
+-- >
 -- > {-# LANGUAGE OverloadedStrings #-}
--- > 
+-- >
 -- > module Main where
--- > 
+-- >
 -- > import Dhall.Core (Expr(..), ReifiedNormalizer(..))
--- > 
+-- >
 -- > import qualified Data.Text.IO
 -- > import qualified Dhall
 -- > import qualified Dhall.Context
 -- > import qualified Lens.Family   as Lens
--- > 
+-- >
 -- > main :: IO ()
 -- > main = do
 -- >     text <- Data.Text.IO.getContents
--- > 
+-- >
 -- >     let startingContext = transform Dhall.Context.empty
 -- >           where
 -- >             transform = Dhall.Context.insert "Natural/equal" naturalEqualType
--- > 
+-- >
 -- >             naturalEqualType =
 -- >                 Pi "_" Natural (Pi "_" Natural Bool)
--- > 
+-- >
 -- >     let normalizer (App (App (Var "Natural/equal") (NaturalLit x)) (NaturalLit y)) =
 -- >             Just (BoolLit (x == y))
 -- >         normalizer _ =
 -- >             Nothing
--- > 
+-- >
 -- >     let inputSettings = transform Dhall.defaultInputSettings
 -- >           where
 -- >             transform =
 -- >                   Lens.set Dhall.normalizer      (Just (ReifiedNormalizer (pure . normalizer)))
 -- >                 . Lens.set Dhall.startingContext startingContext
--- > 
+-- >
 -- >     x <- Dhall.inputWithSettings inputSettings Dhall.auto text
--- > 
+-- >
 -- >     Data.Text.IO.putStrLn x
 --
 -- Here is an example use of the above program:
@@ -1925,22 +1945,22 @@ import Dhall
 -- expressions containing unexpected free variable such as @Natural/equal@:
 --
 -- > $ dhall <<< 'Natural/equal 2 (1 + 1)'
--- > 
+-- >
 -- > Use "dhall --explain" for detailed errors
--- > 
+-- >
 -- > Error: Unbound variable
--- > 
--- > Natural/equal 
--- > 
+-- >
+-- > Natural/equal
+-- >
 -- > (input):1:1
 --
 -- You will need to either:
--- 
+--
 -- * create your own parallel versions of these tools, or:
 -- * < https://github.com/dhall-lang/dhall-lang/blob/master/.github/CONTRIBUTING.md#how-do-i-change-the-language try to upstream your built-ins into the language>
--- 
+--
 -- The general guidelines for adding new built-ins to the language are:
--- 
+--
 -- * Keep built-ins easy to implement across language bindings
 -- * Prefer general purpose built-ins or built-ins appropriate for the task of program configuration
 -- * Design built-ins to catch errors as early as possible (i.e. when type-checking the configuration)
@@ -1959,8 +1979,8 @@ import Dhall
 --
 -- > -- example.dhall
 -- >
--- > let Result = < Failure : Integer | Success Text >
--- > in Result.Failure 1
+-- > let Result = < Failure : Integer | Success : Text >
+-- > in Result.Failure +1
 --
 -- Right now it is quite easy to keep these two definitions (the one in Haskell source and the one in the Dhall file) synchronized:
 -- If we implement a new feature in the Haskell source we update the corresponding type in the Dhall file.
@@ -1971,14 +1991,14 @@ import Dhall
 -- > resultDecoder = Dhall.auto
 -- >
 -- > resultType :: Expr Src Void
--- > resultType = Dhall.expected resultDecoder
+-- > resultType = maximum $ Dhall.expected resultDecoder
 -- >
 -- > resultTypeString :: String
 -- > resultTypeString = show $ pretty resultType
 --
 -- Now we just have to inject that type into the Dhall code and we are done. One common way to do that is to wrap the import of example.dhall in a let expression:
 --
--- > Dhall.input (Dhall.auto :: Dhall.Decoder Result) ("let Result = " ++ Data.Text.pack resultTypeString ++ " in ./example.dhall")
+-- > Dhall.input (Dhall.auto :: Dhall.Decoder Result) ("let Result = " <> Data.Text.pack resultTypeString <> " in ./example.dhall")
 --
 -- Now we can omit the definition of Result in our example.dhall file. While this will work perfectly Dhall provides a cleaner solution for our \"injection problem\":
 --
@@ -2018,17 +2038,17 @@ import Dhall
 -- The @not@ function is just a UTF8-encoded text file hosted online with the
 -- following contents
 --
--- > $ $ dhall <<< 'https://prelude.dhall-lang.org/Bool/not as Text'
+-- > $ dhall <<< 'https://prelude.dhall-lang.org/Bool/not as Text'
 -- > ''
 -- > {-
--- > Flip the value of a `Bool`
+-- > Flip the value of a Bool
 -- > -}
 -- > let not : Bool → Bool = λ(b : Bool) → b == False
--- > 
+-- >
 -- > let example0 = assert : not True ≡ False
--- > 
+-- >
 -- > let example1 = assert : not False ≡ True
--- > 
+-- >
 -- > in  not
 -- > ''
 --
@@ -2073,11 +2093,11 @@ import Dhall
 -- > Returns `True` if a number if even and returns `False` otherwise
 -- > -}
 -- > let even : Natural → Bool = Natural/even
--- > 
+-- >
 -- > let example0 = assert : even 3 ≡ False
--- > 
+-- >
 -- > let example1 = assert : even 0 ≡ True
--- > 
+-- >
 -- > in  even
 -- > ''
 --
@@ -2225,10 +2245,10 @@ import Dhall
 -- convenience:
 --
 -- > $ dhall
--- >     let Prelude = https://prelude.dhall-lang.org/package.dhall
+-- > let Prelude = https://prelude.dhall-lang.org/package.dhall
 -- >
--- > in    λ(x : Text)
--- >     → Prelude.List.length Text (Prelude.List.replicate 10 Text x)
+-- > in  \(x : Text) ->
+-- >       Prelude.List.length Text (Prelude.List.replicate 10 Text x)
 -- > <Ctrl-D>
 -- > ∀(x : Text) → Natural
 -- >
@@ -2242,6 +2262,35 @@ import Dhall
 --
 -- > $ dhall <<< 'https://prelude.dhall-lang.org/package.dhall'
 
+-- $limitations
+-- #limitations#
+--
+-- `FromDhall` has a limitiation: It won't work for recursive types.
+-- If you instantiate these using their Generic instance you end up with one of
+-- the following two cases:
+--
+-- If the type appears in it's own definition like
+--
+-- > data Foo = Foo Foo
+-- >     deriving Generic
+-- >
+-- > instance FromDhall Foo
+--
+-- the resulting Decoder will throw an exception /ON USAGE/.
+-- However, if the recursion is indirect like
+--
+-- > data Foo = Foo Bar
+-- >     deriving Generic
+-- >
+-- > instance FromDhall Foo
+-- >
+-- > data Bar = Bar Foo
+-- >     deriving Generic
+-- >
+-- > instance FromDhall Bar
+--
+-- the resulting Decoder will __NOT TERMINATE__ /ON USAGE/.
+
 -- $conclusion
 --
 -- By this point you should be able to use the Dhall configuration language to
@@ -2250,6 +2299,11 @@ import Dhall
 --
 -- <https://github.com/dhall-lang/dhall-haskell/issues>
 --
+-- Also, you might want to check out the
+-- <https://docs.dhall-lang.org/tutorials/Language-Tour.html Language Tour>, a
+-- more comprehensive tour of Dhall which may cover features missing from this
+-- Haskell tutorial.
+--
 -- You can also request features, support, or documentation improvements on the
 -- above issue tracker.
 --
