@@ -1,5 +1,7 @@
+{-# LANGUAGE NamedFieldPuns    #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards   #-}
+{-# LANGUAGE TypeApplications  #-}
 
 -- | Shared utility functions
 
@@ -10,6 +12,7 @@ module Dhall.Util
     , _ERROR
     , Censor(..)
     , Input(..)
+    , Transitivity(..)
     , OutputMode(..)
     , Output(..)
     , getExpression
@@ -17,27 +20,35 @@ module Dhall.Util
     , getExpressionAndHeaderFromStdinText
     , Header(..)
     , CheckFailed(..)
+    , MultipleCheckFailed(..)
+    , handleMultipleChecksFailed
+    , renderExpression
     ) where
 
-import Control.Exception (Exception(..))
-import Control.Monad.IO.Class (MonadIO(..))
-import Data.Bifunctor (first)
-import Data.Monoid ((<>))
-import Data.String (IsString)
-import Data.Text (Text)
-import Data.Text.Prettyprint.Doc (Doc, Pretty)
-import Dhall.Parser (ParseError, Header(..))
-import Dhall.Pretty (Ann)
-import Dhall.Syntax (Expr, Import)
-import Dhall.Src (Src)
+import Control.Exception      (Exception (..))
+import Control.Monad.IO.Class (MonadIO (..))
+import Data.Bifunctor         (first)
+import Data.Either            (lefts)
+import Data.Foldable          (toList)
+import Data.List.NonEmpty     (NonEmpty (..))
+import Data.String            (IsString)
+import Data.Text              (Text)
+import Dhall.Parser           (Header (..), ParseError)
+import Dhall.Pretty           (Ann, CharacterSet)
+import Dhall.Src              (Src)
+import Dhall.Syntax           (Expr, Import)
+import Prettyprinter          (Doc, Pretty)
 
 import qualified Control.Exception
 import qualified Data.Text
 import qualified Data.Text.IO
-import qualified Data.Text.Prettyprint.Doc                 as Pretty
-import qualified Data.Text.Prettyprint.Doc.Render.Terminal as Pretty.Terminal
 import qualified Dhall.Parser
 import qualified Dhall.Pretty
+import qualified Prettyprinter                 as Pretty
+import qualified Prettyprinter.Render.Terminal as Pretty.Terminal
+import qualified Prettyprinter.Render.Text     as Pretty.Text
+import qualified System.Console.ANSI           as ANSI
+import qualified System.IO                     as IO
 
 -- | Utility function to cut out the interior of a large text block
 snip :: Text -> Text
@@ -107,17 +118,17 @@ get
     -> InputOrTextFromStdin
     -> IO a
 get parser censor input = do
-    inText <- do
+    inText <-
         case input of
             Input_ (InputFile file) -> Data.Text.IO.readFile file
             Input_ StandardInput    -> Data.Text.IO.getContents
-            StdinText text          -> pure text
+            StdinText _ text        -> pure text
 
     let name =
             case input of
                 Input_ (InputFile file) -> file
                 Input_ StandardInput    -> "(input)"
-                StdinText _             -> "(input)"
+                StdinText inputName _   -> inputName
 
     let result = parser name inText
 
@@ -139,10 +150,23 @@ throws (Right r) = return r
 data Censor = NoCensor | Censor
 
 -- | Path to input
-data Input = StandardInput | InputFile FilePath
+data Input = StandardInput | InputFile FilePath deriving (Eq)
 
 -- | Path to input or raw input text, necessary since we can't read STDIN twice
-data InputOrTextFromStdin = Input_ Input | StdinText Text
+data InputOrTextFromStdin
+    = Input_ Input
+    | StdinText String Text
+    -- ^ @StdinText name text@ where name is a user-friendly name describing the
+    -- input expression, used in parsing error messages
+
+{-| Specifies whether or not an input's transitive dependencies should also be
+    processed.  Transitive dependencies are restricted to relative file imports.
+-}
+data Transitivity
+    = NonTransitive
+    -- ^ Do not process transitive dependencies
+    | Transitive
+    -- ^ Process transitive dependencies in the same way
 
 -- | Path to output
 data Output = StandardOutput | OutputFile FilePath
@@ -153,15 +177,25 @@ data Output = StandardOutput | OutputFile FilePath
 -}
 data OutputMode = Write | Check
 
+-- | A check failure corresponding to a single input.
+-- This type is intended to be used with 'MultipleCheckFailed' for error
+-- reporting.
+newtype CheckFailed = CheckFailed { input :: Input }
+
 -- | Exception thrown when the @--check@ flag to a command-line subcommand fails
-data CheckFailed = CheckFailed { command :: Text, modified :: Text }
+data MultipleCheckFailed = MultipleCheckFailed
+  { command :: Text
+  , modified :: Text
+  , inputs :: NonEmpty Input
+  }
 
-instance Exception CheckFailed
+instance Exception MultipleCheckFailed
 
-instance Show CheckFailed where
-    show CheckFailed{..} =
-         _ERROR <> ": ❰dhall " <> command_ <> " --check❱ failed\n\
-        \\n\
+instance Show MultipleCheckFailed where
+    show MultipleCheckFailed{..} =
+         _ERROR <> ": ❰dhall " <> command_ <> " --check❱ failed on:\n\
+        \\n" <> files <>
+        "\n\
         \You ran ❰dhall " <> command_ <> " --check❱, but the input appears to have not\n\
         \been " <> modified_ <> " before, or was changed since the last time the input\n\
         \was " <> modified_ <> ".\n"
@@ -170,6 +204,32 @@ instance Show CheckFailed where
 
         command_ = Data.Text.unpack command
 
+        files = unlines . map format $ toList inputs
+
+        format input = case input of
+            StandardInput -> "↳ (stdin)"
+            InputFile file -> "↳ " <> file
+
+-- | Run IO for multiple inputs, then collate all the check failures before
+-- throwing if there was any failure
+handleMultipleChecksFailed
+    :: (Foldable t, Traversable t)
+    => Text
+    -> Text
+    -> (a -> IO (Either CheckFailed ()))
+    -> t a
+    -> IO ()
+handleMultipleChecksFailed command modified f xs = post =<< mapM f xs
+  where
+    post results =
+        case lefts (toList results) of
+            [] -> pure ()
+            cf:cfs -> Control.Exception.throwIO $ MultipleCheckFailed
+                { command
+                , modified
+                , inputs = fmap input (cf:|cfs)
+                }
+
 -- | Convenient utility for retrieving an expression
 getExpression :: Censor -> Input -> IO (Expr Src Import)
 getExpression censor = get Dhall.Parser.exprFromText censor . Input_
@@ -181,6 +241,35 @@ getExpressionAndHeader censor =
 
 -- | Convenient utility for retrieving an expression along with its header from
 -- | text already read from STDIN (so it's not re-read)
-getExpressionAndHeaderFromStdinText :: Censor -> Text -> IO (Header, Expr Src Import)
-getExpressionAndHeaderFromStdinText censor =
-    get Dhall.Parser.exprAndHeaderFromText censor . StdinText
+getExpressionAndHeaderFromStdinText
+    :: Censor -> String -> Text -> IO (Header, Expr Src Import)
+getExpressionAndHeaderFromStdinText censor inputName =
+    get Dhall.Parser.exprAndHeaderFromText censor . StdinText inputName
+
+{-| Convenient utility to output an expression either to a file
+    or to stdout.
+-}
+renderExpression :: Pretty a => CharacterSet -> Bool -> Maybe FilePath -> Expr Src a -> IO ()
+renderExpression characterSet plain output expression = do
+    let document = Dhall.Pretty.prettyCharacterSet characterSet expression
+
+    let stream = Dhall.Pretty.layout document
+
+    case output of
+        Nothing -> do
+            supportsANSI <- ANSI.hSupportsANSI IO.stdout
+
+            let ansiStream =
+                    if supportsANSI && not plain
+                    then fmap Dhall.Pretty.annToAnsiStyle stream
+                    else Pretty.unAnnotateS stream
+
+            Pretty.Terminal.renderIO IO.stdout ansiStream
+
+            Data.Text.IO.putStrLn ""
+
+        Just file_ ->
+            IO.withFile file_ IO.WriteMode $ \h -> do
+                Pretty.Text.renderIO h stream
+
+                Data.Text.IO.hPutStrLn h ""
