@@ -1,77 +1,107 @@
+{-# LANGUAGE CPP                 #-}
 {-# LANGUAGE DataKinds           #-}
+{-# LANGUAGE FlexibleInstances   #-}
+{-# LANGUAGE LambdaCase          #-}
+{-# LANGUAGE NamedFieldPuns      #-}
 {-# LANGUAGE OverloadedStrings   #-}
 {-# LANGUAGE RankNTypes          #-}
 {-# LANGUAGE RecordWildCards     #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE TypeOperators       #-}
-{-# LANGUAGE ViewPatterns #-}
+{-# LANGUAGE ViewPatterns        #-}
 
 {-# OPTIONS_GHC -fno-warn-orphans #-}
 
+-- Build without optimizations to prevent out-of-memory situations in Hydra CI
+{-# OPTIONS_GHC -O0 #-}
+
 module Dhall.Test.QuickCheck where
 
-import Data.Either (isRight)
-import Data.Either.Validation (Validation(..))
-import Data.Monoid ((<>))
-import Data.Void (Void)
-import Dhall (ToDhall(..), FromDhall(..), auto, extract, inject, embed, Vector)
-import Dhall.Map (Map)
-import Dhall.Core
-    ( Binding(..)
-    , Chunks(..)
-    , Const(..)
-    , Directory(..)
-    , DhallDouble(..)
-    , Expr(..)
-    , File(..)
-    , FilePrefix(..)
-    , Import(..)
-    , ImportHashed(..)
-    , ImportMode(..)
-    , ImportType(..)
-    , PreferAnnotation(..)
-    , Scheme(..)
-    , URL(..)
-    , Var(..)
+import Data.Either            (isRight)
+import Data.Either.Validation (Validation (..))
+import Data.Text              (Text)
+import Data.Void              (Void)
+import Dhall
+    ( FromDhall (..)
+    , ToDhall (..)
+    , Vector
+    , auto
+    , embed
+    , extract
+    , inject
     )
+import Dhall.Core
+    ( Binding (..)
+    , Chunks (..)
+    , Const (..)
+    , DhallDouble (..)
+    , Directory (..)
+    , Expr (..)
+    , FieldSelection (..)
+    , File (..)
+    , FilePrefix (..)
+    , FunctionBinding (..)
+    , Import (..)
+    , ImportHashed (..)
+    , ImportMode (..)
+    , ImportType (..)
+    , PreferAnnotation (..)
+    , RecordField (..)
+    , Scheme (..)
+    , URL (..)
+    , Var (..)
+    )
+import Dhall.Map              (Map)
 
-import Data.Functor.Identity (Identity(..))
-import Data.Typeable (Typeable, typeRep)
-import Data.Proxy (Proxy(..))
-import Dhall.Set (Set)
-import Dhall.Parser (Header(..), createHeader)
-import Dhall.Pretty (CharacterSet(..))
-import Dhall.Src (Src(..))
-import Dhall.Test.Format (format)
-import Dhall.TypeCheck (Typer, TypeError)
-import Generic.Random (Weights, W, (%), (:+)(..))
+import Data.Functor.Identity     (Identity (..))
+import Data.Proxy                (Proxy (..))
+import Data.Typeable             (Typeable, typeRep)
+import Dhall.Parser              (Header (..), createHeader)
+import Dhall.Pretty              (CharacterSet (..))
+import Dhall.Set                 (Set)
+import Dhall.Src                 (Src (..))
+import Dhall.Test.Format         (format)
+import Dhall.TypeCheck           (TypeError, Typer)
+import Generic.Random            (ConstrGen (..), W, Weights, (%), (:+) (..))
 import Test.QuickCheck
-    ( Arbitrary(..), Gen, Positive(..), Property, NonNegative(..)
-    , genericShrink, suchThat, (===), (==>))
+    ( Arbitrary (..)
+    , Gen
+    , NonNegative (..)
+    , Positive (..)
+    , Property
+    , genericShrink
+    , suchThat
+    , (===)
+    , (==>)
+    )
 import Test.QuickCheck.Instances ()
-import Test.Tasty (TestTree)
-import Test.Tasty.QuickCheck (QuickCheckTests(..))
-import Text.Megaparsec (SourcePos(..), Pos)
+import Test.Tasty                (TestTree)
+import Test.Tasty.QuickCheck     (QuickCheckTests (..))
+import Text.Megaparsec           (Pos, SourcePos (..))
 
 import qualified Control.Spoon
-import qualified Data.List
-import qualified Data.Sequence
-import qualified Data.SpecialValues
+import qualified Data.Char
+import qualified Data.Foldable         as Foldable
+import qualified Data.HashMap.Strict   as HashMap
 import qualified Data.HashSet
-import qualified Data.Set
-import qualified Data.Text as Text
+import qualified Data.List
 import qualified Data.Map
-import qualified Data.HashMap.Strict as HashMap
+import qualified Data.Sequence
+import qualified Data.Set
+import qualified Data.SpecialValues
+import qualified Data.Text             as Text
+import qualified Data.Time             as Time
 import qualified Dhall.Binary
 import qualified Dhall.Context
 import qualified Dhall.Core
 import qualified Dhall.Diff
 import qualified Dhall.Map
-import qualified Dhall.Parser as Parser
+import qualified Dhall.Parser          as Parser
 import qualified Dhall.Set
 import qualified Dhall.TypeCheck
 import qualified Generic.Random
-import qualified Numeric.Natural as Nat
+import qualified Lens.Family           as Lens
+import qualified Numeric.Natural       as Nat
 import qualified Test.QuickCheck
 import qualified Test.Tasty
 import qualified Test.Tasty.QuickCheck
@@ -138,18 +168,21 @@ integer =
         , (1, fmap (\x -> x - (2 ^ (64 :: Int))) arbitrary)
         ]
 
-instance Arbitrary CharacterSet where
-    arbitrary = Test.QuickCheck.elements [ ASCII, Unicode ]
-
-instance Arbitrary Header where
-    arbitrary = do
+whitespace :: Gen Text
+whitespace = do
       let commentChar =
               Test.QuickCheck.frequency
                   [ (20, Test.QuickCheck.elements [' ' .. '\DEL'])
                   , ( 1, arbitrary)
                   ]
 
-          commentText = Text.pack <$> Test.QuickCheck.listOf commentChar
+          noInteriorBlockComments text =
+              not (Text.isInfixOf "{-" text || Text.isInfixOf "-}" text)
+
+          commentText =
+              suchThat
+                  (Text.pack <$> Test.QuickCheck.listOf commentChar)
+                  noInteriorBlockComments
 
           multiline = do
               txt <- commentText
@@ -170,14 +203,24 @@ instance Arbitrary Header where
               , newlines
               ]
 
-      pure . createHeader $ Text.unlines comments
+      pure $ Text.unlines comments
 
-    shrink (Header txt) = createHeader . Text.pack <$> shrink (Text.unpack txt)
+shrinkWhitespace :: Text -> [Text]
+shrinkWhitespace "" = []
+shrinkWhitespace _  = [""]
 
-instance (Ord k, Arbitrary k, Arbitrary v) => Arbitrary (Map k v) where
+instance Arbitrary CharacterSet where
+    arbitrary = Test.QuickCheck.elements [ ASCII, Unicode ]
+
+instance Arbitrary Header where
+    arbitrary = createHeader <$> whitespace
+
+    shrink (Header text) = Header <$> shrinkWhitespace text
+
+instance (Arbitrary v) => Arbitrary (Map Text v) where
     arbitrary = do
         n   <- Test.QuickCheck.choose (0, 2)
-        kvs <- Test.QuickCheck.vectorOf n ((,) <$> arbitrary <*> arbitrary)
+        kvs <- Test.QuickCheck.vectorOf n ((,) <$> label <*> arbitrary)
         -- Sorting the fields here because serialization needs them in order
         return (Dhall.Map.fromList (Data.List.sortOn fst kvs))
 
@@ -187,12 +230,20 @@ instance (Ord k, Arbitrary k, Arbitrary v) => Arbitrary (Map k v) where
         .   Dhall.Map.toList
 
 instance (Arbitrary s, Arbitrary a) => Arbitrary (Binding s a) where
-    arbitrary =
-        let adapt = fmap ((,) Nothing)
-            f a b   = Binding Nothing "_" Nothing (adapt a) Nothing b
-            g a b c = Binding Nothing a   Nothing (adapt b) Nothing c
+    arbitrary = do
+        bindingSrc0 <- arbitrary
 
-        in  Test.QuickCheck.oneof [ lift2 f, lift3 g ]
+        variable <- Test.QuickCheck.oneof [ pure "_", label ]
+
+        bindingSrc1 <- arbitrary
+
+        annotation <- arbitrary
+
+        bindingSrc2 <- arbitrary
+
+        value <- arbitrary
+
+        return Binding{..}
 
     shrink = genericShrink
 
@@ -228,6 +279,23 @@ instance (Arbitrary s, Arbitrary a) => Arbitrary (PreferAnnotation s a) where
             , pure PreferFromCompletion
             ]
 
+instance (Arbitrary s, Arbitrary a) => Arbitrary (RecordField s a) where
+    arbitrary = lift4 RecordField
+
+    shrink = genericShrink
+
+instance (Arbitrary s, Arbitrary a) => Arbitrary (FunctionBinding s a) where
+    arbitrary = do
+        l <- label
+        type_ <- arbitrary
+        return $ FunctionBinding Nothing l Nothing Nothing type_
+
+    shrink = genericShrink
+
+instance Arbitrary s => Arbitrary (FieldSelection s) where
+    arbitrary = FieldSelection <$> pure Nothing <*> label <*> pure Nothing
+    shrink = genericShrink
+
 instance (Arbitrary s, Arbitrary a) => Arbitrary (Expr s a) where
     arbitrary =
         Test.QuickCheck.suchThat
@@ -238,17 +306,31 @@ instance (Arbitrary s, Arbitrary a) => Arbitrary (Expr s a) where
             standardizedExpression
       where
         customGens
-            :: Gen Integer    -- Generates all Integer fields in Expr
-            :+ Gen Text.Text  -- Generates all Text fields in Expr
+            :: ConstrGen "Lam" 1 (FunctionBinding s a)
+            :+ ConstrGen "Pi" 1 Text
+            :+ ConstrGen "Field" 1 (FieldSelection s)
+            :+ ConstrGen "Project" 1 (Either [Text] (Expr s a))
+            :+ Gen Integer  -- Generates all Integer fields in Expr
+            :+ Gen Text     -- Generates all Text fields in Expr
             :+ ()
         customGens =
-               integer
+               ConstrGen arbitrary
+            :+ ConstrGen label
+            :+ ConstrGen arbitrary
+            :+ ConstrGen projection
+            :+ integer
                -- 'Lam's and 'Pi's are encoded differently when the binding is
                -- the special string "_", so we generate some of these strings
                -- to improve test coverage for these code paths.
             :+ Test.QuickCheck.oneof [pure "_", arbitrary]
             :+ ()
 
+        projection =
+            Test.QuickCheck.oneof
+                [ fmap Left (Test.QuickCheck.listOf label)
+                , arbitrary
+                ]
+
         -- These weights determine the frequency of constructors in the generated
         -- Expr.
         -- They will fail to compile if the constructors don't appear in the order
@@ -293,7 +375,14 @@ instance (Arbitrary s, Arbitrary a) => Arbitrary (Expr s a) where
             % (1 :: W "Text")
             % (1 :: W "TextLit")
             % (1 :: W "TextAppend")
+            % (1 :: W "TextReplace")
             % (1 :: W "TextShow")
+            % (1 :: W "Date")
+            % (1 :: W "DateLiteral")
+            % (1 :: W "Time")
+            % (1 :: W "TimeLiteral")
+            % (1 :: W "TimeZone")
+            % (1 :: W "TimeZoneLiteral")
             % (1 :: W "List")
             % (1 :: W "ListLit")
             % (1 :: W "ListAppend")
@@ -307,8 +396,6 @@ instance (Arbitrary s, Arbitrary a) => Arbitrary (Expr s a) where
             % (1 :: W "Optional")
             % (7 :: W "Some")
             % (1 :: W "None")
-            % (1 :: W "OptionalFold")
-            % (1 :: W "OptionalBuild")
             % (1 :: W "Record")
             % (7 :: W "RecordLit")
             % (1 :: W "Union")
@@ -337,17 +424,37 @@ standardizedExpression (ListLit (Just _ ) xs) =
     Data.Sequence.null xs
 standardizedExpression (Note _ _) =
     False
-standardizedExpression (Combine (Just _) _ _) =
+standardizedExpression (Combine _ (Just _) _ _) =
     False
 standardizedExpression With{} =
     False
-standardizedExpression (Prefer PreferFromCompletion _ _) =
+standardizedExpression (Prefer _ PreferFromCompletion _ _) =
     False
-standardizedExpression (Prefer (PreferFromWith _) _ _) =
+standardizedExpression (Prefer _ (PreferFromWith _) _ _) =
     False
+-- The following three expressions are valid ASTs, but they can never be parsed,
+-- because the annotation will associate with `Merge`/`ListLit`/`ToMap` with
+-- higher precedence
+standardizedExpression (Annot (Merge _ _ Nothing) _) =
+    False
+standardizedExpression (Annot (ListLit Nothing _) _) =
+    False
+standardizedExpression (Annot (ToMap _ Nothing) _) =
+    False
+standardizedExpression (TimeZoneLiteral (Time.TimeZone _ b t)) =
+    not b && null t
+standardizedExpression (TimeLiteral (Time.TimeOfDay _ _ ss) precision) =
+    isInteger (ss * magnitude)
+  where
+    magnitude = 10 ^ precision
+
+    isInteger x = x == fromInteger (round x)
 standardizedExpression _ =
     True
 
+chooseCharacter :: (Char, Char) -> Gen Char
+chooseCharacter = Test.QuickCheck.chooseEnum
+
 instance Arbitrary File where
     arbitrary = lift2 File
 
@@ -359,9 +466,13 @@ instance Arbitrary FilePrefix where
     shrink = genericShrink
 
 instance Arbitrary Src where
-    arbitrary = lift3 Src
+    arbitrary = do
+        lift2 Src <*> whitespace
 
-    shrink = genericShrink
+    shrink (Src start end text) =
+            (Src <$> shrink start <*> pure end   <*> pure text)
+        ++  (Src <$> pure start   <*> shrink end <*> pure text)
+        ++  (Src <$> pure start   <*> pure end   <*> shrinkWhitespace text)
 
 instance Arbitrary SourcePos where
     arbitrary = lift3 SourcePos
@@ -374,8 +485,21 @@ instance Arbitrary Pos where
 instance Arbitrary ImportType where
     arbitrary =
         Test.QuickCheck.oneof
-            [ lift2 Local
-            , lift5 (\a b c d e -> Remote (URL a b c d e))
+            [ do  prefix <- arbitrary
+
+                  let nonEmptyText =
+                          fmap Text.pack (Test.QuickCheck.listOf1 arbitrary)
+
+                  components <- Test.QuickCheck.listOf nonEmptyText
+
+                  file <- nonEmptyText
+
+                  let directory = Directory{ components }
+
+                  let path = File{ file, directory }
+
+                  return (Local prefix path)
+            , lift1 Remote
             , lift1 Env
             , lift0 Missing
             ]
@@ -409,27 +533,88 @@ instance Arbitrary Scheme where
     shrink = genericShrink
 
 instance Arbitrary URL where
-    arbitrary = lift5 URL
+    arbitrary = do
+        scheme <- arbitrary
 
-    shrink = genericShrink
+        -- TODO: the authority generator could be more precise, but it currently
+        -- seems good enough.
+        let validAuthorityChar =
+                arbitrary `suchThat` \ch ->
+                    not (Data.Char.isSpace ch) && not (Data.Char.isPunctuation ch)
+
+        authority <- Text.pack <$> Test.QuickCheck.listOf validAuthorityChar
+
+        let validPChar =
+                Test.QuickCheck.frequency
+                    [ (26, chooseCharacter ('A', 'Z'))
+                    , (26, chooseCharacter ('a', 'z'))
+                    , (10, chooseCharacter ('0', '9'))
+                    , (17, Test.QuickCheck.elements "-._~!$&'*+;=:@")
+                    ]
+
+        let component = fmap Text.pack (Test.QuickCheck.listOf validPChar)
+
+        components <- Test.QuickCheck.listOf component
+
+        file <- component
+
+        let directory = Directory{ components }
+
+        let path = File{ file, directory }
+
+        let validQueryCharacters =
+                Test.QuickCheck.frequency
+                    [ (79, validPChar)
+                    , ( 2, Test.QuickCheck.elements "/?")
+                    ]
+
+        query <- Test.QuickCheck.oneof
+            [ pure Nothing
+            , fmap (Just . Text.pack) (Test.QuickCheck.listOf validQueryCharacters)
+            ]
+
+        headers <- arbitrary
+
+        return URL{..}
+
+    shrink _ = []
 
 instance Arbitrary Var where
     arbitrary =
         Test.QuickCheck.oneof
             [ fmap (V "_") (getNonNegative <$> arbitrary)
-            , lift1 (\t -> V t 0)
-            , lift1 V <*> (getNonNegative <$> arbitrary)
+            , fmap (\t -> V t 0) label
+            , V <$> label <*> (getNonNegative <$> arbitrary)
             ]
-
     shrink = genericShrink
 
+label :: Gen Text
+label = fmap Text.pack (Test.QuickCheck.listOf labelCharacter)
+  where
+    labelCharacter =
+        Test.QuickCheck.frequency
+            [ (64, chooseCharacter ('\x20', '\x5F'))
+            , (30, chooseCharacter ('\x61', '\x7e'))
+            ]
+
 binaryRoundtrip :: Expr () Import -> Property
 binaryRoundtrip expression =
         Dhall.Binary.decodeExpression (Dhall.Binary.encodeExpression denotedExpression)
     === Right denotedExpression
   where
     denotedExpression :: Expr Void Import
-    denotedExpression = Dhall.Core.denote expression
+    denotedExpression = denote' expression
+
+    denote' :: Expr a Import -> Expr b Import
+    denote' = Dhall.Core.denote . fmap denoteHttpHeaders
+
+    denoteHttpHeaders import_@(Import importHashed _)
+        | Remote url <- importType importHashed
+        = let headers' = denote' <$> headers url
+              importType' = Remote url { headers = headers' }
+              importHashed' = importHashed { importType = importType' }
+          in  import_ {importHashed = importHashed'}
+        | otherwise = import_
 
 everythingWellTypedNormalizes :: Expr () () -> Property
 everythingWellTypedNormalizes expression =
@@ -443,12 +628,17 @@ isNormalizedIsConsistentWithNormalize :: Expr () Import -> Property
 isNormalizedIsConsistentWithNormalize expression =
     case maybeProp of
         Nothing -> Test.QuickCheck.discard
-        Just prop -> prop
+        Just v -> v
   where
+      denotedExpression :: Expr Void Import
+      denotedExpression = Dhall.Core.denote expression
+
       maybeProp = do
           nf <- Control.Spoon.spoon (Dhall.Core.normalize expression)
           isNormalized <- Control.Spoon.spoon (Dhall.Core.isNormalized expression)
-          return $ isNormalized === (nf == expression)
+          -- Dhall.Core.isNormalized ignores 'Note's and other annotations.
+          -- So we do the same when checking the result of 'normalize'.
+          return $ isNormalized === (nf == denotedExpression)
 
 normalizeWithMIsConsistentWithNormalize :: Expr () Import -> Property
 normalizeWithMIsConsistentWithNormalize expression =
@@ -494,6 +684,18 @@ normalizingAnExpressionDoesntChangeItsInferredType expression =
     filterOutEmbeds :: Typer a
     filterOutEmbeds _ = Const Sort -- This could be any ill-typed expression.
 
+noDoubleNotes :: Expr () Import -> Property
+noDoubleNotes expression =
+    length
+        [ ()
+        | e <- Foldable.toList parsedExpression
+        , Note _ (Note _ _) <- Lens.toListOf Dhall.Core.subExpressions e
+        ] === 0
+  where
+    text = Dhall.Core.pretty expression
+
+    parsedExpression = Parser.exprFromText "" text
+
 embedThenExtractIsIdentity
     :: forall a. (ToDhall a, FromDhall a, Eq a, Typeable a, Arbitrary a, Show a)
     => Proxy a
@@ -510,10 +712,16 @@ embedThenExtractIsIdentity p =
 
 idempotenceTest :: CharacterSet -> Header -> Expr Src Import -> Property
 idempotenceTest characterSet header expr =
-    let once = format characterSet (header, expr)
-    in case Parser.exprAndHeaderFromText mempty once of
-        Right (format characterSet -> twice) -> once === twice
-        Left _ -> Test.QuickCheck.discard
+        not (any hasHttpHeaders expr)
+    ==> let once = format characterSet (header, expr)
+        in case Parser.exprAndHeaderFromText mempty once of
+            Right (format characterSet -> twice) -> once === twice
+            Left _ -> Test.QuickCheck.discard
+  where
+    -- Workaround for https://github.com/dhall-lang/dhall-haskell/issues/1925.
+    hasHttpHeaders = \case
+        Import (ImportHashed _ (Remote (URL { headers = Just _ }))) _ -> True
+        _                                                             -> False
 
 tests :: TestTree
 tests =
@@ -547,24 +755,23 @@ tests =
           , Test.QuickCheck.property normalizingAnExpressionDoesntChangeItsInferredType
           , adjustQuickCheckTests 10000
           )
-        , embedThenExtractIsIdentity (Proxy :: Proxy (Text.Text))
+        , ( "Parsing an expression doesn't generated doubly-nested Note constructors"
+          , Test.QuickCheck.property noDoubleNotes
+          , adjustQuickCheckTests 100
+          )
+        , embedThenExtractIsIdentity (Proxy :: Proxy (Text))
         , embedThenExtractIsIdentity (Proxy :: Proxy [Nat.Natural])
         , embedThenExtractIsIdentity (Proxy :: Proxy (Bool, Double))
         , embedThenExtractIsIdentity (Proxy :: Proxy (Data.Sequence.Seq ()))
         , embedThenExtractIsIdentity (Proxy :: Proxy (Maybe Integer))
         , embedThenExtractIsIdentity (Proxy :: Proxy (Data.Set.Set Nat.Natural))
-        , embedThenExtractIsIdentity (Proxy :: Proxy (Data.HashSet.HashSet Text.Text))
+        , embedThenExtractIsIdentity (Proxy :: Proxy (Data.HashSet.HashSet Text))
         , embedThenExtractIsIdentity (Proxy :: Proxy (Vector Double))
         , embedThenExtractIsIdentity (Proxy :: Proxy (Data.Map.Map Double Bool))
         , embedThenExtractIsIdentity (Proxy :: Proxy (HashMap.HashMap Double Bool))
         , ( "Formatting should be idempotent"
           , Test.QuickCheck.property idempotenceTest
-
-            -- FIXME: While this test is flaky, we set the number of test cases
-            -- to 0 by subtracting the default number of tests (100).
-            -- To run the test manually, use e.g.
-            --    --quickcheck-tests 1000
-          , Test.Tasty.adjustOption (subtract (QuickCheckTests 100))
+          , adjustQuickCheckTests 10000
           )
         ]
 
